/* #define TEST */
/*++++++++++++++
.IDENTIFICATION catfiles.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT    Unix
.KEYWORDS       Convert Catalogue Number
.VERSION  1.0   18-Feb-1992
.VERSION  1.1   02-Jul-1992: Refined ls edition (via tighten function)
.VERSION  1.2   18-Dec-1992: Corrected index creation bug
.			Symptom: Crash when first file commented out
.VERSION  1.3   27-Feb-1999: Accept empty files....
.VERSION  1.4   04-May-2002: Keep a SINGLE line in NOT_
.VERSION  1.5   02-Nov-2004: Accept fbytes terminated by b K M G
.VERSION  1.6   04-Mar-2006: Test whether inxistent or inaccessible
.VERSION  1.7   08-May-2006: Pb if lines not trimmed...
.VERSION  1.8   25-Aug-2009: in 'anafile' package
.COMMENTS       Interpret .files, and generates the following output:
		Col.  1: + (chosen copy) or Number
		Col.  2: Catalogue designation  e.g. II/140/f1
		Col.  3: Complete Filename
		Col.  4: Records
		Col.  5: Blksize
		Col.  6: Lrecl
		Col.  7: Bytes
		Col.  8: "bin" indication for a Binary File
		Col.  9: Bytes as Unix File
		Col. 10: Bytes as Compressed
		Col. 11: Designation (Comments)
  Example of a "ls" output :
         1         2         3         4         5         6         7
12345678901234567890123456789012345678901234567890123456789012345678901234567890
            File:     NrecxLrec      FB     Unix        Z  Explanation
          V/3/f1:     3522x144    507168     973b     351b 5003-Data FK4 Sup
V/3       =====>:     3522x144    507168     973b     351b 
---------------*/

#include <stdio.h>
#include <ctype.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#ifndef O_BINARY
#define O_BINARY 0
#endif

#ifdef MSDOS	/* ======= MSDOS Definitions ===== */
#define DIRSEP '\\'	/* / for Unix \ for MS-DOS */
#define FILES  "-FILES"
#else
#define DIRSEP '/'	/* / for Unix \ for MS-DOS */
#include <unistd.h>	/* to get status of file   */
#include <sys/stat.h>	/* to get status of file   */
#include <sys/types.h>	/* to get status of file   */
#define FILES  ".files"
#endif

typedef int (*SORT_FCT)() ;
static char FS	= ':';	/* Separation character between fields */
/*static int  lname = 16;  Length of "Name" field in ls	       */
static char thecat[40];
static char thedir[128];	/* Directory containing the Files */
#define MAX(x,y)        ((x) >= (y) ? (x) : (y))
#define ITEMS(a)        sizeof(a)/sizeof(a[0])

typedef struct {
	char *fd;	/* File Designation 	*/
	char *fm;	/* Media		*/
	char *fc;	/* File Characteristics */
	char *explain;	/* Explanations		*/
	int  lrec;
	unsigned blk2;	/* log2(blksize)	*/
	unsigned long nrec;
	unsigned long fbytes;
	unsigned long ubytes;
	unsigned long Zbytes;
   }	CFILE;

static CFILE *cfiles;
static char *input;
static char *obsolete = (char *)0;
static char *NOT_     = (char *)0;
static unsigned long file_bytes;

static int verbop   = 0;  /* Verbose option       */

static char *rm[] = {"X", "I", "II", "III", "IV", 
		"V", "VI", "VII", "VIII", "IX" };
static char pgm[] = "catfiles";
static char *cats;	/* getenv("CATS") */
static char unit[] = "b\011k\012m\024g\036";	/* Block Kilo Mega Giga */
static char usage[] = 
	"Usage: %s [-dx] [-ls] [-c5] [catalogues...]\n";

/*==================================================================
		Internal Routines
 *==================================================================*/
static void error(text1,text2)
/*++++++++++++++++
.PURPOSE  Error exit
.RETURNS  -
-----------------*/
  char *text1, *text2;     /* IN: Text before error */
{
  
    fprintf(stderr, "*** %s: %s%s\n", pgm, text1, text2);
    fprintf(stderr, usage, pgm);

    exit(1);
}

static void help()
/*++++++++++++++++
.PURPOSE  Display help, then exit
.RETURNS  To OS
-----------------*/
{
  static char *text[] = {
	"  catalog: Catalogues to look for file classification",
	"The output is made of columns separated by colons as follows",
	"  Col.  1: + (chosen copy) or Number",
	"  Col.  2: Catalogue designation  e.g. II/140/f1",
	"  Col.  3: Complete Filename",
	"  Col.  4: Records",
	"  Col.  5: Blksize",
	"  Col.  6: Lrecl",
	"  Col.  7: Bytes",
	"  Col.  8: 'bin' indication for a Binary File",
	"  Col.  9: Bytes as Unix File",
	"  Col. 10: Bytes as Compressed",
	"  Col. 11: Designation (Comments)",
	"Options:",
	"   -dx : use character x as delimiter (default :) ",
	"   -ls : Make a list of the files making the catalogue", 
	"   -c5 : Convert into a 5-columns name:file_name:rxl:b/Z:Desc",
  };
  int i;
  
    fprintf(stderr, usage, pgm);
    for (i=0; i < ITEMS(text); i++)       fprintf(stderr, "%s\n", text[i]);

    exit(0);
}

/*==================================================================*/
 
static int strloc(text, c)
/*++++++++++++++++
.PURPOSE  Locate specified character
.RETURNS  Index of located char
-----------------*/
  char *text;       /* IN: String to interpret */
  int  c;		/* IN: Char to look for */
{
  char *s;

    for ( s = text; *s; s++)	if (*s == c)	break;
    return(s-text);
}

static int strcopy(dest, sour)
/*++++++++++++++++
.PURPOSE  Copy, returning the length
.RETURNS  Length
-----------------*/
  char *dest; 	/* OUT: Destination */
  char *sour;	/* IN: Source */
{
  char *d, *s;
    for (d=dest, s=sour; *s; s++, d++)	*d = *s;
    *d = 0;
    return(d-dest);
}

#if 0
static int strbloc(text, c)
/*++++++++++++++++
.PURPOSE  Locate specified character, from right to left
.RETURNS  Index of located char
-----------------*/
  char *text;       /* IN: String to interpret */
  int  c;		/* IN: Char to look for */
{
  char *s;
    for ( s = text + strlen(text); --s >= text; )	if (*s == c)	break;
    return(s-text);
}
#endif

static int tighten(str, len)
/*++++++++++++++++
.PURPOSE  Compress: suppress spaces if str is too large
.RETURNS  Length of Compressed
-----------------*/
  char *str;       /* IN: String to interpret */
  int  len;	/* IN: Maximal String Length */
{
  char *p, *s;
  int ls;
    ls = strlen(str);
    for (p = s = str; (*p) && (ls > len); p++) {
	if ((p[0] == ' ') && (p[1] == ' ')) ls--;
	else *(s++) = *p;
    }
    if (p != s)  strcopy(s, p);
  
    return(ls);
}

static int compact(text)
/*++++++++++++++++
.PURPOSE  Compress (suppress spaces)
.RETURNS  Length of Compressed
-----------------*/
  char *text;       /* IN: String to interpret */
{
  char *p, *s, c;

    for (p = s = text, c = ' '; *s; s++) {
	if (isspace(*s)) {
		if (isspace(c))	continue;
	}
	c = *s;
	*(p++) = c;
    }
    if (isspace(c) && (p != text)) --p;
    *p = 0;

    return(p-text);
}

/*==================================================================*/
static char *trN(numeric)
/*++++++++++++++++
.PURPOSE  Convert Numeric Catalogue to e.g. V/NNN
.RETURNS  Converted
-----------------*/
  char *numeric;       /* IN: String to interpret */
{
  char *p;
  int nc, cl;

    for (p = numeric; isspace(*p); p++) ;
    nc = strcopy(thecat, p);
    if (! isdigit(*p))	{ 
	if (! nc) return(thecat);
	p = &thecat[nc-1];
	if (*p == DIRSEP)	*p = 0;
	return(thecat); 
    }

    cl = *(p++) - '0'; nc = atoi(p); 
    if(cl > 0)	while(isdigit(*p)) p++;

    if (cl >= 10)	sprintf(thecat, "%d%c%d%s", cl, DIRSEP, nc, p);
    else if (cl >= 1)	sprintf(thecat, "%s%c%d%s", rm[cl], DIRSEP, nc, p);
    else		sprintf(thecat, "%s%c%s",   rm[0],  DIRSEP, p);

    p = thecat + strlen(thecat) - 1;
    if (*p == DIRSEP)	*p = 0;

    return(thecat);
}

/*==================================================================*/
static int idx(text, text_len) 
/*++++++++++++++++
.PURPOSE  Create an array of pointers to text
.RETURNS  Number of items
-----------------*/
  char *text;		/* IN: Text to index  */
  unsigned text_len;	/* IN: Length of Text */
{
  int nl, i, j;
  char *p, *pe, *pc, *next;
  CFILE cf;

	/* Count number of interesting lines */
    for (nl=0, p=text, pe = text+text_len; p < pe; p += i) {
	i = strloc(p, '\n')+1;
	if((i>1) && isspace(*p)) nl++;
    }
    cfiles = (CFILE *)malloc(i=nl*sizeof(CFILE));

	/* Fill the array */
    memset(&cf, 0, sizeof(cf)); 	/* No file specified */
    for (i=0, nl=0, p=text; p < pe; p = next) {
	pc = p + strloc(p, '\n'); next = pc+1;
	while ((pc >= p) && isspace(*pc)) *(pc--) = 0;
	switch(*p) {
	case '#': 
		continue;
	case '\0': 
		cf.fd = (char *)0;	/* No file specified */
		continue;
	case '\t': case ' ':
		if (!cf.fd) continue;
		pc = p; p += strlen(p);
		pc[strloc(pc, '(')] = 0; compact(pc);
		cfiles[i] = cf;
		cfiles[i].fm = pc;
#ifdef TEST
		printf("Created Record #%d\n  (fd)%s\n  (fm)%s  (fc)%s\n", 
			i, cf.fd, pc, cf.fc);
#endif
		i++;
		continue;
	default:
		cf.fd = p; p += strloc(p, FS); *(p++) = 0;
		cf.fc = p; p += strloc(p, FS); *(p++) = 0;
		pc    = p; p += strloc(p, FS); *(p++) = 0;
		cf.explain = p; p += strlen(p);
		cf.blk2 = 0;
		while (*pc == ' ') pc++;
		cf.ubytes = atol(pc); 
		while (isdigit(*pc)) pc++;
		if (isalpha(*pc)) { 
		    j = strloc(unit, tolower(*pc));
		    if (unit[j]) cf.blk2 = unit[j+1];
		}
		pc += strloc(pc, DIRSEP); if (*pc) pc++;
		cf.Zbytes = atol(pc);
		pc = cf.fc; 
		cf.nrec = atol(pc);   pc += strloc(pc, 'x'); if (*pc) pc++;
		cf.lrec = atol(pc); while(isspace(*pc)) pc++;
		while (isdigit(*pc)) pc++; compact(pc); cf.fc = pc;
		while (*pc != 'b') {
			if (!*pc) break;
			pc += strloc(pc, ' ');
			if (*pc) pc++;
		}
		if (*pc == 'b') cf.fc = pc;
		compact(cf.fd); compact(cf.explain);
		cf.fbytes = cf.nrec * cf.lrec;
		cf.fm = "?";
	}
    }
    nl = i;
    return(nl);
}

/*==================================================================*/
static int fdiff(c1, c2)
/*++++++++++++++++
.PURPOSE  Compare two strings, taking care of numeric values
.RETURNS  Difference
-----------------*/
  char *c1;	/* IN: First file     */
  char *c2;	/* IN: Second file     */
{
  int diff;
  
    if (c1 == c2)	return(0);	/* Strings are obviously identical ! */
    for (diff=0; (diff == 0) && (*c1 || *c2); ) {
	if (isdigit(*c1) && isdigit(*c2)) {
		diff = atoi(c1) - atoi(c2);
		while (isdigit(*c1)) c1++;
		while (isdigit(*c2)) c2++;
		continue;
	}
	diff = *(c1++) - *(c2++);
    }
    return(diff);
}

/*==================================================================*/
static int cfcmp(cf1, cf2)
/*++++++++++++++++
.PURPOSE  Compare two files, order them
.RETURNS  Difference
-----------------*/
   CFILE *cf1;	/* IN: First file     */
  CFILE *cf2;	/* IN: Second file     */
{
  int diff;
  char c1, c2;

    if ((diff = fdiff(cf1->fd, cf2->fd))) return(diff);

    c1 = *(cf1->fm); if (c1 == '.') c1 = 1;	/* Very First Char */
    c2 = *(cf2->fm); if (c2 == '.') c2 = 1;	/* Very First Char */

    if ((diff = (c1 - c2)))	return(diff);

    return(strcmp(cf1->fm + 1, cf2->fm + 1));
}

/*==================================================================*/
static char *loadfile(fn, first_line) 
/*++++++++++++++++
.PURPOSE  Load a file into memory
.RETURNS  The memory copy. The number of bytes is the static file_bytes
-----------------*/
  char *fn;		/* IN: File to load */
  int first_line; 	/* IN: non-zero for loading first line only */
{
  char *pf;
  int nf;

#ifdef TEST
    printf("Trying to load file %s\n", fn);
#endif
    pf = (char *)0;
    nf = open(fn, O_BINARY);
    if (nf <= 0)	return(pf);
    file_bytes = lseek(nf, 0L, 2); lseek(nf, 0L, 0); 
    pf = malloc(file_bytes+2); 

    if (file_bytes == 0) pf[file_bytes++] = '\n'; 
    else read(nf, pf, file_bytes); 
    close(nf);

    /* Be sure file is terminated by a newline... */
    if (pf[file_bytes-1] != '\n') pf[file_bytes++] = '\n' ;
    pf[file_bytes] = 0;

    /* Keep only first line */
    nf = strloc(pf, '\n') ;
    if (first_line) pf[nf] = 0;

    return(pf);
}

static int getfiles(fn) 
/*++++++++++++++++
.PURPOSE  Load .files, and sort it.
.RETURNS  Number of files, -1 for Errors
-----------------*/
  char *fn;	/* IN: File to examine */
{
  unsigned nc;
  int nf, i, k;

	/* Open file, & Load it */

    input = loadfile(fn, 0);
    if (!input)	{ 
#ifdef MSDOS
  	printf("+++: %s : =====>Inexistant or Unavailable <=====\n", thecat);
#else
      struct stat buf ;
        i = strlen(fn);
	if ((i>8) && (strcmp(fn+i-7,"/.files") == 0)) {	/* Try directory */
	    i -= 7;
	    fn[i] = 0;
	    k = stat(fn, &buf);
	    if ((k == 0) && S_ISDIR(buf.st_mode)) ;	/* Success, is dir */
	    else k = 1;
	    printf("+++: %s : =====> %s <=====\n", thecat, 
		k ? "Inexistant" : "Inaccessible");
	    fn[i] = '/';
	}
#endif
	return(-1);
    }
    nc = file_bytes;

	/* Create the Index of CFILE */
  
    nf = idx(input, nc);

	/* Sort now the index: Name First, then according to filename */

    qsort(cfiles, nf, sizeof(CFILE), (SORT_FCT)cfcmp);

    return(nf);
}

/*==================================================================*/
static int ls(nf) 
/*++++++++++++++++
.PURPOSE  ls the files
.RETURNS  Number of files.
.REMARK   .fix ARE IGNORED 
-----------------*/
  int nf;	/* IN: Number of files (elements in cfiles) */
{
  char *fd;
  CFILE *cf;
  int i, j, Lrec, len;
  long total[3], Nrec;	/* total[1] and total[2] in blocks */
  long blocks[2];	/* Length in blocks */
  char thefile[80], edited[256];
  char al[20], an[20] ;


    fd = ""; Nrec = total[0] = total[1] = total[2] = 0; Lrec = 0;
    printf("%20s:%8sx%-4s%8s %8s %8s  %s\n", "File", "Nrec","Lrec", 
	"FB","Unix","Z","Explanation");
    for (cf = cfiles, i = nf; --i >= 0; cf++) {
	if (cf->fd == fd)  continue;
	/* if (strcmp(cf->fd, fd) == 0) continue; */
	fd = cf->fd; len = strlen(fd);
	if (len > 4) {	/* Check against .fix */
		if (strcmp(fd + len - 4, ".fix") == 0) continue;
	}
	sprintf(thefile, "%s%c%s", thecat, DIRSEP, fd);
				/* Replace zero by a dot */
	al[0] = an[0] = '.' ; al[1] = an[1] = 0 ;
	if (cf->nrec) sprintf(an, "%lu", cf->nrec) ;
	if (cf->lrec) sprintf(al, "%d", cf->lrec) ;
	j = (int)cf->blk2 - 9;
	blocks[0] = blocks[1] = 1;
	if (j>0) {
	    blocks[0] = cf->ubytes << j;
	    blocks[1] = cf->Zbytes << j;
	}
	else {
	    j = -j;
	    blocks[0] = cf->ubytes >> j;
	    blocks[1] = cf->Zbytes >> j;
	    j = (1<<j)-1;
	    if (cf->ubytes) {
		blocks[0] += 1;
		if (cf->ubytes&j) blocks[0] += 1;
	    }
	    if (cf->Zbytes) {
		blocks[1] += 1;
		if (cf->Zbytes&j) blocks[1] += 1;
	    }
	}
	sprintf(edited, "%20s: %7sx%-4s %8lu %7ldb %7ldb", thefile, an, al, 
		cf->fbytes, blocks[0], blocks[1]);
	tighten(edited, 61);
	printf("%s %s\n", edited, cf->explain);
	Nrec += cf->nrec; if (Lrec < cf->lrec) Lrec = cf->lrec;
	total[0] += cf->fbytes; 
	total[1] += blocks[0];
	total[2] += blocks[1];
    }
    if (obsolete) printf("%-13s =====> obsolete <===== %ld %s\n", 
	thecat, total[1], obsolete);
    else if (NOT_) printf("%-13s ==>NOT AVAILABLE <=== %ld %s\n", 
	thecat, total[1], NOT_);
    else printf("%-13s %6s: %7ldx%-4d %8ld %7ldb %7ldb %s\n", 
	thecat, "=====>", Nrec, Lrec, total[0], total[1], total[2], "Total");

    return(nf);
}
  
/*==================================================================*/
static int cv1(nf) 
/*++++++++++++++++
.PURPOSE  Edit the .files as 11 columns
.RETURNS  Number of files.
.REMARK   .fix ARE IGNORED 
-----------------*/
  int nf;	/* IN: Number of files (elements in cfiles) */
{
  char *fd;
  CFILE *cf;
  int i, k=0, len;


    fd = "";
    for (cf = cfiles, i = nf; --i >= 0; cf++) {
	if (cf->fd != fd)  fd=cf->fd, k=0;
	else k++;
	len = strlen(fd);
	if (len > 4) {	/* Check against .fix */
		if (strcmp(fd + len - 4, ".fix") == 0) continue;
	}
	printf("%c%c", k ? '0'+k : '+', FS);	/* Col. 1 */
	printf("%s%c%s%c", thecat, DIRSEP, 
				  cf->fd, FS);  /* Col. 2 */
	if (strcmp(cf->fm, ".") == 0)		/* Col. 3 */
		printf("%s%s%c", thedir, fd, FS);
  	else	printf("%s%c", cf->fm, FS);
  	printf("%ld%c", cf->nrec, FS);		/* Col. 4 */
  	printf("Blk%c", FS);			/* Col. 5 */
  	printf("%d%c", cf->lrec, FS);		/* Col. 6 */
  	printf("%ld%c", cf->fbytes, FS);	/* Col. 7 */
  	printf("%s%c", cf->fc, FS);		/* Col. 8 */
  	printf("%ld%c", cf->ubytes, FS);	/* Col. 9 */
  	if (cf->Zbytes) 
	      	printf("%ld", cf->Zbytes);	/* Col.10 */
	putchar(FS);
  	printf("%s\n", cf->explain);		/* Col.11 */
    }
    if (obsolete) {
  	printf("+++: %s : =====>obsolete<=====\n%s\n", thecat, obsolete);
  	return(-1);
    }
    return (nf);
}

/*==================================================================*/
static int cv5(nf) 
/*++++++++++++++++
.PURPOSE  Edit the .files as 5 columns
.RETURNS  Number of files.
.REMARK   .fix ARE IGNORED 
-----------------*/
  int nf;	/* IN: Number of files (elements in cfiles) */
{
  char *fd;
  CFILE *cf;
  int i, len;


    fd = "";
    for (cf = cfiles, i = nf; --i >= 0; cf++) {
	if (cf->fd == fd)  continue;
	fd=cf->fd; len = strlen(fd);
	if (len > 4) {	/* Check against .fix */
		if (strcmp(fd + len - 4, ".fix") == 0) continue;
	}
	printf("%s%c%s%c", fd, FS, cf->fm, FS);  		/* Col. 1+2 */
	printf("%ldx%d %s%c", cf->nrec, cf->lrec, cf->fc, FS);  /* Col. 3   */
	printf("%ld/%ld%c", cf->ubytes, cf->Zbytes, FS);  	/* Col. 4   */
  	printf("%s\n", cf->explain);				/* Col. 5   */
    }
    if (obsolete) {
  	printf("+++: %s : =====>obsolete<=====\n%s\n", thecat, obsolete);
  	return(-1);
    }
    return (nf);
}

/*==================================================================
		Program
 *==================================================================*/
int main (int argc, char **argv)
{
  char line[160], *p, *parm, *e;
  int  len;
  static int (*ed)() = cv1;	/* Function for edition */

    if ((p = getenv("CATS"))) {
  	len = strlen(p); 
  	cats = malloc(len+2); strcopy(cats, p); 
  	p = cats + len;
  	if (p[-1] != DIRSEP) *(p++) = DIRSEP, *p = 0;
    } else cats = "";
    /* printf("cats=%s\n", cats); */
  
	/* Check the Arguments */
#ifdef TEST
    cv1(getfiles("3078.fil")); free(input); free(cfiles);
    ls(getfiles("3078.fil"));
#endif
  
    while (-- argc > 0) {
	parm = *++argv;
	if (*parm == '-') switch (*++parm) {
	  case 'F': case 'd':
		FS = *++parm;
		continue;
	  case 'v':                     /* Verbose      */
		verbop = 1;
		/* fprintf(stderr, "... %s: Examine argument ", pgm+4); */
		continue;
	  case 'c':
		switch(*++parm) {
		  case '5': ed = cv5; continue;
		  default: goto bad_option;
		}
	  case 'l':			/* lscat	*/
		ed = ls;
		continue;
	  case 'h':                     /* Help         */
		help();
		continue;
	  default:
	  bad_option:
		error("Bad option: ", parm);
	} else {	/* A name starting with punctuation is not a cat. */
		p = line;
		if (isalnum(*parm)) p += strcopy(p, cats); 
		p += strcopy(p, trN(parm)); *(p++) = DIRSEP; e = p;
		*e = 0; 
		p = thedir;	/* Fill the Directory name */
		if (*line != DIRSEP) {
		    if (!getcwd(thedir, sizeof(thedir))) perror("#***getcwd");
		    p = thedir + strlen(thedir);
		    *(p++) = DIRSEP;
	    	}
		strcopy(p, line);
		strcopy(e, "=obsolete="); obsolete = loadfile(line, 1);
		strcopy(e, "=NOT_PUBLIC="); NOT_ = loadfile(line, 1);
		strcopy(e, FILES);
		(*ed)(getfiles(line));
		free(input); free(cfiles);
	}
    }

    exit(0);

}
