/*++++++++++++++
.IDENTIFICATION lss.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT    Unix
.KEYWORDS       Convert Catalogue Number
.VERSION  1.0   23-Jan-2000
.VERSION  1.1   17-Feb-2000: message when directory not readable
.VERSION  1.2   02-Jun-2007: Cosmetics (-Wall)
.VERSION  1.3   05-Dec-2012: If file specifed => accept a .
.COMMENTS       List file & directories newer than some limit
---------------*/

#define VERSION  "1.3"

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>

#ifndef int4
#define int4 int
#endif

#ifndef LEN_CWD		/* Max length of files */
#define LEN_CWD	512
#endif

#define MAX(x,y)        ((x) >= (y) ? (x) : (y))
#define ITEMS(a)        sizeof(a)/sizeof(a[0])
typedef int (*SORT_FCT)(void *, void *) ;

static int  optR;		/* Allowed depth */
static char opta;
static char optv;
static char optu;
static char optl;
static time_t mtime = -1;	/* -m option */

/* static char *rm[] = {"X", "I", "II", "III", "IV", 
		"V", "VI", "VII", "VIII", "IX" }; */
static char *pgm;

static char usage[] = "\
Usage: lss [-aluvR[n]] [-mUnixTime] [file...]\n\
" ;
static char help[] = "\
   -a: List also the hidden files (starting by .)\n\
   -l: List the file size, protections, modification times\n\
   -R: Recursive list of files, eventually limited to depth n\n\
   -u: List also the last access time (default is modification time only)\n\
   -v: List also the files, even when not readable by everybody\n\
   -m: files more recent than the specified date (Unix seconds)\n\
" ;
  
/*==================================================================
		Internal Routines
 *==================================================================*/

#if 0
static int strbloc (text, c)
/*++++++++++++++++
.PURPOSE  Locate (backwards) specified character
.RETURNS  Index of located char / -1
-----------------*/
  char *text;       /* IN: String to interpret */
  int  c;		/* IN: Char to look for */
{
    char *s;
    for (s = text + strlen(text); --s >= text; ) if (*s == c)    break;
    return(s-text);
}

static int strloc(text, c)
/*++++++++++++++++
.PURPOSE  Locate specified character
.RETURNS  Index of located char
-----------------*/
  char *text;       /* IN: String to interpret */
  int  c;		/* IN: Char to look for */
{
  char *s;
    for ( s = text; *s; s++)	if (*s == c)	break;
    return(s-text);
}
#endif

static char *basename(name, sub)
/*++++++++++++++++
.PURPOSE  Extract the 'name' from a complete path
.RETURNS  Converted
-----------------*/
  char *name;	/* IN: Full file name */
  int  sub;	/* IN: Number of Subdirectories to take into account */
{
  char *p;
    for (p = name + strlen(name); --p >= name; ) {
	if (*p == '/') { if (--sub < 0) break; }
    }
    return(++p);
}

/*==================================================================*/
#if 0
static int filcmp(c1, c2)
/*++++++++++++++++
.PURPOSE  Compare two strings, taking care of numeric values
.RETURNS  Difference
-----------------*/
  char **c1;	/* IN: First file     */
  char **c2;	/* IN: Second file     */
{
  char *p, *q;
  int diff;
  
    p = *c1; q = *c2;
    if (p == q)	return(0);	/* Strings are obviously identical ! */
    for (diff=0; (diff == 0) && (*p || *q); ) {
	if (isdigit(*p) && isdigit(*q)) {
	    diff = atoi(p) - atoi(q);
	    while (isdigit(*p)) p++;
	    while (isdigit(*q)) q++;
	    continue;
	}
	diff = *(p++) - *(q++);
    }
    return(diff);
}
#endif

static void edit_buf(struct stat *buf, int level, char *name)
/*++++++++++++++++
.PURPOSE  Edit one entry
.RETURNS  ---
-----------------*/
{
  char tf ;
    if (S_ISREG(buf->st_mode)) tf = '-' ;
    else if (S_ISDIR(buf->st_mode)) tf = 'd' ;
    else if (S_ISLNK(buf->st_mode)) tf = 'l' ;
    else if (S_ISCHR(buf->st_mode)) tf = 'c' ;
    else if (S_ISBLK(buf->st_mode)) tf = 'b' ;
    else if (S_ISFIFO(buf->st_mode)) tf = 'f' ;
    else if (S_ISSOCK(buf->st_mode)) tf = 's' ;
    else tf = '?' ;
    printf("%2d %c", level, tf) ;
    if (optl) printf("%03o %10ub %10us", 
       buf->st_mode&0777, (int4)buf->st_size, (int4)buf->st_mtime) ;
    if (optu) printf(" %10da", (int4)buf->st_atime) ;
    printf(" %s\n", name) ;
}

static int exec_ls(char *name, int level)
/*++++++++++++++++
.PURPOSE  Get directories with specified name
.RETURNS  Number of edited files
-----------------*/
{
  static int recu = 0 ;			/* Recusrsivity Counter	*/
  DIR *dirp; struct dirent *dire;
  struct stat buf;
  char cwd[LEN_CWD] ;
  char *p;
  int ns, i ;

    getcwd(cwd, sizeof(cwd));
    ns = 0;
    if (lstat(name, &buf) < 0) {
      	fprintf(stderr, "****lss ") ;
      	p = name ? name : "(nil)" ;
      	perror(p) ; 
        exit(1) ; 
    }

    if ((!optv) && ((buf.st_mode & 4) == 0)) {	/* Unreadable */
	if (recu == 0) fprintf(stderr, 
	    "++++lss: file not world readable (use option -v): %s\n", name) ;
        return(0) ;
    }
    recu++ ;

    if (S_ISDIR(buf.st_mode)) {
    	edit_buf(&buf, level, name) ;

	i = level+1 ;
	if (i > optR) { recu--; return(1) ; }

	dirp = opendir(name) ;
	if (!dirp) {
	    perror("\n++++");
	    fprintf(stderr, "++++Can't opendir '%s' from '%s'\n", name, cwd);
	    recu-- ;
	    return(0);
	}
	if (chdir(name)) {
	    perror("\n++++");
	    fprintf(stderr, "++++Can't chdir from '%s' to '%s'\n", cwd, name);
	    closedir(dirp) ;
	    recu-- ;
	    return(0);
	}

        while ((dire = readdir(dirp))) {
	    if (dire->d_name[0] == '.') {
	        if (!dire->d_name[1]) continue ;	/* . directory */
		if ((dire->d_name[1] == '.') && (!dire->d_name[2]))
		    continue ;
		if (!opta) continue ;
	    }
	    ns += exec_ls(dire->d_name, i) ;
	}
	putchar('\n') ;
	closedir(dirp) ;
	chdir(cwd) ;
    }
    else if ((mtime > 0) && (buf.st_mtime < mtime)) {
	recu-- ;
        return(0) ;
    }
    else {
        if ((*name == '.') && (!opta) && (recu>1)) { recu--; return(0) ; }
        ns++, edit_buf(&buf, level, name) ;
    }

    recu-- ;
    return(ns) ;

}

/*==================================================================
		Program
 *==================================================================*/
int main (int argc, char **argv)
{
  int nc;
  char *p ;

    nc  = 0;

	/* Check the Arguments */
  
    pgm = basename(*argv, 0);

    while (-- argc > 0) {
	p = *++argv;
	if (*p == '-') for (++p; *p; p++) switch (*p) {
	  case 'h': case '?':
	  	printf("%s%s", usage, help) ;
		exit(0) ;
	  case 'a' :            
		opta = 1 ;
		continue ;
	  case 'R' :		/* Recursive */
		optR = 9999 ;
		if (isdigit(p[1])) for (optR=0; isdigit(p[1]); p++) 
		    optR = optR*10 + (p[1]-'0') ;
		continue ;
	  case 'l' :            /* Just 1 catalog/line	*/
		optl = 1 ;
		continue ;
	  case 'u' :            /* Just 1 catalog/line	*/
		optu = 1 ;
		continue ;
	  case 'v' :            /* Verbose, all files	*/
	        if (strncmp(p, "ver", 3) == 0) {
	            printf("lss (CDS) Version %s\n", VERSION);
		    exit(0);
	        }
		optv = 1 ;
		continue ;
	  case 'm' :
		mtime = 0;  	/* Limit in Unix time 	*/
		while (isdigit(p[1])) mtime = mtime*10 + (p[1]-'0'),p++ ;
		continue;
	  default:
	  /*bad_option:*/
		fprintf(stderr, "****Bad option: %s\n%s", p, usage);
		exit(1) ;
	} else {
	    if (isdigit(*p) && (mtime == 0)) {
	        mtime = atol(p) ;
		continue ;
	    }
	    exec_ls(p, 0) ;
	    nc++ ;
	    continue ;
	}
    }

    if (nc == 0) { /* Just execute from the current directory */
	if (!optR) optR = 1 ;
        exec_ls(".", 0) ;
    }
    exit(0) ;
}
