/*++++++++++++++
.IDENTIFICATION preadme.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT    Unix, CDS Catalogues
.KEYWORDS       Table, FITS
.VERSION  1.0   01-Dec-1995
.VERSION  1.1   26-Apr-1996: -cgi option
.VERSION  1.2   08-Apr-1999: -a option / Add link to Files
.VERSION  1.2   07-May-1999: accept - option
.VERSION  1.3   20-Jul-1999: -csv option
.VERSION  1.4   22-Jun-2000: -par option
.VERSION  1.5   09-Dec-2000: -t option  *** A FAIRE ***
.VERSION  1.6   20-Apr-2001: Finalized the -par (ADS) option
.VERSION  1.7   11-Jul-2001: Added %d (date of file) %O (Objects)
.VERSION  1.8   22-Oct-2001: Added -b  option (for bibcode)
.VERSION  1.9   26-Mar-2002: Accept a tables also in standard mode
.VERSION  1.91  01-May-2004: -keys option
.VERSION  1.92  03-Jun-2004: Accept * in -par
.VERSION  1.93  02-Jun-2007: Cosmetics (-Wall)
.VERSION  1.94  30-Sep-2007: Edit ALL files (except description file)
.VERSION  1.95  05-Jun-2008: Use journals.h
.VERSION  1.96  15-Dec-2009: 64-bit
.VERSION  1.97  07-May-2011: Pb with DOS files!
.VERSION  1.98  09-Jan-2013: Use quiet by default
.COMMENTS       Print out the contents of ReadMe
---------------*/

#define VERSION "1.98"

#ifndef DEBUG
#define DEBUG 0
#endif
#ifndef TEST
#define TEST 0
#endif
/*static int vartest = TEST;*/

#include <catstd.h>
#include <journals.h>
#include <filutil.h>
#include <strings.h>	/* Standard String manipulations */
#include <string.h>	/* Standard String manipulations */
#include <fcntl.h>		/* File Control	   */
#ifndef O_BINARY
#define O_BINARY 0
#endif

/* static char journals[] = 	-- List of Journals coded in VizieR
  "ApJ/ApJS/A+A/A+AS/AJ/PASJ/PASP/MNRAS/AZh/PAZh/other/" ;
*/

#define MAX(x,y)        ((x) >= (y) ? (x) : (y))
#define MIN(x,y)        ((x) >= (y) ? (y) : (x))
#define ITEMS(a)        sizeof(a)/sizeof(a[0])

static char verbop = DEBUG;	/* Verbose option       */
static char oped   = 0;		/* Edition Option	*/
static int  errors = 0;
static int  max_len = 255 ;	/* Maximal length of Explanataion / Note */

static FILE *fout ;		/* Output file		*/
static char *ddat;		/* Directory with data	 */
static char *intro;		/* Name of Description File */
static char cwd[81];		/* Working directory	*/

/*static char catid[80];-- Catalogue designation */
/*static char *Tbf, *Trm;-- \\bf and \\rm flags	 */
/*static char Tget[2];	-- (A) (F) Anchors	 */
/*static char rule_char = '-';	-- Rule Char	 */
static char csv ;	/* -csv option ...       */
static char *atable ;	/* Table to explain..	 */
static char  *bcode ;	/* -b option		 */

#if DEBUG || TEST
static char inbyte;
#endif

static char *keys[] = {	/* List of Keys to retrieve */
    "ADC", 
    "Mission", 
    "wl", 
    "", 
    "\\cUsualName",
    "\\cObjectType",
    "\\cCenters",
    "\\cSimbad",
    "\\cServer",
    "\\vizMore",
    "\\vizIgnoreTables",
    "\\vizDisplayColumns",
};

/* List of Keys for Parfile -- see file ~cats/lib/bibdef.tex	*/
static char parkeys[] = "\
%A = Authors\n\
%B = Abstract (or Description)\n\
%b = Bulletin CDS where Catalogue was Announced\n\
%c = Catalogue Type (Observation/[Biblio/Critical/General]Compilation/Model)\n\
%D = Dates when catalogue was added / removed\n\
%d = Date of last modification (change in the description)\n\
%F = List of Files\n\
%I = CDS/ADC Identification of Catalogue\n\
%J = Reference of Catalogue (J= uses SIMBAD/NED Code)\n\
%K = Keyword(s) of Catalogue (ADC_Keywords and Mission_Name)\n\
%M = Name of the Column which contains the ID as known in Dic\n\
%N = Notes and Remarks about the Catalogue\n\
%O = Individual Object\n\
%Q = Query possibilities (VizieR data-base)\n\
%R = BibCode of ReadMe (mainly for ADS)\n\
%S = Size of Catalogue (as Unix files)\n\
%T = Title of Catalogue\n\
%t = Short Title of Catalogue\n\
%U = Usual Designation\n\
%V = Records\n\
%W = Working Server (Catalogue Server by Position)\n\
%w+ = Tables to keep\n\
%w- = Tables to ignore\n\
%X = comments (obsolete)\n\
%Y = Availability: B=Bejing I=India L=LaPlata M=Moscow N=NASA T=Tokyo\
 f=ftp v=VizieR\n\
%Z = Status of Catalogue in Simbad\n\
" ;


static char usage[] = "\
Usage: preadme [-v] [-cgi|-c2|-csv|-par[=list]] [-d dir] [-i file] [-b catid]\n\
       [-a anchor_prefix] [-o output] [-t table] [ReadMe_file]         #v1.9#\n\
" ;

static char thelp[] = "\
      -v: verbose option\n\
    -cgi: edit for cgiprint (-c2 for \\boxa)\n\
    -csv: edit the ReadMe in TSV format\n\
    -par: edit the ReadMe in PARFILE format (for ADS);\n\
	  can select a list of keys to edit (list by preadme -keys)\n\
  -d dir: specifies the directory containing the data files\n\
 -i file: specifies the file with description (default 'ReadMe')\n\
 -b code: specifies the numeric code to use for -par edition\n\
 -a pref: set a text which prefixes the internal anchors (default 'sRM')\n\
 -o file: Output file (default: stdout)\n\
 -t desc: Output only for a specified table (exit 1 when not described)\n\
    file: file concerned\n\
       -: ReadMe file in standard input\n\
\n\
   This programs edits, for HTML, the contents of the 'ReadMe' file\n\
 describing a catalogue. The result is to be processed by cgiprint.\n\
 The -a option is useful when several descriptions are to be concatenated.\n\
" ;

/*==================================================================
		Internal Utilities
 *==================================================================*/

static void help()
/*++++++++++++++++
.PURPOSE  Detailed help
.RETURNS  -
-----------------*/
{
    fprintf(stdout, "%s%s", usage, thelp) ;
    /* fprintf(stdout,  "\nThe following 'standards' are known:\n"); */
}

static void pr_noline(text, len, nl)
/*++++++++++++++++
.PURPOSE  Print out, but replace '\n' by the 'nl' string. Remove the blanks.
.RETURNS  -
-----------------*/
  char *text ;	/* IN: Text to issue    */
  int len;	/* IN: Length of Text   */
  char *nl ;	/* IN: Substitute of \n */
{
  char *p, *e;
  int lnl, imax ;
  char c = ' ' ;

    if (len < 0) len = strlen(text) ;
    imax = max_len ;
    lnl = strlen(nl)-1 ;

    for (p=text, e=p+len; (p<e) && *p; p++) {
	if (isspace(*p)) {
	    if (*p == '\n') imax -= lnl, fprintf(fout, "%s", nl) ;
	    else if (c != ' ') fputc(' ', fout); 
	    c = ' ' ;
	}
	else fputc(*p, fout), c = *p ;
	if (--imax <= 0) break ;
    }
}

static SECTION *sec_class(sec, class)
/*++++++++++++++++
.PURPOSE  Find the next section corresponding to specified class
.RETURNS  The section / NULL
-----------------*/
  SECTION *sec;	/* IN: First Section to look at */
  int class ;	/* IN: Class type to search...	*/
{
  SECTION *ps;
    for (ps = sec; ps ; ps = ps->next) {
        if (ps->class == class) return(ps) ;
    }
    return((SECTION *)0) ;
}

/*==================================================================
		Edit the Table Contents in TSV Format
 *==================================================================*/
static int print_csv(T)
/*++++++++++++++++
.PURPOSE  Print out the "ReadMe" file
.RETURNS  Number of lines
-----------------*/
  CATINFO *T;           /* IN: Various details about position of description */
{
  SECTION *ps; 
  FILINFO *pf, *pe ; 
  COLINFO *pc, *pd ;
  TABINFO *pt ;
  char *p;
  int len, idno; 

    /* Print out the main Title */
    ps=T->section[0]; 
    p = (char *)0 ;
    len = -1 ;
    if (ps->class == SEC_Title) {
	p = ps->header ;
	if (!p) {
	    p = ps->abody ;
	    len = strloc(p, '\n') ;
	    if ((p+len) > ps->ebody) len = ps->ebody - p ;
	}
    }
    if (p) {	/* Title preceded by '#' when only single table asked */
	if (atable) fputc('#', fout) ;
	pr_noline(p, len, "") ;
    }
    else {
	if (ddat) {
	    fprintf(fout, "%s", ddat) ;
	    if (ddat[strlen(ddat)-1] != '/') fputc('/', fout) ;
	}
	else fprintf(fout, "./") ;
	if (T->ffile) fprintf(fout," ('%s' file)", T->ffile) ;
    }
    fprintf(fout, "\n#Contains filename, datatype, label, units, explain\n") ;

    /* Print the various files */
    if (atable)   len = 1, pf = cad_findfile(T, atable);
    else  len = T->nfinfo, pf = T->afinfo ;
    for (pe=pf+len; pf<pe; pf++) {
	if (!(pt = pf->tabform)) continue ;
	fprintf(fout, "\n#Table %s:\t%s\n", pf->dfile, pf->text) ;
	idno = 0 ;	/* No Identifier Yet */
	for (pc=pt->afmt, pd=pc+pt->nfmt; pc<pd; pc++) {
	    if ((pc->flags & F_ACCU) && (!(pc->flags&F_ACCU2))) /* Cplx...*/
		continue ;
	    p = pc->unit ;
	    if (pc->flags&F_ACCU2) p = "{CPLX}" ;
	    fprintf(fout,"%s%c%c%c%s%c%s%c", pf->dfile, csv,
		pc->ff, csv, pc->label, csv, p, csv) ;
	    /* Skip the Range definitions */
	    p = pc->text; 
	    if (*p == '*') p++; 
	    if (*p == '[') {
	        if (*++p == ']') p++;
	        while (*p && *p != ']') p++ ;
		if (*p == ']') p++ ;
	    }
	    if ((*p == '?') && (p[1] == '=')){	/* Ignore the =... */
	        fputc('?', fout) ;
	        for (++p; isgraph(*p); p++) ;
	    }
	    pr_noline(p, -1, " ") ;
	    /* Special F_ACCU2 */
	    if (pc->flags&F_MPOS) {
		p = " {Coosys: " ;
		if (pf->equinox) fprintf(fout, "%s Equinox=%s", 
		    p, pf->equinox), p = "" ;
		if (pf->epoch) fprintf(fout, "%s Epoch=%s", 
		    p, pf->epoch), p = "" ;
		if (!*p) fputc('}', fout) ;
	    }
	    /* Print the note, if any */
	    if ((ps = pc->note)) {
	        fputc(csv, fout) ;
	        if ((p = ps->header)) pr_noline(p, -1, " ") ;
	        if ((p = ps->abody)) pr_noline(p, ps->ebody - ps->abody, " ") ;
	    }
	    fputc('\n', fout) ;
	}
    }
    return (0) ;
}


/*==================================================================
		Edit the Table Contents in TXT Format
 *==================================================================*/
static int print_txt(T)
/*++++++++++++++++
.PURPOSE  Print out the "ReadMe" file
.RETURNS  Number of lines
-----------------*/
  CATINFO *T;           /* IN: Various details about position of description */
{
  SECTION *ps, **a, **anotes; 
  FILINFO *pf, *pe, filinfo ; 
  COLINFO *pc, *pd ;
  TABINFO *pt ;
  char *p, *text, septable ;
  int len, idno, i, notes, lenlab, lenunit; 
  char fmt[40] ;

    /* Print out the main Title */
    ps=T->section[0]; 
    p = (char *)0 ;
    len = -1 ;
    if (ps->class == SEC_Title) {
	p = ps->header ;
	if (!p) {
	    p = ps->abody ;
	    len = strloc(p, '\n') ;
	    if ((p+len) > ps->ebody) len = ps->ebody - p ;
	}
    }
    if (p) {	/* Title preceded by '#' when only single table asked */
	if (atable) fputc('#', fout) ;
	pr_noline(p, len, "") ;
    }
    else {
	if (ddat) {
	    fprintf(fout, "%s", ddat) ;
	    if (ddat[strlen(ddat)-1] != '/') fputc('/', fout) ;
	}
	else fprintf(fout, "./") ;
	if (T->ffile) fprintf(fout," ('%s' file)", T->ffile) ;
    }
    fprintf(fout, "\n") ;

    /* Print the various files */
    if (atable)   len = 1, pf = cad_findfile(T, atable);
    else  len = T->nfinfo, pf = T->afinfo ;
    if (!pf) {		/* This may happen for files in subdirectories */
	pf = &filinfo ;
	filinfo.dfile = atable ;
	filinfo.tabform = pt = cad_findformat(T, atable) ;
	filinfo.equinox = pt->equinox ;
	filinfo.epoch   = pt->epoch   ;
	filinfo.text    = "(no explanation for this specific file)" ;
    }
    septable = (len > 1) ;		/* One blank line between tables */
    for (pe=pf+len; pf<pe; pf++) {
	if (!(pt = pf->tabform)) continue ;
	if (septable) fputc('\n', fout) ;
	fprintf(fout, "#Table %s: %s\n", pf->dfile, pf->text) ;
	idno = 0 ;	/* No Identifier Yet */
	/* Compute the longest length of unit and label */
	lenlab = lenunit = notes = 0 ;
	for (pc=pt->afmt, pd=pc+pt->nfmt; pc<pd; pc++) {
	    if ((pc->flags & F_ACCU) && (!(pc->flags&F_ACCU2))) /* Cplx...*/
		continue ;
	    if (pc->flags&F_ACCU2) i = 6 ;
	    else i = strlen( pc->unit) ;
	    if (i > lenunit) lenunit = i ;
	    i = strlen(pc->label) ;
	    if (i > lenlab) lenlab = i ;
	    if (pc->note) notes++ ;
	}
	if (lenlab < 8) lenlab=8 ;
	if (lenunit< 3) lenunit=3 ;
	anotes = a = malloc((1+notes)*sizeof(char *)) ;
	*a = (SECTION *)0 ;
	sprintf(fmt, "%%-%ds %%%ds : ", lenunit, lenlab) ;
	for (pc=pt->afmt, pd=pc+pt->nfmt; pc<pd; pc++) {
	    if ((pc->flags & F_ACCU) && (!(pc->flags&F_ACCU2))) /* Cplx...*/
		continue ;
	    if ((pc->label[0] == '-') && (pc->label[1] == '-'))
		continue ;
	    p = pc->unit ;
	    if (pc->flags&F_ACCU2) p = "{CPLX}" ;
	    else if (*p == '-') p = "" ;
	    fprintf(fout, fmt, p, pc->label) ;
	    /* Skip the Range definitions */
	    p = pc->text; 
	    if (*p == '*') p++; 
	    if (*p == '[') {
	        if (*++p == ']') p++;
	        while (*p && *p != ']') p++ ;
		if (*p == ']') p++ ;
	    }
	    if ((*p == '?') && (p[1] == '=')){	/* Ignore the =... */
	        fputc('?', fout) ;
	        for (++p; isgraph(*p); p++) ;
	    }
	    text = p ;
	    /* Special F_ACCU2 */
	    if (pc->flags&F_MPOS) {
		p = "{Coosys: " ;
		if (pf->equinox) fprintf(fout, "%s Equinox=%s", 
		    p, pf->equinox), p = "" ;
		if (pf->epoch) fprintf(fout, "%s Epoch=%s", 
		    p, pf->epoch), p = "" ;
		if (!*p) {
		    fprintf(fout, "}\n") ;
		    i = strloc(fmt, ':') ; fmt[i] = ' ' ;
		    fprintf(fout, fmt, "", "") ;
		    fmt[i] = ':' ;
		}
	    }
	    fprintf(fout, "%s\n", text) ;
	    /* Save teh note to print */
	    if ((ps = pc->note)) {
		for (a=anotes; *a && (*a != ps); a++) ;
		if (! *a) { 		/* A new note */
		    *(a++) = ps ;
		    *a = (SECTION *)0 ;
		}
	    }
	    /* fputc('\n', fout) ; */
	}
	if (*anotes) for (a=anotes; *a; a++) {
	    fputc('\n', fout);
	    ps = *a ;
	    if ((p = ps->header)) fprintf(fout, "%s\n", p) ;
	    if ((len = ps->ebody -  ps->abody)) {
		fwrite( ps->abody, len, 1, fout) ;
		fputc('\n', fout) ;
	    }
	    fputc('\n', fout) ;
	}
	free(anotes) ;
    }
    return (0) ;
}

/*==================================================================
		Edit the ReadMe in ParFile Format
 *==================================================================*/

static int set_bibcode(bibcode, catid)
/*++++++++++++++++
.PURPOSE  Transform the catid (e.g. J/A+AS/...) into a BibCode
.RETURNS  0 / -1 (can't)
.REMARKS  author name is set to \0
-----------------*/
  char *bibcode ;	/* OUT: Modified Bibcode     */
  char *catid ;		/* IN: Catalogue designation */
{
  int i, len;
  char *p, *b=(char *)0;
  int4 bid ;

    bid = bcode ? atoi(bcode) : 0 ;
    strcpy(bibcode+4, "yCat.") ;

    bibcode[4] = 'y' ;	/* Always this signature... */
    p = catid ;
    if ((catid[0] == 'I') || (catid[0] == 'V')) {
	for (i = 0 ; *p == 'I'; p++) i++ ;
	if (*p == 'V')  p++,  i =  5-i ;
	else if (*p == 'X') p++, i = 10-i ;
	while (*p == 'I') i++, p++ ;
	if (*p == '/') p++ ;
	i = i*1000 + atoi(p) ;
	sprintf(bibcode+9, "%d....0", i) ;
	return(0) ;
    }

    if (catid[0] == 'J') {
	p += 2 ; i = 1 ;
	if (*p == 'o') {	/* J/other, must have the -b */
	    if (bid < 1000000000) {
		fprintf(stderr, "****Missing -b catid for '%s'\n", catid) ;
		return(-1) ;
	    }
	    sprintf(bibcode+8, "%d", bid) ;
	    bibcode[8] = 'p' ;
	    return(0) ;
	}
	if (bid == 0) {		/* Assume from ApJ ApJS A+A A+AS AJ PASP MN */
	    len = strloc(p, '/') ;
#if 0	/* Modified in V1.95 */
	    for (b = journals, i=1; *b; b += strloc(b, '/'), i++) {
		if (*b == '/') b++ ;
		if (strncmp(b, p, len) == 0) break ;
	    }
	    if (!*b) {
		fprintf(stderr, "****Missing -b catid for '%s' (%s)\n",
		  catid, "unknown journal") ;
		return(-1) ;
	    }
#else
	    for (i=0; journal[i]; i += 2) {
		b = journal[i];
		if (b[0] == 'o') break;  	/* J/other notcomputable */
		if (strncmp(b, p, len) == 0) break;
	    }
	    if ((b[0] == 'o') || (!journal[i])) {
		fprintf(stderr, "****Missing -b catid for '%s' (%s)\n",
			 catid, "unknown journal") ;
		return(-1) ;
	    }
	    i = 1 + (i/2);
#endif
	    bid = i * 10000000 ;
	    p += len; if (*p == '/') p++;
	    bid += 10000*atoi(p) ;
	    len = strloc(p, '/') ;
	    p += len; if (*p == '/') p++;
	    if (isalpha(*p)) p++, bid += 9000 ;	/* Letter page */
	    bid += atoi(p) ;
	}
    }
    else {
        /* General list: must be explicitely associated a number (-b option) */
        if (bid < 100000) {
	    fprintf(stderr, "****Missing -b catid for '%s'\n", catid) ;
	    return(-1) ;
        }
    }
    sprintf(bibcode+9, "%9d", bid) ;
    for (i = strloc(bibcode, ' '); bibcode[i] == ' '; i++) 
	bibcode[i] = '.';
    return(0) ;
}

static char *end_date(T)
/*++++++++++++++++
.PURPOSE  Standardize the date of ReadMe file
.RETURNS  Internal buffer with date YYYY-mmm-dd
-----------------*/
  CATINFO *T ;		/* IN: The ReadMe file	 */
{
  static char ymd[20] ;
  SECTION *ps ;
  char *s, *p;
  int len ;

    if ((ps = sec_class(T->section[0], SEC_End))) p = ps->header ;
    else {
	fprintf(stderr, "**** (End) missing\n") ;
	return ((char *)0) ;
    }
    if (!p) return ((char *)0) ;
    p += strloc(p, '\n') ;
    --p;	/* Last byte */
    while (isspace(*p) || ispunct(*p)) --p;
    for (s = p; isalnum(*s) || (*s == '-') || (*s == '.'); s--) ; ++s ;
    len = 1+(p-s) ;		/* Length of date */
    if (atoi(s) > 1900) 	/* yyyy-mmm-dd ........ */
	strncpy(ymd, s, len) ;
    else {		/* dd-mmm-yyyy	*/
	strncpy(ymd, p-3, len=4) ;
	ymd[len++] = '-' ;
	for (p=s; isalnum(*p); p++) ;
	if (ispunct(*p) || isspace(*p)) p++ ;
	while (isalnum(*p)) ymd[len++] = *(p++) ;	/* Copy month */
	ymd[len++] = '-' ;
	for (p=s; isalnum(*p); p++) ymd[len++] = *p ;	/* Copy day   */
    }
    ymd[len] = 0 ;
    return(ymd) ;
}

static int print1(text, len, par)
/*++++++++++++++++
.PURPOSE  Write out (stdout) text, replacing empty lines par "par"
.RETURNS  Number of lines
-----------------*/
  char *text ;		/* IN: Text to write out */
  int len ;		/* IN: Length of text    */
  char *par ;		/* IN: Replacement for enpty lines */
{
  char *e, *p, *n;
  int lines = 0;
  int nl ;
    for (p=text, e=p+len; p<e; p=n) {
	if (*p == '\n') {	/* Check \n\n */
	    for (n=p+1, nl=1; (n<e) && isspace(*n); n++) { 
		if (*n == '\n') nl++ ;
	    }
	    if (nl>1) printf("%s", par) ;
	    else fwrite(p, 1, n-p, stdout) ;
	    lines++ ;
	    continue ;
	}
	for (n=p; (n<e) && (*n != '\n'); n++) ;
	fwrite(p, 1, n-p, stdout) ;
    }
    return(lines) ;
}

static int print_par(T, items)
/*++++++++++++++++
.PURPOSE  Print out the "ReadMe" file for ParFile (specified items)
.RETURNS  Number of linesA/ -1 when error
-----------------*/
  CATINFO *T;           /* IN: Various details about position of description */
  char *items ;		/* IN: Items to list out */
{
  SECTION *ps; 
  FILINFO *pf, *pe ; 
  char bibcode[20] ;
  char *p, *e, *b, *f, *s, *pitem, *line1, *tit1, *aut1;
  int c, i, c1=0, len, nitems, status=0; 

    ps = T->section[0] ;
    line1 = ps->header; if (!line1) line1 = ps->abody ;

    /* Find Title, Authors */
    tit1 = ps->abody ;
    if (tit1 == line1) { tit1 += strloc(tit1, '\n'); if (*tit1) tit1++; }
    if (*tit1 == '=') { tit1 += strloc(tit1, '\n'); if (*tit1) tit1++; }
    aut1 = tit1 + strloc(tit1, '\n') ;
    while ((aut1 < ps->ebody) && *aut1) {
        if (*aut1 == '\n') { aut1++ ; continue ; }
	if (*aut1 == ' ') break ;
	if (*aut1 == '=') { aut1 = "  Anonymous" ; break ; }
	aut1 += strloc(aut1, '\n') ;
    }
    while (*aut1 == ' ') aut1++ ;

    putchar('\n') ;

    for (pitem = items; *pitem; pitem++) switch(c = *pitem) {
      case ',': case ' ': continue ;

      case 'I':	/* Main Identification of the Catalogue */
        printf("%%%c ", c) ;
	p = line1 ;
	while(isgraph(*p)) putchar(*p), p++ ;
	putchar('\n') ; continue ;

      case 't':	/* Short Title */
	ps = T->section[0] ;
        printf("%%%c ", c) ;
	for (p=line1; isgraph(*p); p++) ;
	while (*p == ' ') p++ ;
	e = p + strloc(p, '\n') ;
	while (p < e) {
	    if (*p == ' ') {
	        putchar(' ') ;
		while (*p == ' ') p++ ;
		continue ;
	    }
	    if (isgraph(*p)) while(isgraph(*p)) putchar(*p), p++ ;
	    else { /* Should never happen -- except in \r\n sequences! */
	      static int count=0;
	        if(++count<10) fprintf(stderr, 
		    "#+++Control char '\\x%02X' in ReadMe\n", (*p)&0xff);
		p++;
	    }
	}
	p = ps->header; if (!p) p = ps->abody ;
	while (isgraph(*p)) p++ ;	/* Skip %I */
	putchar('\n') ; continue ;

      case 'T': /* Full  Title */
	ps = T->section[0] ;
	p = tit1; e = ps->ebody ;
        printf("%%%c", c) ;
	while((p<e) && isgraph(*p)) {
	    len = strloc(p, '\n') ;
	    putchar(' ') ;
	    fwrite(p, 1, len, stdout) ;
	    p += len+1 ;
	}
	putchar('\n') ; continue ;

      case 'J': 	/* BibCodes */
	c1 = '=' ;	/* The '=' indicates the bibcode */
	ps = T->section[0] ;
	e = ps->ebody ;
	nitems = 0 ;
      case_J:
	for (p=aut1; p<e ; p += 1+strloc(p, '\n')) {
	    if (*p != ' ') continue;
	    while ((p<e) && (*p==' ')) p++ ;
	    if (*p != c1) continue;
	    printf("%%%c ", c) ;
            if (*p == '=') { p++ ; len = 19 ; }
	    else { len = strloc(p, '>'); if (p[len]) len++; }
	    fwrite(p, 1, len, stdout) ;
	    p += len ;
	    nitems++ ;
	    putchar('\n') ;
	}
	if (nitems == 0) {
	    /* No BibCode, use the Reference */
	    if (c1 != '<') { c1 = '<'; goto case_J; }
	}
	continue ;

      case 'A': /* Authors */
	ps = T->section[0] ;
	p = aut1; e = ps->ebody ;
        printf("%%%c ", c) ;
	len = strloc(aut1, '\n') ;
	fwrite(aut1, 1, len, stdout) ;
	p = aut1 + len ;
	/* Author lines are indented, start with alphabetic */
	while((p<e) && (p[0] == '\n') && (p[1] == ' ')) {
	    while (isspace(*p)) p++ ;
	    len = strloc(p, '\n') ;
	    if (!isalpha(*p)) break ;
	    putchar(' ') ;
	    fwrite(p, 1, len, stdout) ;
	    p += len ;
	}
	putchar('\n') ; continue ;

      case 'D':	/* Date is the date at the bottom of the ReadMe file */
	printf("%%%c ", c) ;
	p = end_date(T) ;
	printf("%s\n", p ? p : "****-***-**" ) ;
	continue ;

      case 'R':	/* ReadMe Ref. */
        printf("%%%c ", c) ;
	p = end_date(T) ;
	strncpy(bibcode, p ? p : "****", 4) ;
	if (set_bibcode(bibcode, line1) < 0) {
	    fprintf(stderr, "%s", usage) ;
	    exit(1) ;
	}
	p = aut1; 
	/* In some cases, no author-- take first letter, or A for Anonymous */
	if (isalpha(*p)) ; else if (*p) p++ ;
	bibcode[18] = isalpha(*p) ? toupper(*p) : 'A' ;
	bibcode[19] = 0 ;
	printf("%s\n", bibcode) ;
	continue ;

      case 'X': /* Comments (obsolete) */
	if (T->flags&CAT_obs) {
            printf("%%%c Obsoleted", c) ;
	    for (p=T->more+strlen(T->more)-1; (p>T->more) && isspace(*p); p--) ;
	    while ((p>T->more) &&isgraph(*p)) --p;
	    ++p ;
	    if (isgraph(*p)) {
		printf(" by ") ;
		while (isgraph(*p)) putchar(*p), p++ ;
	    }
	    putchar('\n') ;
	}
	continue ;

      case 'd': /* Date of ReadMe File */
	dateset(T->credate) ;
        printf("%%%c ", c) ;
	p = date('.') ;
	while (*p && (*p != ',')) putchar(*p), p++ ;
	putchar('\n') ; continue ;

      case 'B': /* Abstract... */
	ps = sec_class(T->section[0], SEC_Description) ;
	if (!ps) ps = sec_find(T->section[0], "Introduction") ;
	if (!ps) ps = sec_find(T->section[0], "Abstract") ;
        printf("%%%c ", c) ;
	if (ps) {
	    for (p = ps->abody; *p == ' '; p++) ;
	    print1(p, ps->ebody - p, "\n \\par\n") ;
	}
	else printf ("(no description available)") ;
	putchar('\n') ; continue ;

      case 'F': /* Files ....  */
        pf=T->afinfo, pe=pf+T->nfinfo; 
        for (++pf; pf<pe; pf++) {	/* Skip first(description) file */
            printf("%%%c ", c) ;
	    if (pf->reclen) printf("%-12s %8dx%-4d %s\n", 
		pf->dfile, pf->records, pf->reclen, pf->text) ;
	    else printf("%-12s %8d.%-4s %s\n",
		pf->dfile, pf->records, "", pf->text) ;
	}
	continue ;

      case 'O':	/* List of Objects */
	for (ps = T->section[0]; ps && (ps->class != SEC_Objects);
	    ps = ps->next) ;
	if (!ps) continue ;		/* No object... */
	e = ps->ebody ;
	b = ps->abody ;			/* First line -----------------	*/
	len = 80 ;			/* Maximal length of obj. line  */
	while (b < e) {
	    if (*b == '\n') { b++ ; continue ; }
	    p = b; while (*p == ' ') p++ ;
	    if (isdigit(*p)) break ;	/* Data line (RA part!)		*/
	    s = b ; b += strloc(b, '\n');
	    if (*p == '-') continue ;	/* Rule ------------- (skip it)	*/
	    /* Find the location of the object name in the table	*/
	    while (isgraph(*p)) p++ ;	/* Title "RA" --skip it!	*/
	    while (*p == ' ') p++ ;
	    if (*p && (!isalpha(*p)))	/* The Equinox in the title...	*/
		while (isgraph(*p)) p++ ;
	    while (*p == ' ') p++ ;
	    while (isgraph(*p)) p++ ;	/* Title "Dec" -- skip it!	*/
	    while (*p == ' ') p++ ;
	    c1 = p - s ;		/* Where the designation starts	*/
	    while (isgraph(*p)) p++ ;	/* Title "Designations"		*/
	    while (*p == ' ') p++ ;
	    if (isgraph(*p)) len = p - s ;
	}
	while (b < e) {
	    if (*b == '\n') { b++ ; continue ; }
	    s = b ; 		/* Start = Left margin   */
	    f = s + len;	/* End of object on line */
	    p = b + c1 ;	/* Start of obj. on line */
	    b += strloc(b, '\n');
	    if (f > b) f = b;	/* Actual end of Object  */
	    while (*s == ' ') s++ ;
	    if (*s == '-') continue ;	/* Rule ------------- (skip it)	*/
	    while (*--f == ' ') ; ++f ;
	    if (f > p) {
	        printf("%%%c ", c) ;
	        fwrite(p, 1, f-p, stdout) ;
	        putchar('\n') ;
	    }
	}
	continue ;

      case 'K':	/* ADC Keywds  */
	for (i=0; *keys[i]; i++) {
	    p = cad_keywords(T, keys[i]) ;
	    if (!p) continue ;
	    printf("%%%c ", c) ;
	    while (*p) {
	      if (*p == '\t') printf("; ") ;
	      else if (*p == '\n') { if (p[1] == '\n') break; }
	      else putchar(*p) ;
	      p++ ;
	    }
	    putchar('\n'); 
	}
	continue ;

      case 'U':	/* Usual Name  */
	p = viz_getkey(T, "\\cUsualName");
	if (!p) continue;
	c1 = '\t';	/* Indicates to write a new line */
	i = 0;		/* Counter of Names */
	while (*p) {
	    if (c1 == '\t') { 
		if (i) putchar('\n'); i++;
		printf("%%%c ", c);
	    }
	    c1 = *(p++);
	    if (c1 != '\t') putchar(c1);
	}
	putchar('\n');
	continue;

      case 'c':	/* Catalog Type */
	if ((p = viz_getkey(T, "\\cType")))
	   printf("%%%c %s\n", c, p);
	continue;

      case 'V':	/* # Records */
	if ((p = viz_getkey(T, "\\cRecords")))
	   printf("%%%c %s\n", c, p);
	continue;

      case 'Y':	/* Availability */
	if ((p = viz_getkey(T, "\\cCenters")))
	   printf("%%%c %s\n", c, p);
	continue;

      default :
        printf("%%%c **** Can't interpret this key in parfile output ***\n", c);
	status = -1 ;
	continue ;
    }
    return (status) ;
}

#define pVal(text)      \
	if (! *++p) {                           	\
	    if (--argc <= 0) errors++, error(text);   	\
	    else    p = *++argv;            		\
	}

/*==================================================================
		Test Program
 *==================================================================*/
int main (int argc, char **argv)
/*++++++++++++++++
.PURPOSE  Just a TEST program
-----------------*/
{
  CATINFO *tinfo; int fop, i;
  void *exists ;
  char *p, *getenv(), *getcwd();
  char *t, *par=(char*)0 ;

    if (!getcwd(cwd, sizeof(cwd)-1))
    	perror(*argv), terminate("Can't get cwd");
    fout = stdout ;

	/* Check the Arguments */
    fop = FOP_quiet ;		/* V1.98 */
    while (-- argc > 0) {
	p = *++argv;
	if (*p != '-') break;	/* Options exhausted */
	switch (*++p) {
	  case 'a' :
		pVal("Missing prefix text in '-a' option") ;
		sec_prefix(p) ;
		oped |= P_a ;		/* Also absolute anchors */
		continue ;
	  case 'b':
		pVal("Missing code for Bibliography in '-b' option") ;
		bcode = p;
		continue ;
	  case 'o' : 			/* Tell the output file	*/
	  	pVal("Missing -o(utput) file");
	  	fout = fopen(p, "w");
	  	if (!fout) { perror(p); terminate((char *)0); }
	  	continue;
	  case 'd' :			/* Tell  the Directory 	*/
		pVal("Missing -d Directory");
		ddat = p; 
		continue;
	  case 'i' :			/* Tell which "ReadMe" 	*/
		pVal("Missing -i FileName");
		intro = p; i = strbloc(p, dirsep);
		if ((!ddat) && (i>=0)) {  /* Keep the path 	*/
		    ddat = strdup(intro);
		    ddat[i+1] = 0;
		}
		continue;
	  case 0:
		intro = p ;
		continue ;
	  case 'v':                     /* Verbose -- v1=CatID	*/
		if (strncmp(p, "ver", 3) == 0) {
		    printf("preadme (CDS) Version %s\n", VERSION);
		    exit(0);
		}
		fop &= ~FOP_quiet;	/* V1.98 */
		if (isdigit(p[1])) 
		     verbop = p[1] - '0' ;
		else verbop = 10; 
		if (verbop > 1) fop |= FOP_v;
		/* osmtrace(1); */
		continue;
	  case 't':			/* Choice of Tables */
		pVal("Missing -t TableName");
		atable = p ;
		continue ;
	  case 'p':
	  	par = "RDdItTJXABKFO" ;
		if ((t = strchr(p, '='))) par = t + 1 ;
		else if ((t = strchr(p, '%'))) par = t ;
		if (strchr(par, '*')) par = "IRtTAJcUDdXOBKFVY";
		continue;

	  case 'c':
		if (strncmp(p, "csv", 3) == 0) {
		    csv = p[3] ;
		    if (csv == 0) csv = '\t' ;
		    continue ;
		}
		fop |= FOP_cgi ;
		oped |= P_cgi;
		while (isalpha(*p)) p++;
		if (isdigit(*p)) oped = atoi(p);
		continue;
	  case 'h': 
		if (strcmp(p, "help") == 0) { help(); exit(0); }
	  	goto bad_option;
	  case 'k': 
		if (strcmp(p, "keys") != 0) goto bad_option;
		printf("====List of keys in 'parfile' edition:\n%s\n", parkeys);
		exit(0);
	  default:
	  bad_option:
		fprintf(stderr, "\n****Bad option: -%s\n", p);
		errors++;
		fprintf(stderr, "%s", usage);
		terminate((char *)0);
	} 
    } 
    
#if DEBUG
    printf("----%d arguments remain", argc);
    if (argc>0) printf(", first is %s\n", *argv);
#endif

    if ((!ddat) && (!intro) && (argc>0)) {	/* File may be Dir/File */
	p = *argv;
	i = strbloc(p, dirsep);
	if (i >= 0) {
	    ddat = p; 
	    p += i;
	    *(p++) = 0;
	}
	intro = p;
	argc--; argv++;
    }
    if (argc > 0) {
	fprintf(stderr, "\n****Bad argument: %s\n", *argv);
	terminate((char *)0);
    }

    if (ddat) {
    	if (chdir(ddat)) perror(ddat), 
    	    terminate("Can't find the data directory");
    }

	/* Set up the default variables */
    if (! intro) {
    	intro = "ReadMe";
    	if (!comfile(intro)) intro = "Intro";
    }

    if (verbop>1) fprintf(stderr, 
	"....call cad_new(\"%s\", 0x%X)\n", intro, fop);

    tinfo = cad_new(intro, fop);	/* Format File */
    if (!tinfo) terminate("Errors in \"ReadMe\", terminated");
    if (verbop) {
	fprintf(stderr, "----") ;
	if ((p = tinfo->section[0]->abody)) {
	    while (isgraph(*p)) fputc(*p, stderr), p++ ;
	}
	fputc('\n', stderr) ;
    }
    tinfo->ffile = intro;
    if (verbop>1) {
	cad_check(tinfo);
	printf("====Keywords===\n");
	for (i=0; *keys[i]; i++) {
	    p = (char *)0 ;
	    if (*keys[i] != '\\') p = cad_keywords(tinfo, keys[i]) ;
	    if (p) printf("    %7s=%s\n", keys[i], p) ;
	}
	printf("====Other Keys===\n");
	for (++i; i < ITEMS(keys); i++) {
	    p = (char *)0 ;
	    if (*keys[i] == '\\') p = viz_getkey(tinfo, keys[i]) ;
	    if (p) printf("    %s=%s\n", keys[i], p) ;
	}
    }

    if (atable) {
	exists = cad_findfile(tinfo, atable) ;
	if (!exists) exists = cad_findformat(tinfo, atable) ;
	if (!exists) {
	    fprintf(stderr, "++++File not found/described in catalog %s: %s\n",
	       cad_name(tinfo), atable) ;
	    exit(1) ;
	}
    }

    if (par) print_par(tinfo, par) ;
    else if (csv) print_csv(tinfo) ;
    else if (atable) print_txt(tinfo) ;
    else cad_print(tinfo, stdout, oped);
    
    /* cad_kill(tinfo); -- May cause problems at INASAN, has to be checked ! */
    exit(0);
}
