/*++++++++++++++
.IDENTIFICATION tabform.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT    Unix, CDS Catalogues
.KEYWORDS       Table, FITS
.VERSION  1.0   15-May-1991 (anafile)
.VERSION  4.0   28-Feb-1994 New package
.VERSION  4.1   14-Oct-1994 Bug corrected in tbf_load
.VERSION  4.2   09-Nov-1994: Find Epochs
		(Syndrome: charset includes ! or ?)
.VERSION  4.3   13-Dec-1994: Don't stop if no format found
.VERSION  4.4   04-Feb-1995: Better algorithm to  find Equinox + Epoch
.VERSION  4.5   04-Apr-1995: posfmt in TABINFO structure
.VERSION  4.6   21-Sep-1995: Check not dplicated labels
.VERSION  4.7   12-Oct-1995: added tbf_reset
.VERSION  4.8   19-Oct-1995: posfmt is a table
.VERSION  4.9   10-Jun-1997: Flag F_ACCU, F_SIGN
.VERSION  4.91  16-Jul-1997: Be sure that [1] is limit lower + Upper
.VERSION  4.92  22-Jul-1997: Flag Proper Motions with F_MPOS
.VERSION  4.93  15-Jul-1998: Accept pm + pmPA
.VERSION  4.94  31-Dec-1998: Remove F_MPOS when one is missing
.VERSION  4.95  16-Dec-1999: tbf_reset revisited
.VERSION  5.0   16-Feb-2000: Deeper verification of labels
.VERSION  5.1   05-Oct-2000: Accept references as r_   OR   ref(...)
.VERSION  5.2   05-Apr-2001: tbf_mpos updated (case MPOS & NODB)
.VERSION  5.3   13-Jul-2001: tbf_mpos not in error for NOTES files
.VERSION  5.4   26-Feb-2002: Propage to DE- the F_NULL / F_REQD flags
.VERSION  5.5   01-Apr-2002: Accept TABMAP
.VERSION  5.6   04-Aug-2002: Accept "month" as DATE
.VERSION  5.7   03-Nov-2002: tbf_addcol; tbf_mpos has a list argument
.VERSION  5.8   11-May-2003: tbf_addcol: at least 1 byte for addcolumn
.VERSION  5.9   30-Jun-2004: tbf_mpos: accept suffixes RAsuffix
.VERSION  5.91  16-Jul-2005: label comparison for ACCU
.VERSION  5.92  18-Jul-2005: Accept ?=* ?=~ for NULLs in numbers
.VERSION  5.93  19-Oct-2005: Better check "special" units "date" etc
.VERSION  5.94  05-Dec-2005: F_SSEP installed in complex nuimbers
.VERSION  5.95  13-Mar-2007: Accept charset limits [[...]]
.VERSION  5.96  02-Jun-2007: Cosmetics + accept RAsec DEsec
.VERSION  5.97  17-Oct-2007: Accept in tbf_order (pos='*') for radix
.VERSION  5.98  25-Nov-2007: Add oid
.VERSION  5.99  08-Jan-2008: Accept NULL (?=---) in char strings
.VERSION  5.991 21-Jan-2008: oid changed in 
.VERSION  6.0   07-Aug-2008: tbf_order accept wild cards
.VERSION  6.01  19-Oct-2008: Accept lowercase GLon / GLat
.VERSION  6.02  23-Nov-2008: Question of "datime": set both F_DATE F_TIME
.VERSION  6.03  06-Jun-2009: Look for equinox / epoch in notes
.VERSION  6.04  01-Oct-2009: Bug several positions
.VERSION  6.05  09-Oct-2009: use \yAref
.VERSION  6.06  15-Dec-2009: 64-bit
.VERSION  6.07  31-Aug-2010: Accept complex labels with Z (zones with units)
.VERSION  6.08  02-Sep-2010: free ==> reset to 0
.VERSION  6.09  03-Nov-2010: Comments in ReadMe are also lines starting by [..]
.VERSION  6.10  20-Jun-2011: Modifications labels.h => modified lab_find()
.VERSION  6.11  02-Jul-2011: Accept RAdeg and pmRA
.VERSION  6.12  30-Jul-2011: Bug
.VERSION  6.13  30-Dec-2011: bug in tbf_mpos (in case RA + SPD)
.VERSION  6.14  13-Feb-2012: tbf_addcol: attach the note if terminated by (G1)
.VERSION  6.15  17-Feb-2012: tbf_addcol: bug fix for pasted table
.VERSION  6.16  29-Mar-2012: bug in tbf_order
.VERSION  6.17  18-Jun-2012: bug in tbd_mpos
.VERSION  6.18  28-Jun-2012: DEsuffixes (verify)
.VERSION  6.19  28-Aug-2012: (0) is NOT a note
.VERSION  6.20  03-Oct-2012: tbf_addcol set the b0/b1 values
.VERSION  6.21  16-Feb-2013: accept suffix "sec" for seconds (cat. I/85)
.VERSION  6.22  22-Feb-2013: tbf_mpos: accept a single name (use following one)
.VERSION  6.23  25-Feb-2013: bug in tbf_cpy
.VERSION  6.24  11-Mar-2013: 64-bit
.VERSION  6.25  28-May-2013: Check components of sexagesimal if compatible unit
.VERSION  6.26  29-Jun-2013: Remove also equinox/epoch in tbf_mpos
.COMMENTS       Interpret and displays COLINFO
---------------*/

#ifndef DEBUG
#define DEBUG 0
#endif
#ifndef MAX_ERRORS
#define MAX_ERRORS 256
#endif

static char verbop = DEBUG;
#include <catstd.h>
#include <labels.h>		/* Standard labels */
#include <fcntl.h>		/* File Control	   */
#ifndef O_BINARY
#define O_BINARY 0
#endif

/* Globals (for edition)*/
int (*FPUTC)();  	/* Function which writes a character */

#define MAX(x,y)        ((x) >= (y) ? (x) : (y))
#define MIN(x,y)        ((x) >= (y) ? (y) : (x))
#define ITEMS(a)        sizeof(a)/sizeof(a[0])
#define NULL4		0x80000000
#define NULL8		0x8000000000000000LL
/*static char nan8[8] = { -1, -1, -1, -1, -1, -1, -1, -1 };*/
#define isff(c)		((c)=='A')||((c)=='I')||\
			((c)=='D')||((c)=='E')||((c)=='F')	/* Fortran */
#define bzero(a,n)		memset(a, 0, n)

#define isReadMe_comment(c)  ((c)=='#')||((c)=='[')
#define checkapos(apos)	if (apos) if(apos->flags&F_NODB) \
				apos->flags &= ~(F_MPOS|F_DISP), \
				apos = (COLINFO *)0

static char unitless[] = "---";
static char rule_char = '-';	/* Rule Char	 */
static char *unit_A;		/* Unit Anchor	 */
#define RAsuffix pos_suffix[0]
#define DEsuffix pos_suffix[1]
/* Acceptable suffixes for DEclination with a length>1 */
static char *pos_suffix[] = {
    "deg,mas,hr,hour,sec,ds,cs,ms,d",
    "deg,mas,rad,sec,ds,cs,ms,d,d,-," } ;	/* Mod. V6.21 */
#if DEBUG
static char inbyte;
#endif

/*==================================================================
		Dealing with LABEL
 *==================================================================*/
static int check_label(label, lineno)
/*++++++++++++++++
.PURPOSE  Check the validity of a label
.RETURNS  Number of Warnings
-----------------*/
  char *label;	/* IN: Label name to look at */
  int lineno;	/* IN: Line number in ReadMe */
{
  static char forbid[]  = "$@!`\\^~& \t" ;	/* Always forbidden */
  static char forbid1[] = "+-*/=.#_'\";,";	/* Never as first char */
  static char forbid9[] = ",;";			/* Never as last  char */
  static char *nomatch[] = { "()", "[]", "{}" } ;
  char *p ;
  int op[3], nw, i ;

    p = label ; nw = 0 ;
    /* A single-character label must br alphabetic */
    if (!p[1]) {
	if (!isalpha(*p)) nw++, fprintf(stderr,
	    "++++ line %d: 1-character label must be alphabetic: %s\n", lineno,
	    label) ;
	return(nw) ;
    }

    /* First byte */
    if (strcmp(label, "---") == 0) return(0) ;
    if (strchr(forbid1, *p)) nw++, fprintf(stderr,
	"++++ line %d: forbidden first character '%c' of label: %s\n",
	    lineno, *p, label) ;
    while (*p) {
	if (strchr(forbid, *p)) nw++, fprintf(stderr,
	    "++++ line %d: forbidden character '%c' in label: %s\n",
	    lineno, *p, label) ;
	p++ ;
    }

    /* Last Byte */
    --p ;
    if (strchr(forbid9, *p)) nw++, fprintf(stderr,
	"++++ line %d: forbidden  last character '%c' of label: %s\n",
	lineno, *p, label) ;

    /* Verify the parenthesis */
    for (p=label, op[0]=op[1]=op[2]=0; *p; p++) switch(*p) {
	case '(':  op[0]++; continue ;
	case ')':  op[0]--; continue ;
	case '[':  op[1]++; continue ;
	case ']':  op[1]--; continue ;
	case '{':  op[2]++; continue ;
	case '}':  op[2]--; continue ;
    }
    for (i=0; i<3; i++) if (op[i]) nw++, fprintf(stderr,
	"++++ line %d: unbalanced %s symbols in label: %s\n",
	lineno, nomatch[i], label) ;

    return(nw) ;

}

static LABEL *lab_find(label)
/*++++++++++++++++
.PURPOSE  Lookup in the known labels
.RETURNS  The matching LABEL / NULL
.REMARKS Modified V6.10 (examine only with/without prefix)
-----------------*/
  char *label;	/* IN: Label name to look at */
{
  LABEL *pl, *pe;
  int len;

    pl = known_label; pe = pl + i2known_label;
    if (label[0] && (label[1]) == '_') { 	/* Label like e_label */
	pl = pe;
	pe = known_label + ITEMS(known_label);
    }
    while (pl < pe) {
	len = strlen(pl->name);
	if (strncmp(pl->name, label, len) == 0){ /* label match */
	    if ((len == 2) && (label[1] == '_')) return(pl);
	    else if (!isalpha(label[len])) return(pl);
	}
	pl++;
    }
    return((LABEL *)0);
}

static int lab_len(label)
/*++++++++++++++++
.PURPOSE  Find out the length of radix part
.RETURNS  0 (doesn't match) / length
-----------------*/
  char *label ;	/* IN: Label to compare */
{
  char *p ; int len;
    if (*label == '-') return(0) ;
    len =  strlen(label);
    p = label + len - 1 ;
    switch(*p) {	/* Look for possible suffixes */
      case 'D':	/* Can be .YD = day of year */
	if (p[-1] == 'Y') p--;
	/* NO BREAK */
      case '.':
      case '-':	/* Can be a sign ... */
      case 'Y': case 'M':
      case 'y':
      case 'h': case 'd':
	break ;
      case 'm':
	if (p[-1] == 'd') p-- ;
	break ;
      case 's':
	if ((p[-1] == 'd') || (p[-1] == 'c') || (p[-1] == 'm')) p-- ;
	break ;
      case 'e':	/* Can be '.date' '.time' */
	if (len<=5) return(0);
	p -= 3;
	if (strmatch("[Dd]ate", p)) break;
	if (strcmp(p, "time") == 0) break;
	return(0);
      case 'z': case 'Z':	/* V6.07 -- but MUST be preceded by a . */
	if (ispunct(p[-1])) break;
      default:
	return(0) ;
    }
    /* if (p[-1] == '.') p-- ; -- Removed V5.91 */
    /* Accept as a generic label ONLY when preceding is a digit, punctuation
       or uppercase -- in other words can't work for lowercase termination
    */
    if (islower(p[-1])) return(0);
    return(p-label) ;
}

static int lab_match(radix, label, len)
/*++++++++++++++++
.PURPOSE  Compare: can the label be part of radix ?
.RETURNS  0 (doesn't match) / 1 (match)
-----------------*/
  char *radix ;	/* IN: The 'main' radix */
  char *label ;	/* IN: Label to compare */
  int  len ;	/* IN: Length of radix part */
{
    if (len < 2) return(0) ;	/* Don't accept shorter than RAh */
    if (strncmp(radix, label, len) != 0) return(0) ;
    if (!label[len]) return (0) ;
    if (lab_len(label) == len) return(1) ;
    return(0) ;

}

/*==================================================================
		Dealing with Equinox + Epoch
 *==================================================================*/
#if 0
static char *dup_epoch(s)
/*++++++++++++++++
.PURPOSE  Find out the length of an equinox/epoch
.RETURNS  A cloned copy of the equinox/epoch
-----------------*/
  char *s;	/* IN: String to interpret */
{
  char *p, *r; int len;
    if (!s) return(s);
    p = s;
    if (isalpha(*p)) p++;
    while (isdigit(*p)) p++;
    if (*p == '.') for (++p; isdigit(*p); p++) ;
    len = (p-s);
    r = malloc(len+1);
    strncpy(r, s, len); r[len] = 0;
    return(r);
}
#endif

static int set_accu1(pf, pl)
/*++++++++++++++++
.PURPOSE  Check if any of the specified marked columns are a date
.RETURNS  F_ACCU1  /  0
-----------------*/
  COLINFO *pf ;	/* IN: First column to examine */
  COLINFO *pl ;	/* IN: Last  column to examine */
{
  COLINFO *p ; int accu1, k ;
    for (p=pf, accu1=0, k=2 ; (accu1 == 0) && (--k >= 0) ; p = pl) {
	if (p->flags & F_SIGN) continue ;
	switch(p->unit[0]) {
	  case '-': case '"': case '\0':
	    accu1 = F_ACCU1 ;
	    continue ;
	  case 'y':
	    if (p->unit[1] == 'r') accu1 = F_ACCU1 ;	/* Unit yr */
	    continue ;
	  case 'a':
	    if (!p->unit[1])       accu1 = F_ACCU1 ;	/* Unit yr */
	    continue ;
	}
    }
    if (accu1)  while (pf <= pl) {
	if (pf->flags&(F_ACCU|F_ACCU2)) pf->flags |= F_ACCU1 ;
	pf++ ;
    }
    return(accu1) ;
}

static int mark_accu(T, verbop)
/*++++++++++++++++
.PURPOSE  Mark out the complex (sexagesimal) values
.RETURNS  Number of complex numbers found
.REMARKS  The first is marked F_ACCU2 (and eventually F_SIGN), the next F_ACCU
	F_ACCU1 is added when it's a data (can't just sum components...)
-----------------*/
  TABINFO *T; 	/* IN: The complete format */
  int verbop;	/* IN: Option to display   */
{
  COLINFO *pf, *pe, *p1 ;
  char *plab, udim[20];
  int   llab, n, accu1  ;

    /* Remove first the flags related to complex numbers */
    for (pf=T->afmt, pe=pf+T->nfmt; pf<pe ; pf++)
        pf->flags &= ~(F_SIGN|F_ACCU2|F_ACCU1|F_ACCU);

    p1 = (COLINFO *)0;	/* Column of first column of complex value */
    plab = (char *)0;	/* Name (label) of complex value */
    llab = n = accu1 = 0 ;
    udim[0] = 0;	/* V6.25 */
    for (pf=T->afmt, pe=pf+T->nfmt; pf<pe ; pf++) {
	if (pf->label[0] == '-') {
	    if (! isalnum(pf->label[1])) continue ;
	}
	if (pf->label[1] == '_') /* Just a flag  ... */
	    continue ;
	if (plab) {	/* Compare with preceding */
	    /* Can be complex values only when units differ...	*/
	    if (lab_match(plab, pf->label, llab) && strcmp(pf->unit, p1->unit)){
		if (!(p1->flags&F_ACCU2)) {
		    if (verbop && (n==0)) fprintf(stderr,
			"----accu markers in table described at line %d:",
			T->lineno) ;
		    n++ ;
		    if (verbop) fprintf(stderr, "\n    %s", plab) ;
		    udim[0]=0;	/* V6.25: compare dimansions */
		}
		p1->flags |= F_ACCU2 ;
		/* In sign, propagate the REQD / NULL flags */
		if (p1->label[strlen(p1->label)-1] == '-') {
		    p1->flags |= F_SIGN ;
		    p1->elims &= ~(F_NULL|F_REQD) ;
		    p1->elims |= pf->elims&(F_NULL|F_REQD) ;
		}
		pf->flags |= (F_ACCU|accu1) ;
		/* V6.25: verify compatible dimensions! */
		if ((udim[0]==0) && ((p1->flags&F_SIGN)==0)) 
		    strcpy(udim, unit_dim(p1->unit));
		if (udim[0]) {
		    char *u = unit_dim(pf->unit);
		    if (strdiff(udim, u)) {
			if (verbop) fprintf(stderr, "\n");
			fprintf(stderr, 
			    "#***Incompatible units: %s[%s] / %s[%s]",
			    pf->label, pf->unit, p1->label, p1->unit);
			fprintf(stderr, ": dimensions are %s / %s\n",
			    u, udim);
			if (verbop) fprintf(stderr, "#  ");
		    }
		}
		if (pf->label[strlen(pf->label)-1] == '-')  
		    pf->flags |= F_SIGN ; /* V6.07: sign AFTER value */
		if (!accu1) accu1 = set_accu1(p1, pf) ;
		if (verbop) fprintf(stderr, " %s", pf->label) ;
		continue ;
	    }
	    plab = (char *)0 ;
	}
	llab = lab_len(pf->label) ;
	if (llab == 0) continue ;
	plab = pf->label ;
	p1 = pf ;
	accu1 = 0 ;
    }
    if (verbop && n) fprintf(stderr, "\n") ;
    return(n) ;
}

int tbf_mpos(T, pos)
/*++++++++++++++++
.PURPOSE  Flag the columns with Main Position, if any
.RETURNS  0 (none)  1 (Equinox found)  2(Epoch found)  3(Both)
.REMARKS  When several positions exist, we keep the main one.
	Columns with Epoch of Positions are flagged by F_MPOS
-----------------*/
  TABINFO *T; 	/* IN: The complete format */
  char *pos;	/* IN: The list of chosen positions -- NULL for Choice */
{
static char note_pos[] = "\
    (the full equatorial position requires RAh RAm RAs DE- DEd DEm DEs)" ;
  COLINFO *pf, *pe, *apm, *apos;
  char *p, *a, *n, *u, *v, *lof=(char *)0, *coolab[2], *saved_ep[2];
  char sec[2], re_search=0;
  char **parm=0, latname[20];
  double factor, fact1 ;
  int i, j, k, len, icoo, ilat, lcmp[2];
  int needs_eq=1;	/* Equinox required */
  SECTION *section ;

    /* When the list is provided, remove first the existing */
    apos = (COLINFO *)0;
    pf=T->afmt, pe=pf+T->nfmt ;
    if (pos) {
	saved_ep[0] = T->equinox;
	saved_ep[1] = T->epoch  ;
	if (stumatch("*icrs*", pos)||stumatch("*glon*", pos))
	    needs_eq = 0;
	for (i=0; T->posfmt[i] >= 0; i++) {
	    pf = T->afmt + T->posfmt[i];
	    pf->flags &= ~F_MPOS;
	    T->posfmt[i] = -1;
	}
	for (p=pos, k=0; *p && (k<7); ) {
	    if (isspace(*p)) { p++; continue; }
	    for (n=p; isgraph(*n); n++) ;
	    re_search = *n; *n = 0;
	    pf = tbf_findlabel(T, p);
	    if (pf) pf->flags |= F_MPOS, T->posfmt[k++] = pf - T->afmt;
	    else fprintf(stderr, "++++tbf_mpos: can't find column '%s'\n", p);
	    if (!apos) apos = pf;
	    *n = re_search;
	    p = n;
	}
	if (k==1) {	/* V6.22: take next column as second components */
	    i = T->posfmt[0]+1;
	    if (i<T->nfmt) {
		T->posfmt[k++] = i;
		pf++;
		pf->flags |= F_MPOS;
	    }
	    else fprintf(stderr, 
		    "#+++tbf_mpos: single position specified: %s\n", pos);
	}
	T->posfmt[k] = -1;	/* Terminate list */
	if (T->posfmt[0] < 0) { /* No position at all... (Mod. V6.26) */
            if (T->equinox) free(T->equinox); 
            if (T->epoch)   free(T->epoch);
	    T->equinox = T->epoch = (char *)0;
	    return(0);
	}
	else parm = epochs(apos->text, 1);	/* Find equinox / epoch */
	if (!parm[0]) parm[0] = saved_ep[0];
	if (!parm[1]) parm[1] = saved_ep[1];
	goto SCAN_FOR_PM; 
    }

    /* Find the Coordinates from the Byte-by-Byte Description */
    if (T->equinox) free(T->equinox), T->equinox = (char *)0;
    if (T->epoch)   free(T->epoch),   T->epoch = (char *)0;
    coolab[0] = coolab[1] = (char *)0;
    /* List of files from ReadMe */
    lof = "*" ;
    if ((section = T->section)) {
	if ((p = section->header)) lof = p + strloc(p, ':') ;
	if (*lof) lof++ ;
	while (*lof == ' ') lof++ ;
    }
    T->posfmt[0] = -1;
    if (verbop) {
	fprintf(stderr, "....Examine ") ;
	p = T->section->header ;
	if (!p) p = "" ;
	while (*p && (*p != ':')) fputc(*p, stderr), p++ ;
	if (*p) fputc(*p, stderr), p++ ;
	while (isspace(*p)) fputc(*p, stderr), p++ ;
	while (isgraph(*p)) fputc(*p, stderr), p++ ;
	if (*p == ' ') fprintf(stderr, " ...") ;
    }

    /* Look first for "RA"; among RA's (must be expressed in time units,
       or be unitless); keep the shortest label */
    apos = (COLINFO *)0; lcmp[0] = lcmp[1] = 0 ; sec[0] = 0 ;
    /* Look for Basic Definitions ... and remove the MPOS for NODB columns */
    re_search = 1;	/* Accept Declination comes BEFORE Right Ascension */
    apos = tbf_findlabel(T, "RAdeg") ; checkapos(apos) ;
    if (!apos) apos = tbf_findlabel(T, "RArad") ; checkapos(apos) ;
    if (!apos) apos = tbf_findlabel(T, "RAmas") ; checkapos(apos) ;
    if (!apos) re_search = 0 ;
    if (apos) lcmp[0] = strlen(apos->label) ;
    else for (; pf<pe ; pf++) {
	if (pf->flags&F_NODB) { pf->flags &= ~(F_MPOS|F_DISP); continue; }
    	if ((pf->label)[0] != 'R') continue;
    	if ((pf->label)[1] != 'A') continue;
	if (*(pf->unit) == '-')  continue;	/* No Unit */
	u = unit_si(pf->unit, &factor);
	if ((u[0]==0) || ((u[0]=='s')&&(u[1]==0))) ; else continue;
	/* Verify possible label -- the 'normal' one is 'h' */
	len = strlen(pf->label); j=1;
	if (pf->label[len-1] != 'h') {
	    for (p=RAsuffix; *p; p += j) {
	        j = strloc(p, ',');
	        if (j == 0) { j=1; continue; }
	        if (strncmp(pf->label+len-j, p, j)) continue;
		break;		/* The label is OK */
	    }
	    if (!*p) 		/* No valid suffix found here */
		continue;
	    if ((*p == 'h') || (*p == 's'))	/* RAhour RAsec */
		re_search = 1 ;	/* Accept Declination BEFORE RA */
	}

	if (apos) {		/* There is already another RA	*/
	    /* keep this one if shorter ... */
	    if ((len-j) >= lcmp[0]) continue ;
	}
	apos = pf; lcmp[0] = len-j;
	sec[0] = u[0]; sec[1] = u[1]; /*strcpy(sec, u) ;*/
    }
    if (apos) { 	/* Found RA; the Dec name has "RA" changed into "DE" */
	strcpy(latname, apos->label); latname[0] = 'D'; latname[1] = 'E';
	coolab[0] = apos->label; coolab[1] = latname;     
	lcmp[1] = lcmp[0];
    }

    /* If no "RA", look for "[GE]LON" -- which must be dimensionless */
    else for (pf=T->afmt; (pf<pe) && (!apos) ; pf++) {
	if (pf->flags&F_NODB) continue ;
	if ((pf->label[0] != 'G')&&(pf->label[0] != 'E')) continue;
	if (toupper((pf->label)[1]) != 'L') continue;
	if (toupper((pf->label)[2]) != 'O') continue;
	if (toupper((pf->label)[3]) != 'N') continue;
	u = unit_si(pf->unit, &factor);
	if (*u) continue;		/* Must be unitless	*/
	apos = pf; v = pf->label; lcmp[0] = strlen(v)-1;
	strncpy(latname, v, sizeof(latname)-1);
	/* V6.01: Lon and Lat must be written with same case */
	latname[2] = islower(latname[2]) ? 'a' : 'A';
	latname[3] = islower(latname[3]) ? 't' : 'T';
	coolab[0] = apos->label; coolab[1] = latname;     
	lcmp[1] = lcmp[0];
	if (pf->label[0] == 'G') needs_eq = 0;
	re_search = 1 ;			/* Accept GLAT before GLON */
    }

    if (!apos) {		/* No position at all... */
    	if ((apos = tbf_findlabel(T, "_RA"))) {
	    /* COMPUTED RA + Dec 	*/
	    coolab[0] = "_RA"; coolab[1] = "_DE";
	    lcmp[1] = lcmp[0] = 3;
	}
    }
    if (!apos) {		/* No position at all... */
	if (verbop) fprintf(stderr, " (no position)\n") ;
	return(0) ;
    }

    /* Find out the various components, as follows:
       => check matching "DE"; if yes, continue only with DE;
       => verify identical physical types
       => verify that, if RAdeg is provided, we find DEdeg
       => verify that names match, with the exception of lowercase
	  definitions of components, e.g. RAh vs RAdm
    */
    /* Not necessary ...
    if (strmatch("*deg",  apos->label)) m = "*deg" ;
    else if (strmatch("*rad", apos->label)) m = "*rad" ;
    else if (strmatch("*mas", apos->label)) m = "*mas" ;
    else if (strmatch("*sec", apos->label)) m = "*sec" ;
    else m = "*" ;   */
    T->posfmt[0] = (apos - T->afmt);
    pf = apos+1; 
    i = 0;   /* Index in T->posfmt */
    icoo = 0;	/* 0=RA, 1=Dec */
    
  SCAN_FOR_MPOS:
    /* I know, that's bad to have a goto to this label, it's for
       testing other names than the standard DE
    */
    fact1 = 999999 ;
    ilat = -1;	/* Index of latitude part in T->posfmt array */
    /* Here we start just after the RA found. */
    for ( ; (pf<pe) && (i<6); pf++) {
	if (pf->flags&F_NODB) continue ;
	if ((icoo==0) && (strncmp(coolab[1], pf->label, lcmp[1]) == 0)) {
	    icoo = 1; sec[0] = 0; ilat = i+1;    /* ilat: where Dec/lat starts*/
	    fact1 = 999999;
	}
	/*if (!strmatch(m, pf->label)) continue ; */
	if (strncmp(coolab[icoo], pf->label, lcmp[icoo])) continue;
	/* V6.18: Verify the suffix is OK */
	p = pf->label + lcmp[icoo];	/* Suffix */
	if (strlen(p)>1) {
	    if (!strstr(pos_suffix[icoo], p)) continue;
	}
	/* End of addition for V6.18 */
	if (strcmp(sec, unit_si(pf->unit, &factor))) {
	    fprintf(stderr,
	       "#+++Column '%s': Unit '%s' can't match a position\n",
	       pf->label, pf->unit) ;
	    continue;
	}
	/* We assume the values are always in decreasing order - the sign
	 * may hwoever be at the end... */
	if (pf->flags&F_SIGN) ;
	else if (factor >= fact1) continue ;
	fact1 = factor ;
#if 0
	a = apos->label + lcmp[0]; p = pf->label + lcmp[1];
	while (*a && (*a == *p)) a++, p++;
	if (islower(*a) && (islower(*p) || (*p == '-'))) {
	    while (islower(*a)) a++;
	    if (*p == '-') p++;
	    else while (islower(*p)) p++;
	}
	while (*a && (*a == *p)) a++, p++;
	if ((!*a) && (!*p)) { 	/* Match !!	*/
	    T->posfmt[++i] = (pf - T->afmt);
	    if (m[1]) break;	/* *deg *mas *rad: a single number (V6.04)*/
	    /* if (verbop) fprintf(stderr, "....Matching %s (i=%d)\n",
		pf->label, i) ;
	    */
	}
#else
	T->posfmt[++i] = (pf - T->afmt);
	/* Push the sign BEFORE the declination part */
	if ((pf->flags&F_SIGN) && (i>ilat) && icoo) {
	    for (j=i-1; j>=ilat; j--) T->posfmt[j+1] = T->posfmt[j];
	    j = T->posfmt[ilat];
	    fprintf(stderr, 
		    "#+++Sign of latitude (%s) follows its value (%s)\n",
		    pf->label, T->afmt[j].label);
	    T->posfmt[ilat] = (pf - T->afmt);
	}
#endif

    }
    if ((icoo == 0) && (re_search)) {
	re_search = 0 ;
	pf = T->afmt ; pe = apos;
	goto SCAN_FOR_MPOS ;
    }

    /* Verify Declination found --- otherwise try NPD / SPD (Polar Distance) */
    pe=T->afmt+T->nfmt;
    if ((icoo == 0) && (coolab[0][0] == 'R') && (coolab[0][1] == 'A')) 
    for (pf=T->afmt+T->posfmt[i]+1; (pf<pe) && (i<6); pf++) {
	if (pf->flags&F_NODB) continue ;
	if (!strmatch("[NS]PD*", pf->label)) continue;
	if (*unit_si(pf->unit, &factor))    continue;
	if (factor >= 1.0)                  continue;
	memmove(coolab[1]+1, coolab[1], strlen(coolab[1])+1);
	coolab[1][0] = pf->label[0];
	coolab[1][1] = pf->label[1];
	coolab[1][2] = pf->label[2];
	icoo = 1; lcmp[1] += 1;
	goto SCAN_FOR_MPOS;
    }
    T->posfmt[++i] = -1;	/* Terminate the list	*/

    /* Set the flag F_MPOS in the specified fields		*/
    if (verbop) fprintf(stderr, "\n----Position made of (i=%d):", i) ;
    /* In k, save position names ; there must finally be 2 bytes used	*/
    k = 0 ;
    for (i=0; T->posfmt[i] >= 0; i++) {
	pf = T->afmt + T->posfmt[i];
	pf->flags |= (F_MPOS|F_DISP);
	if (verbop) fprintf(stderr, " %s", pf->label) ;
	p = pf->label ;
	if (*p == '_') p++ ;
    	if (!apos) apos = tbf_findlabel(T, "_RArad") ;
	if ((*p == 'G') || (*p == 'E') || (*p == 'S')) p += 2 ;
	if ((char)k != *p) k = (k<<8) | *p ;
    }
    if ((k&0xff00) == 0) {
	/* Accept not an error for NOTES or REFS files */
	p = strstr(lof, "note") ;
	if (!p) {	/* It IS an error ! */
	    fprintf(stderr,
	       "++++Only ONE component of position, remove MPOS flag\n%s\n",
	           note_pos) ;
	    fprintf(stderr, "    for: %s\n", lof) ;
	}
	for (i=0; T->posfmt[i] >= 0; i++) {
	    pf = T->afmt + T->posfmt[i];
	    pf->flags &= ~F_MPOS ;
	    T->posfmt[i] = -1 ;
	}
    }
    re_search = 0 ;
    pf = apos+1 ;		/* Starting Search for pm */

  SCAN_FOR_PM:
    /* Flag the proper motions, if existing */
    apm = (COLINFO *)0;
    for (i=0; (pf<pe) && ((i&3)!=3); pf++) {
	if (pf->flags&F_NODB) continue ;
	if (pf->flags&F_MPOS) continue ;
	if (pf->label[0] != 'p') continue ;
	if (pf->label[1] != 'm') continue ;
	k = 0 ;
	if (((i&1)==0) && (strncmp(&pf->label[2],apos->label,lcmp[0])==0)) 
	    k = 1 ;
	else if (((i&2)==0) && (strncmp(&pf->label[2], latname, lcmp[1])==0))
	    k = 2;
	else if (pf->label[2] == 0) k=5 ;
	else if (strcmp(pf->label, "pmPA") == 0) k=6 ;
	if (!k) continue ;
	i |= k ;
	apm = pf ;
	pf->flags |= F_MPOS ;
	if (verbop) fprintf(stderr, ", %s", pf->label) ;
    }
    if (verbop) fputc('\n', stderr) ;
    if ((i == 0) && (re_search<2)) {	/* pm not found, give a second chance */
	pf = T->afmt ;
	re_search += 1 ;
	if (re_search == 2) { 		/* V6.11: 3rd chance if RAdeg/pmRA */
	    if ((apos[0].label[0] == 'R') && (apos[0].label[1] == 'A')
	      &&(lcmp[0]>2)) lcmp[0] = lcmp[1] = 2;
	    else re_search++;	/* Useless */
	}
	if (re_search <= 2) goto SCAN_FOR_PM ;
    }
    /* Must have the 2 pm components !!! */
    if (i && ((i&3)!=3)) {
	apm->flags &= ~F_MPOS ;
	if (re_search) fprintf(stderr,
	    "++++Only one proper motion (%s) found for: %s\n",
	    apm->label, lof) ;
	else {
	    re_search += 1 ;
	    pf = T->afmt ;
	    goto SCAN_FOR_PM ;
	}
    }

    /* Locate in "RA" the Equinox and Epoch
       epochs returns the 4 strings:
	 [0] = Value of Equinox
	 [1] = Value of Epoch
	 [2] = "Equinox" text
	 [3] = "Epoch" text
    */
    if (!parm) parm = epochs(apos->text, 1);
    /* V6.03: look for equinox/epoch in notes */
    if (((!parm[0])||(!parm[1])) && (section = (apos->note))) {
        len = section->ebody-section->abody;
        p = a = malloc(strlen(apos->text)+strlen(section->header)+len+4);
        strcpy(p, apos->text); p += strlen(p); *(p++) = '\n';
        strcpy(p, section->header); p += strlen(p); *(p++) = '\n';
        if (len>0) strncpy(p, section->abody, len); p[len] = 0;
        parm = epochs(a, 1);
        free(a);
    }
    /* Set the F_MPOS flag for an Epoch being a column */
    if (parm[1]) {
	if (! *parm[1]) {	/* A blank epoch means 'Epoch' */
	    pf = tbf_findlabel(T, "Epoch");
	    if (!pf) pf = tbf_findlabel(T, "epoch") ;
	    if (pf) parm[1] = pf->label;
	}
	if ((pf = tbf_findlabel(T, parm[1]))) {
	    pf->flags |= F_MPOS ;
	    parm[1] = (char *)0 ;
	}
	else {		/* Not a column ... Is it a valid epoch ? */
	    if (! isepoch(parm[1])) {
		fprintf(stderr, "++++Column '%s' at line %d: %s\n",
		    apos->label, apos->lineno, apos->text);
		fprintf(stderr,
	       "    position has no associated Epoch (last analysed word: '%s')\n",
		parm[1]) ;  /*
		  "    '%s' can't be interpreted as Epoch column or value\n",
		  parm[1]) ;  */
		parm[1] = (char *)0 ;
	    }
	}
    }

    /* Ignore the Equinox for not RA / Ecl positions */
    if (parm[0] && (!needs_eq)
      /*(apos->label[0] != 'R') && (apos->label[0] != 'E') */)
	parm[0] = (char *)0 ;

    /* Error if Right Ascension without Equinox */
    if (needs_eq /*((apos->label[0] == 'R') || (apos->label[0] == 'E')) */
       && (!parm[0])) {
	/* Look: is the Equinox at a specific date ? */
	pf = T->afmt + T->posfmt[0];
	i = 0 ;	/* i set to 2 when just a warning */
	if ((n = pf->text)) {
	    if ((p = strstr(n, "quinox"))) i++ ;
	    if (i && strstr(n, "date")) i++ ;
	}
	if ((i<2) && (pf->note)) {
	    i = 0 ;
	    n = pf->note->header;
	    if ((i == 0) && strstr(n, "quinox")) i++ ;
	    if (i && strstr(n, " date")) i++ ;
	    n = pf->note->abody;
	    if ((i == 0) && strstr(n, "quinox")) i++ ;
	    if (i && strstr(n, " date")) i++ ;
	}
	p = (i >= 2 ? "++++" : "****" ) ;
        fprintf(stderr, "%sEquinox Unknown in Position Column '%s' for: %s\n",
	   p, apos->label, lof) ;
	for (i=0; T->posfmt[i] >= 0; i++) {	/* Remove Position */
	    pf = T->afmt + T->posfmt[i];
	    pf->flags &= ~F_MPOS ;
	    T->posfmt[i] = -1 ;
	}
    }

    /* Keep copies of Equinox & Epoch */
  /*KEEP_EQ_EP: */
    i = 0;
    if (parm[0] != T->equinox) {
	if (T->equinox) free(T->equinox), T->equinox = (char *)0;
	if (parm[0]) T->equinox = strdup(parm[0]), i |= 1;
    }
    if (parm[1] != T->epoch) {
	if (T->epoch  ) free(T->epoch  ), T->epoch   = (char *)0;
	if (parm[1]) T->epoch   = strdup(parm[1]), i |= 2;
    }

    return(i);
}

/*==================================================================
		Internal Utilities
 *==================================================================*/

static int edFULE(os, fout, pF, pU, pL, pE, pnote)
/*++++++++++++++++
.PURPOSE  Edit the Format, Units, Label, Explanations
.RETURNS  Number of Bytes used until Explanation
-----------------*/
  int os[4];		/* IN: Offsets of Format, Units, Label, Explain	*/
  FILE *fout;		/* IN: Opened output file			*/
  char *pF;		/* IN: Format to write out			*/
  char *pU;		/* IN: Units to write out			*/
  char *pL;		/* IN: Label to write out			*/
  char *pE;		/* IN: Explanations to write out		*/
  SECTION *pnote;	/* IN: Note (to generate the Anchor)		*/
{
  char *p; int o, o1, nb;


#if 0
    nb = (os[3] - os[1]) - (strlen(pU)+strlen(pL)+2);
    o1 = os[1]; if (nb < 0) o1 += nb;
#endif
    /* Write out the Format	*/
    o = os[0];
    nb = (os[1] - os[0]) - (strlen(pF)+1);
    if (isalpha(*pF) && isdigit(pF[1]) && (nb > 0)) fputc(' ', fout), o++;
    for (p=pF; *p; p++, o++) fputc(*p, fout);
    fputc(' ', fout); o++ ;

    /* Check Units + Label: nb = number of additional blanks */
    nb = (os[2] - os[1]) - (strlen(pU)+1);
    o1 = os[1];
    if (nb < 0) o1 += nb;
    while (o < o1) fputc(' ', fout), o++;
    if (unit_A) fprintf(fout, "%s{", unit_A);
    for (p=pU; *p; p++, o++) (*FPUTC)(*p, fout);
    if (unit_A) fputc('}', fout);
    fputc(' ', fout), o++;

    /* Write out the  Label	*/
    nb = (os[3] - os[2]) - (strlen(pL)+1);
    o1 = os[2]; if (nb < 0) o1 += nb;
    while (o < o1) fputc(' ', fout), o++;
    /* if (pnote) fprintf(fout, "\\Aref{%p}{", pnote); */
    for (p=pL; *p; p++, o++) (*FPUTC)(*p, fout);
    /* if (pnote) fputc('}', fout); */
    fputc(' ', fout), o++;

    /* Write out the Explanations */
    while (o < os[3]) fputc(' ', fout), o++;
    if (*pE != '*') fputc(' ', fout), o++;
    p = pE ;

    if (pnote) {	/* Make an anchor on Explanation */
	if (*p == '*') fputc(*p, fout), p++;
	if (islim(*p)) {
	    fputc(*p, fout), p++;
	    while (*p && (!islim(*p))) (*FPUTC)(*p, fout), p++;
	}
	while (ispunct(*p)) (*FPUTC)(*p, fout), p++;
	while (isspace(*p)) (*FPUTC)(*p, fout), p++;
	/*fprintf(fout, "\\Aref{%s}{", sec_id(pnote));*/
	fprintf(fout, "\\yAref{move to details of note}{%s}{", sec_id(pnote));
	while (isprint(*p)) fputc(*p, fout), p++;
	fputc('}', fout);
	/* fprintf(fout, "%s\n", p); */
    }
    while (*p) (*FPUTC)(*p, fout), p++;
    fputc('\n', fout);

    return(o);
}

#if DEBUG
static char *ed_lim(format, n)
/*++++++++++++++++
.PURPOSE  Edit the limit (0,1,2)
.RETURNS  Edited format
-----------------*/
  COLINFO *format;	/* IN: Fortran Format */
  int n;		/* IN: 0=lower lim., 1=upper lim., 2=NULL */
{
  static char fmt[8], buf[60];
  double *pd; int4 *pl;

    if (!format) return((char *)0);
    if (format->len == 0) return((char *)0);
    switch(n) {
      case 0:	/* lower */
      	if (!(format->elims & F_IMIN)) return((char *)0);
      	break;
      case 1:	/* upper */
      	if (!(format->elims & F_IMAX)) return((char *)0);
      	break;
      default:
      	n = 2;
      	if (format->elims & F_REQD) {
      	    if (format->elims&F_NULL) break;
      	    return((char *)0);	/* no NULL allowed */
     	}
     	else return(" ");
    }
    switch(format->ff) {
      case 'E': case 'D':
      	sprintf(fmt, "%%%d.%de", format->len, format->len - format->odp -1);
      	break;
      case 'I':
#if int8l == 2
      	sprintf(fmt, "%%%dlld", format->len);
#else
      	sprintf(fmt, "%%%dld", format->len);
#endif
      	break;
      default:
      	sprintf(fmt, "%%%d.%d%c", format->len, format->len - format->odp -1,
      	    tolower(format->ff));
    }
    if (format->elims&F_DBLE) {
      	pd = (double *)format->lims;
    	sprintf(buf, fmt, pd[n]);
    } else {
      	pl = (int8 *)format->lims;
    	sprintf(buf, fmt, pl[n]);
    }
    return(buf);
}
#endif

/*==================================================================
		Dealing with Format File
 *==================================================================*/

int tbf_print(T, fout, P_opt)
/*++++++++++++++++
.PURPOSE  Print out the "Byte-by-Byte description" (set of column formats)
.RETURNS  Number of columns edited
-----------------*/
  TABINFO *T;		/* IN: Various details about position of description */
  FILE   *fout;		/* IN: Opened output file 	*/
  int P_opt;		/* IN: Printing Option		*/
{
  COLINFO *pf; char *p; SECTION *note;
  int i, os[4];
#if DEBUG
  int k;
  static char *tit[] = {"TAMIN", "TAMAX", "TBNUL"};
#endif

    if (P_opt) 	rule_char = '\\';
    else 	rule_char = '-' ;
    unit_A = (char *)0;
    note   = (SECTION *)0;

    /* Look if the Section contains the titles, to compute
       the default offsets of Format, Units, Label and Explanations
    */
    os[3] = os[2] = os[1] = os[0] = 0;
    if ((p = T->section->abody)) {
    	while (*p != ' ' && (p < T->section->ebody))
    	    p += strloc(p, '\n') +1;
    	os[0] = strloc(p, 'F');
    	os[1] = strloc(p, 'U');
    	os[2] = strloc(p, 'L');
    	os[3] = strloc(p, 'E')-1;
    	if (os[3] >= 50) os[3] = 0;
	if ((os[1] - os[0]) < 7) os[3] = 0;
	if ((os[2] - os[1]) < 5) os[3] = 0;
	if ((os[3] - os[2]) < 7) os[3] = 0;
    }
    if (os[3] < 20) {
    	os[0] = 9; 		/* Format starts at Byte 9 	*/
    	os[1] = os[0]+6+1; 	/* width of Format is 6 	*/
    	os[2] = os[1]+7+1; 	/* width of Units  is 7		*/
    	os[3] = os[2]+8+1;	/* width of Label  is 8		*/
    }

    /* Write out the Header " Byte Format Unit " etc...	*/
    fprintf(fout, "%s\n", rule(rule_char));
    fprintf(fout, " %7s ", " Bytes");
    edFULE(os, fout, "Format", "Units", "Label", "Explanations", (SECTION *)0);
    fprintf(fout, "%s\n", rule(rule_char));

    for (i = T->nfmt, pf = T->afmt; --i >= 0; pf++) {
   	if (pf->b0 == pf->b1) fprintf(fout, " %7d ", pf->b0+1);
   	else fprintf(fout, " %3d-%3d ", pf->b0+1, pf->b1+1);
	if (P_opt)
	    note   = pf->note,
	    unit_A = *pf->unit == '-' ? (char *)0 : "\\Unit";
   	edFULE(os, fout, fmt_ed(pf), pf->unit, pf->label, pf->text, note);
#if DEBUG
	fprintf(fout, "\t<%04X>\n", pf->elims & 0xffff);
    	for (k=0; k<=2; k++) {
            if(p=ed_lim(pf, k)) fprintf(fout, "\t%s='%s'\n", tit[k], p);
        }
        if (pf->note) fprintf(fout, "----%s\n%s\n",
        	pf->note->header, pf->note->abody);
#endif
    }
    fprintf(fout, "%s\n", rule(rule_char));
    return(T->nfmt);
}

/*==================================================================*/

COLINFO *tbf_addcol (pt, pos, name, text, value, unit)
/*++++++++++++++++
.PURPOSE  Add a new (constant) column in a Table Description
.RETURNS  The new column created
.REMARKS  unit and name are just USED -- should not be freed !
-----------------*/
  TABINFO *pt;	/* MOD: Various details about position of description */
  int pos;	/* IN: Position >=0 for absolute, <0 for relative from end */
  char *name;	/* IN: Name of column (can't be NULL) */
  char *text;	/* IN: Explanation of Column (can be NULL)  */
  char *value;	/* IN: Value of the column (not NULL = constant column) */
  char *unit;	/* IN: Unit in which value is expressed */
{
  static COLINFO acol;	/* The new column */
  COLINFO *pc, *pe;
  int i, i0, ncols;

    if (strcmp(pt->signature, "\3FIL") == 0)	/* I have a FILINFO structure */
	pt = ((FILINFO *)pt)->tabform;
    if (strcmp(pt->signature, "\3TAB") != 0) {
	fprintf(stderr, "****tbf_addcol to non-TABINFO: %s\n", name);
	return((COLINFO *)0);
    }

    /* Verify the name is not yet existing */
    if (tbf_findlabel(pt, name)) {
	fprintf(stderr, "****tbf_addcol name exists: %s\n", name);
	return((COLINFO *)0);
    }

    /* Prepare the new definition */
    if (!acol.signature) {
        acol.signature = cad_signature("\3COL");
        acol.rep = 1;
        acol.ff = '\1';
	acol.unit = unitless ;
	acol.text = "";
	acol.flags = F_Aname;	/* Name is duplicated */
    }
    acol.oid = cad_oid();	/* A new identifier   */

    /* Compute the absolute position of the new column */
    if (pos >= 0) i0 = pos;
    else i0 = pt->nfmt + 1 + pos;
    if (i0 < 0) i0 = 0;
    if (i0 > pt->nfmt) i0 = pt->nfmt;

    /* ... There is 1 more column ... */
    /* Actual number of colums: pasted tables in lineno */
    i = pt->nfmt ;  pt->nfmt += 1 ;
    /* V6.15: there are hidden columns in case of pasted tables */
    ncols = pt->nfmt+1; if (pt->lineno<0) ncols -= pt->lineno;
    if (unit) {
	if (!*unit) unit = (char *)0;
    }
    pt->afmt = realloc(pt->afmt, ncols*sizeof(COLINFO));

    /* Move the column definitions to insert the new column */
    for (pc=pt->afmt + ncols-1, pe=pt->afmt + i0; pc > pe; pc--)
	memcpy(pc, pc-1, sizeof(COLINFO));
    /* ... and change also the posfmt array !   */
    for (i=0; (i<8) && (pt->posfmt[i] >= 0); i++) {
	if (pt->posfmt[i] >= i0) pt->posfmt[i] += 1;
    }

    acol.b0 = acol.b1 = 0;  /* Added V6.20 */
    if (pc>pt->afmt) {	/* Column inserted not at the beginning */
	--pc ; acol.b0 = acol.b1 = pc->b1; 
	++pc; 
    }
    /* Insert the new definitions (here, pc = pe) */
    memcpy(pc, &acol, sizeof(COLINFO));
    pc->label = strdup(name);
    if (text) {
	pc->text = strdup(text);
	pc->flags |= F_Atext;
	/* V6.14: Attach note... */
	i = strlen(pc->text)-1;
	if ((i>0) && (pc->text[i] == ')')) {
	    for (--i; (i>0)&&isdigit(text[i]); i--) ;
	    if (text[i] == 'G') /* Global note */ i--;
	    if ((i>0) && (text[i] == '(')) { 
		pc->note = sec_findnote(pt->section, text+i);
		if (!pc->note) fprintf(stderr, 
		    "#+++tbf_addcol: Note %s not found for column: %s\n",
		    text+i, name);
	    }
	}
    }
    if (value) {		/* Constant Column */
	pc->len   = strlen(value);
	if (pc->len == 0) { pc->len += 1; pc->value = strdup(" "); }
	else pc->value = strdup(value);
	pc->elims |= F_CCST;
	pc->ff = 'A';
    }
    if (unit) {
	pc->unit = strdup(unit);
	pc->flags |= F_Aunit;
    }

    /* The F_ACCU flags may have changed. Recompute those */
    mark_accu(pt, 0);

    return(pc);
}

int tbf_order (pt, name, pos)
/*++++++++++++++++
.PURPOSE  Order the table: put the column "name" after "pos"
.RETURNS  0 (no change) / 1 (OK, changed) / -1 (impossible)
.REMARKS  Accept pos as "-name" (place before)
-----------------*/
  TABINFO *pt;	/* MOD: Various details about position of description */
  char *name;	/* IN: Name of column (can't be NULL) */
  char *pos;	/* IN: Position >=0 for absolute, <0 for relative from end */
{
  COLINFO acol;	/* temporary column */
  COLINFO *p0, *p1;
  int i, n, i0, i1;
  char after;

    if (strcmp(pt->signature, "\3FIL") == 0)	/* I have a FILINFO structure */
	pt = ((FILINFO *)pt)->tabform;
    if (strcmp(pt->signature, "\3TAB") != 0) {
	fprintf(stderr, "****tbf_order(..., %s, %s) to non-TABINFO\n",
	    name, pos);
	return(-1);
    }
    p0 = tbf_findlabel(pt, name);
    if (!p0) return(-1);
    i0 = p0 - pt->afmt;			/* Position of column to move */
    memcpy(&acol, p0, sizeof(COLINFO));

    if (*pos == '+') after = 1, pos++;
    else if (*pos == '-') after = 0,  pos++;
    else after = 1;	/* Default = after */
    while (*pos == ' ') pos++;
    if (*pos) p1 = tbf_findlabel(pt, pos);
    else {		/* No column specified ==> begin or end.. */
	p1 = pt->afmt;	/* First column */
	if (after) p1 += pt->nfmt-1;	/* Last column */
    }
    if ((!p1) && strchr(pos, '*')) {	/* V6.0: Accept wild card */
      char *a;
	a = malloc(2+strlen(name)+strlen(pos));
	strcpy(a, pos); *strchr(a, '*') = 0;	/* Remove the '*' */
	strcat(a, name[1] == '_' ? name+2 : name);
        p1 = tbf_findlabel(pt, a);
	free(a);
    }
    if (!p1) return(-1);

    n = 0;	/* Counter of number of displacements */
    if (p0 < p1) {			/* Move column to the Right */
	if (after); else --p1;		/* Set p0 BEFORE p1 */
        i1 = p1 - pt->afmt;
	while (p0 < p1) { memcpy(p0, p0+1, sizeof(COLINFO)); p0++; n=1; }
    }
    else {				/* Move column to the Left  */
	if (after) ++p1;
        i1 = p1 - pt->afmt;
	while (p0 > p1) { memcpy(p0, p0-1, sizeof(COLINFO)); p0--; n=1; }
    }
    if (n == 0) return(0);

    memcpy(p1, &acol, sizeof(COLINFO));
    /* Beware, change of indices in posfmt ... (Mod. V6.16) */
#if 0
    if (i0 > i1) { i = i0; i0 = i1; i1 = i; n = 1; }
    else n = -1;
    for (i=0; (i<8) && (pt->posfmt[i] >= 0); i++) {
        if ((pt->posfmt[i] >= i0) && (pt->posfmt[i] <= i1))
    	pt->posfmt[i] += n;
    }
#else
    for (i=0; (i<8) && (pt->posfmt[i] >= 0); i++) { 
	if (pt->posfmt[i] == i0) { pt->posfmt[i] = i1; continue; }
	if ((i0<i1) && (pt->posfmt[i]>i0) && (pt->posfmt[i]<=i1)) 
	    pt->posfmt[i] -= 1; 
	else
	if ((i0>i1) && (pt->posfmt[i]>=i1) && (pt->posfmt[i]<i0))
	    pt->posfmt[i] += 1;
    }
#endif
    return(1);
}

int tbf_load(T, fop, f_op)
/*++++++++++++++++
.PURPOSE  Load the "format" Structure
.RETURNS  Number of errors
-----------------*/
  TABINFO *T;	/* MOD: Various details about position of description */
  int fop;	/* IN: Options FOP_xxx	*/
  int f_op ;	/* IN: Options F_xxxxx: F_isFK is a refs file exists  */
{
  COLINFO *pf, *qf; SECTION *section, *ps;
  char *p, *q, *pp=(char *)0, *e, *eol, *limits, *nullo, x;
  char flagerr ;
  int i, j, nb, errors, oline, count, lineno;
  short bp;		/* Preceding Byte */
  int8 *pl;
  double *pd, factor;

    section = T->section;
    errors = 0;
    if (section->class == SEC_Col) {
	fprintf(stderr,
	  "++++tbf_load: Char-Sep-Values not yet supported at line %d\n  %s\n",
	  section->lineno, section->header);
    }
#if DEBUG
    printf("----load_format(%p,%X) at line %d\n", T, fop, T->lineno+1);
#endif

    /* Count the number of columns in the Byte-by-Byte description */
    for (count = 0, p = section->abody, lineno = T->lineno;
    	p < section->ebody; p += i+1,lineno++) {
    	i = strloc(p, '\n');
    	if (match_rule(p)) 	/* Ignore the rule */
    	    continue;
	if (isReadMe_comment(*p)) continue;	/* Comment, V6.09 */
	/*  if (*p == '#') continue ;	[old] Comment */
    	for (nb=0, q=p; isspace(*q) && (nb < 12); q++) switch(*q) {
    	    case '\t': nb += 6; continue;
    	    case ' ':  nb++;	continue;
    	    case '\n': nb = 80; continue;
    	}
	if (nb >= 12) continue;	/* Long indentations can't be OK */
	if (isdigit(*q) || (fop&FOP_1X)) count++;
    }
    if (count == 0) {
	if ((p = section->header))
	    p = strchr(p, ':');
	if (p) fprintf(stderr, "++++No description was found for:%s\n", p+1);
	else   fprintf(stderr, "++++No description found.\n");
    }

	/* Allocate the Structure */
#if DEBUG
    printf("....There are %d columns to allocate\n", count);
#endif
    T->nfmt = count;
    T->afmt = pf = (COLINFO *) malloc(T->nfmt * sizeof(COLINFO));
    bzero(pf, T->nfmt * sizeof(COLINFO));
    pf->signature = cad_signature("\3COL") ;
    errors = 0;

	/* Fill the COLINFO structure from the input made of fields
 	   Position[X]  [-position]  Fortran-Format  Units  Label  Explain
   	*/
    oline = 0;		/* Offset in line  = relative position 	*/
    bp = -1;		/* End of previous column		*/
    eol= e = (char *)0; /* End-of-line indicator 		*/
    for (lineno=T->lineno, p = section->abody; p < section->ebody; p = e+1) {
	/* Give UP after too many errors ! */
	if (errors > MAX_ERRORS) {
	    fprintf(stderr, "**** line %d: too many errors, giving up!\n",
	      lineno);
	    tbf_free(T);
	    return(errors);
	}
	lineno += 1;
	i = strloc(p, '\n');
	eol = e; e = p+i;	/* eol indicates the end of the explanations */
#if DEBUG
	printf("\nLine%4d: ", lineno); fflush(stdout);
	while(isspace(p[i])) i--;
	write(1,p,i);
#endif
    	if (match_rule(p)) 	/* Ignore the rule: don't set eol */
    	    continue;
	if (isReadMe_comment(*p)) {		/* Comment, V6.09 */
	    /* if (*p == '#') -- Line starting by # is comment  */
	    if (eol) while (isspace(*eol) && (eol >= section->abody))
		*(eol--) = 0;
	    continue ;
	}
    	for (nb=0; isspace(*p) && (nb<12); p++) switch(*p) {
    	    case '\t': nb += 6; continue;
    	    case ' ':  nb += 1;	continue;
    	    case '\n': nb = 80; continue;
    	}
	/* We're describing a new column if not too many blanks
	   and first character is a digit (unless -f1X option).
	   Otherwise we're in a continuation line
	*/
	if ((nb < 12) && (isdigit(*p) || (fop&FOP_1X))) ;
	else { eol = (char *)0; continue; }
	if (eol) while (isspace(*eol) && (eol >= section->abody)) *(eol--) = 0;
#if DEBUG
	printf("\t---->New colum");
#endif

	pf->lineno = lineno;
        if (!pf->oid) pf->oid = cad_oid() ;
	limits = nullo = (char *)0;

	/* With option -f1X, a line can start with non-digit.
	   Here, flagerr is set if column limits are missing.
	   The value i is negative for relative position.
	*/
	flagerr = 0 ;
	i = -1 ;
	if (isdigit(*p)) {
	    i = atoi(q = p) ;
	    while(isdigit(*p)) p++ ;
	    if (isff(*p)) {		   /* Can't be Limits	*/
		i = -1, p = q;
		flagerr = !(fop&FOP_1X) ;
	    }
	    else if (toupper(*p) == 'X')   /* Relative Position	*/
		i = -i, p++;
	}
	else flagerr = !(fop&FOP_1X) ; /* No position specified	*/

	if (fop&FOP_1X) i = -1 ;
	if (i <= 0) oline -= i ;
	else oline = i-1 ;
	if (flagerr) fprintf(stderr,
	    "++++ line %d: no position, 1X assumed\n",
	    lineno);
	pf->b1 = pf->b0 = oline; pf->rep = 1;
	while(iswhite(*p)) p++;
	if (*p == '-') { 		/* End of Column      	*/
	    for (++p; iswhite(*p); p++) ;
	    pf->b1 = atoi(p) - 1;
	    while(isdigit(*p)) p++; while (iswhite(*p)) p++;
	}
	if (isdigit(*p)) {		/* Repetition Factor  */
	    pf->rep = atoi(p);
	    while(isdigit(*p)) p++;
	    while(iswhite(*p)) p++;
	}

	pf->ff = toupper(*p); p++; 	/* Fortran Format     */
	if (*p == '0') pf->flags |= F_ZERO ;
	switch(pf->ff) {
	  case 'D': case 'E': case 'F':
	    pf->elims |= F_DBLE|F_REQD;
	    if (fop&FOP_ra) {	/* Compute Range */
            	pf->range = malloc(2*sizeof(double));
            	pd = (double *)pf->range;
            	pd[0] = 1.e40; pd[1] = -1.e40;
	    } break;
	  case 'A':
	    pf->elims |= F_CHAR;
	    if (fop&FOP_ra) {	/* Compute Range */
	    	pf->range = malloc(256);
	    	bzero(pf->range, 256);
	    } break;
    	  case 'I':
	    pf->elims |= F_REQD;
	    if (fop&FOP_ra) {	/* Compute Range */
            	pf->range = malloc(2*sizeof(int8));
            	pl = (int8 *)pf->range;
            	pl[0] = ~NULL8; pl[1] = NULL8;
	    } break;
	  default:
	    errors++;
	    fprintf(stderr, "**** line %d: '%c' isn't a format letter\n",
	    	lineno, pf->ff);
	}
	pf->len = atoi(p); while(isdigit(*p)) p++;
	pf->odp = 0;
	if (*p == '.') pf->odp = pf->len - atoi(++p) -1;
	while (isdigit(*p)) p++;
	if (!isspace(*p)) errors++,
	    fprintf(stderr, "**** line %d: what is '%c' as a format letter?\n",
	    lineno, *p);
	while(iswhite(*p)) p++;
					/* Compute Column End */
	oline += pf->rep * pf->len;
	if (fop & FOP_1)  pf->b1 = oline - 1;
					/* Unit of Column     */
	if (fop & FOP_dot)  pf->unit = unitless;
	else {
	    pf->unit = p; while(isgraph(*p)) p++;
	    if(*p != '\n') *(p++) = 0;
	    while(iswhite(*p)) p++;
	}
					/* Complex Numbers   */
	if (pf->unit[0] == '"') {
	    /* Verify first it's a known unit
	       pf->flags is set to corresponding unit.
	       Note that "month" doesn't show actual unit!
	       Here i is set to the number of invalid letters in format.
	    */
	    i = 0;
	    if (strcmp(pf->unit, "\"month\"") == 0)
		pf->flags |= F_DATE;
	    else if ((q = unit_si(pf->unit, &factor))) {
		if (*q == 's') {
		     pf->flags |= factor>1.e4 ? F_DATE : F_TIME;
		     if (strstr(pf->unit, "time")) /* Added V6.02 */
			 pf->flags |= F_TIME;
		}
		else pf->flags |= F_ANGL;
		/* Added V5.94: Set the F_SSEP flag */
		for (q=pf->unit+1; isalpha(*q); q++) ;
		if (ispunct(*q) && (*q != '"')) 
		    pf->flags |= F_SSEP;
	    }
	    else for (q=pf->unit+1; *q && (*q != '"'); q++) switch(*q) {
		case 'y': case 'Y': case 'D': case 'M':
			  pf->flags |= F_DATE ; continue ;
		case 'H': case 'h':
		          pf->flags |= F_TIME ; continue ;
		case 'd': if (strncmp(q, "day", 3) == 0) q += 2;
			else pf->flags |= F_ANGL ; continue ;
		case 'm': 		/* Can be angle or time */
			if (strncmp(q, "month", 5) == 0) q += 4;
		case 's': case 'S':	/* Can be angle or time */
			if ((pf->flags&(F_TIME|F_DATE|F_ANGL))==0)
			  pf->flags |= F_TIME ; continue ;
		default:  if (ispunct(*q)) pf->flags |= F_SSEP;
			  else i++;
	    }
	    /* Incorrect unit if simultaneously ANGLE and TIME */
	    if (i>0)
		q = "can't be understood.";
	    else if ((pf->flags&(F_DATE|F_TIME|F_ANGL)) == 0)
		q = "can't be characterized as TIME or ANGL";
	    else if ((pf->flags&(F_DATE|F_TIME)) && (pf->flags&F_ANGL))
		    q = "contains both ANGL and DATE/TIME?";
	    else q = (char *)0;
	    if (q) fprintf(stderr, "Unit %s %s\n", pf->unit, q);
	}
					/* Label of Column   */
	pf->label = p; while(isgraph(*p)) p++;
	if(*p != '\n') *(p++) = 0;  while(iswhite(*p)) p++;

	pf->text = p;			/* Explanations	     */
	if (isnull(*p)) nullo = p++;
	if (*p == '*')  		/* Note (*)	     */
	    pf->note = (SECTION *)pf->label, p++;
	if (isnull(*p)) nullo = p++;
	if (islim(*p)) limits = p;

			    /* Implied limits from "known_label" */
        if (!limits) {  LABEL *plab;
            if ((plab = lab_find(pf->label))) {	/* Found the defaults */
#if DEBUG>1
	    	fprintf(stderr, "....Found known_label '%s' '%s' '%s'",
	    	        plab->name, plab->limits, plab->unit);
#endif
	    	limits = plab->limits;
	    }
	}

	if (limits) { 	/* Compute pp -> closing bracket */
	    if (pf->elims & F_CHAR) {  /* A-format: allowed char set */
	  	pp = limits + (limits[1] == ']');
	    	for (++pp; isprint(*pp) && *pp != ']'; pp++) ;
		/* Added V5.95: accept [[XX89]] */
		if ((limits[1] == '[') && (pp[1] == ']')) pp++;
	    }
	    else {  		/* Numeric limits */
	    	for (pp=limits+1; isprint(*pp) && (!islim(*pp)); pp++) ;
	    }
				/* Char just following the ] */
	    q  = pp;
	    if (islim(*q)) q++;
	    if (ispunct(*q)) nullo = q;
    	}
    		/* Allocate lims if limits given OR a NULL value is provided */
	if (limits || nullo) {	/* Store limits + explicit NULL */
	    if (pf->elims & F_CHAR) nb = 256;
	    /* [0,1]=limits [2]=null_value [3]=value */
	    else /*if (pf->elims & F_DBLE) */ nb = 4*sizeof(double);
	    /*else nb = 4*sizeof(int4); */
	    pf->lims = malloc(nb); bzero(pf->lims, nb);
	}

	if (limits) { 	/* Remind: pp->closing bracket */
	    if (pf->elims & F_CHAR) { /* A-format: allowed character set
	    				 using tr(1) conventions.  */
	    	pf->elims |= F_ATRT;	/* A list of allowed bytes   */
		setset1e(limits+1, pp, pf->lims);
		if (*pp != ']' ) errors++, fprintf(stderr,
		    "**** line %d: unterminated '['char_set']'\n", lineno);
#if DEBUG
		    printf("\n....Column '%s': character set %s]=[",pf->label,
		    	limits);
		    for (i=0; i<256;i++) if(pf->lims[i]) putchar(i);
		    printf("]\n");
#endif
	    }
	    else {  		/* Limits within brackets */
	    	double dval[2], atof(); int n;
	    	static int bound[2] = { F_IMIN, F_IMAX };
	    	q = limits; dval[0] = dval[1] = 0;
#if DEBUG
		{ char *pt, x;
		    for (pt = limits; isgraph(*pt); pt++);
		    x = *pt; *pt = 0;
		    printf(
		    	"\n....Range %s for column: %s\n",
		    	limits, pf->label);
		    *pt = x;
		}
		printf("");
#endif
	    	if (*q == ']') pf->elims |= F_XMIN;
	    	for (j=0; j < 2; j++) {
	    	    for (++q; iswhite(*q); q++);
		    /* V6.10: limit [,*] means UPPER BOUND */
		    if (*q == '*') { /* Interpret as NaN (all bits set to 1) */
			pf->elims |= F_MCMP;
			memset(&(dval[j]), -1, sizeof(dval[j]));
			n = 1;
		    }
		    else {  /* Actual number, get it */
	    	        dval[j] = atof(q); /* limits = q; */
		        n = match_num(q);	/* Length of Number	*/
		    }
		    q += n;		/* Flating-point number */
		    /* Take care of a single number e.g. [2]	*/
	            if (n) pf->elims |= bound[j];
		    if (islim(*q) && (j == 0)) {	/* V4.91*/
		    	dval[1] = dval[0] ;	/* V4.91 */
			if (n) pf->elims |= bound[1] ;
			break;
		    }
		}
		if (!islim(*q)) errors++, fprintf(stderr,
		    "**** line %d: unterminated [inf/sup] limits (char '%c')\n",
		    lineno, *q);
	    	if (*q == '[') pf->elims |= F_XMAX;
	    	if (pf->elims & F_DBLE) { double *lim;
		    lim = (double *)pf->lims;
		    lim[0] = dval[0];
		    lim[1] = dval[1];
	        } else { 		int8 *llim;
		    llim = (int8 *)pf->lims;
		    /* V6.10: limit set to NaN for upper/lower bound */
		    /* WARNING: the cast (int) is REQUIRED! (V6.12)  */
		    llim[0] = dval[0] == dval[0] ? (int8)dval[0] : NULL8; 
		    llim[1] = dval[1] == dval[1] ? (int8)dval[1] : NULL8;
	        }
	    }
				/* Char just following the ] */
	    if (islim(*pp)) pp++;
	    if (isnull(*pp)) nullo = pp++;
	}
	if (nullo) for (pp = nullo; ispunct(*pp); pp++) switch(*pp) {
	    case '!': 	/* REQUIRED value */
		pf->elims |= F_REQD; continue;
	    case '?': 	/* Can be blank   */
		pf->elims &= ~F_REQD;
		if (pp[1] == '=') { 	/* Explicit NULL value */
		  double v, atof(); int n; char *anull;
		    pp += 2;		/* Skip ?= */
		    pf->elims |= F_NULL;
		    /* j contains number of digits */
		    for (n=j=0; isgraph(pp[n]); n++) j += isdigit(pp[n]);
		    if (n==0) errors++, fprintf(stderr,
			"**** line %d: ?= MUST be followed by the NULL value\n",
			lineno);
		    /* V5.99: accept NULL in char strings */
		    if (pf->elims&F_ATRT/*F_CHAR*/) errors++, fprintf(stderr,
		    	"**** line %d: can't have A-format NULL+list ([]?=)\n",
		    	lineno);
		    else if (pf->elims&F_CHAR) { 	/* NULL in char */
			if (n>=20) n=20;
			strncpy(pf->lims, pp, n);
			pf->lims[n] = 0;
		    }
		    else if (j>0) {		/* Numeric value */
		        /* if (n == 0) 	-- V5.92: accept ?=* ?=~ ?== 
		           n = 1;
			 */
		        /* if(!(isdigit(pp[1])||issign(pp[1]))) continue; */
		        v = atof(pp);
			pf->elims |= F_REQD; 	/* Explicit NULL means value */
		    	if (pf->elims & F_DBLE)
		    	    pd = (double *)(pf->lims), pd[2] = v;
		    	else pl = (int8  *)(pf->lims), pl[2] = v;
		    }
		    else {		/* V5.92: one single char for NULL */
			pf->elims |= F_CHAR;	/* Limits in CHARACTER */
			anull = pf->lims + 2* (pf->elims&F_DBLE ?
				  sizeof(double) : sizeof(int4));
			*anull = *pp;	/* Save the char used for null's   */
		    }
		    		/* Skip all chars for floating-point value */
		    pp += n;
#if 0			/* Pre-v5.92: match explicitely the characters */
		    if (issign(*pp)) pp++;
		    while (isdigit(*pp) || *pp == '.') pp++;
		    if (isexpo(*pp)) {
		    	++pp;
		    	if (issign(*pp)) pp++;
		    	while(isdigit(*pp)) pp++;
		    }
#endif
		    pp--;
		} continue;
	    case '-':
		pf->elims |= F_SRTD;	/* NO BREAK */
	    case '+':
		pf->elims |= F_SORT; continue;
	    case '=':
		pf->elims |= F_SRTE; continue;
	    case '(': case ')':		/* Accept ?(name) */
		continue;
	    default:
	    	errors++, fprintf(stderr,
	    	"**** line %d: char <%c> can't be interpreted\n", lineno, *pp);
	}
	if (pf->elims & F_SORT)	{	/* We must keep in memory last field */
	    pf->value = malloc(pf->len+1);
	    bzero(pf->value, pf->len+1);
	    if (pf->elims & F_SRTD) *(pf->value) = '\177';
	    if (!(pf->elims&F_SRTE)) pf->flags |= F_isPK|F_INDX ;
	}

			/* Check the COLINFO */
	if (pf->b0 <= bp)  errors++, fprintf(stderr,
	    "**** line %d: starting column too small: %d < %d\n",
		lineno, pf->b0 + 1, bp);
	if (pf->b1 < pf->b0)  errors++, fprintf(stderr,
	    "**** line %d:  ending  column too small: %d-%d\n",
	    	lineno, pf->b0+1,pf->b1+1);
	i = 1 + pf->b1 - pf->b0;
	if (i != (pf->rep * pf->len)) errors++, fprintf(stderr,
	    "**** line %d: Bad Format (should be %c%d)\n",
	       lineno, pf->ff, i);
	if (pf->note) {	/* Find out the section where note is found */
	    if ((ps = sec_findnote(section, (char *)pf->note)))
	    	ps->used |= SECNOTE_used, pf->note = ps;
	    else pf->note = (SECTION *)0; /* Error detected in tbf_check */
	}
	/* === Assign flags according to label ===
	   A reference may be r_... of ref(...) */
	if (f_op&F_isFK) {	/* A refs file exists. Flag r_ fields */
	    if ((pf->label[0] == 'r') &&
	       ((pf->label[1]=='_') || (strncmp(pf->label, "ref("/*)*/,4)==0)))
		pf->flags |= F_isFK ;
	}

	/* === Verify label is NOT duplicated ==== */
	for (qf = T->afmt; qf < pf; qf++) {
	    if (strcmp(pf->label, qf->label)) continue;
	    if (strcmp(pf->label, "---") == 0) continue;
	    fprintf(stderr, "**** line %d: label '%s' also at line %d\n",
		pf->lineno, qf->label, qf->lineno);
	    errors++;
	}
	bp = pf->b1; pf++;
    }
    if (eol) while (isspace(*eol)) *(eol--) = 0;

    /* Find out the various notes */
    for (pf=T->afmt, i=T->nfmt; --i>=0; pf++) {
    	if (pf->note) continue;
    	p = pf->text; if (!p) continue;
	/* The end of the Explanation can be marked by a long ------ */
	if ((e = strstr(p, "-------------"))) {
	    for (--e; (e>p) && isspace(*e); --e) ;
	    if (e > p) p = e;
	    else p += strlen(p) - 1;
	}
    	else p += strlen(p) - 1;
    	if (*p == ')' && isdigit(p[-1])) {	/* Numbered (<1000!) note */
    	    for (--p; isdigit(*p); p--) ;
	    /* Allow GLOBAL Notes, e.g. (G12) ---
	       but these must be AFTER the table description !!!
	    */
	    if (atoi(p+1) >= 100) p = ""; 	/* Note can't be > 99  */
	    else if ((*p == 'G') || (*p == 'N')) p-- ;
    	    if ((p[0] == '(') && (p[1] != '0')) /* V6.19: (0) is NOT a note */
    	    	pf->note = (SECTION *)p;
    	}
	if (pf->note) {			/* Find out Numbered Note */
	    pf->elims |= F_NOTE;
	    p = (char *)pf->note ;
	    nb = strloc(p, ')') + 1;
	    x = p[nb]; p[nb] = 0 ;
	    if ((ps = sec_findnote(section, (char *)pf->note)))
	    	ps->used |= SECNOTE_used, pf->note = ps;
	    else  pf->note = (SECTION *)0;
	    p[nb] = x ;
	}
    }

    /* Mark the complex (i.e. sexagesimal) values */
    mark_accu(T, fop&FOP_v);

    /* Find the Position (RA, Dec), and Epochs + Equinox */
    tbf_mpos(T, (char *)0);

#if DEBUG
    printf("\n----Format of (%p,%X) at line %d", T, fop, T->lineno+1);
    inbyte = getchar();
    tbf_print(T, stdout, 0);
    inbyte = getchar();
#endif

    return(errors);
}

void tbf_cpy(dest, src)
/*++++++++++++++++
.PURPOSE  Copy a "Byte-by-Byte" to a new place
.RETURNS  0 (OK) / -1 (error)
-----------------*/
  TABINFO *dest;	/* OUT: Duplicated table description */
  TABINFO *src;		/* IN: Source Description */
{
  COLINFO *pc, *pe ;
  int nb ; char *p ;

    /* Duplicate the TABINFO */
    memcpy(dest, src, sizeof(TABINFO)) ;
    if (dest->equinox) dest->equinox = strdup(dest->equinox) ;
    if (dest->epoch)   dest->epoch   = strdup(dest->epoch) ;
    dest->oid = cad_oid();	/* added V5.991 */

    /* Allocate the Columns */
    pc = dest->afmt = (COLINFO *)malloc(nb = dest->nfmt*sizeof(COLINFO)) ;
    memcpy(pc, src->afmt, nb) ;
    for (pe = pc + src->nfmt ; pc < pe; pc++) {
    	if ((p = pc->lims))  {
	    if (pc->elims & F_ATRT) nb = 256;
	    else /*if (pc->elims & F_DBLE) nb = 4*sizeof(double);
	    else */ nb = 4*sizeof(double);
	    pc->lims = malloc(nb);
	    memcpy(pc->lims, p, nb) ;
	}
    	if ((p = pc->value)) {
	    pc->value = malloc(nb = pc->len+1);
	    memcpy(pc->value, p, nb) ;
	}
    	if ((p = pc->range)) {
	    pc->range = malloc(nb = 2*sizeof(int8));
	    memcpy(pc->range, p, nb) ;
	}
    	if (pc->more)  		pc->more = strdup(pc->more) ;
    	if (pc->comment) 	pc->comment = strdup(pc->comment) ;
	if (pc->flags&F_Atext) 	pc->text = strdup(pc->text);
	if (pc->flags&F_Aunit) 	pc->unit = strdup(pc->unit);
	if (pc->flags&F_Aname) 	pc->label= strdup(pc->label);
	pc->oid = cad_oid();	/* added V5.991 */
    }
}

int tbf_free(T)
/*++++++++++++++++
.PURPOSE  Free an allocated TABINFO structure
.RETURNS  0 (OK) / -1 (error)
-----------------*/
  TABINFO *T;	/* IN: Format to free */
{
  TABMAP  *pm, *M, *m;
  COLINFO *pc;
  char *signature, *sigparam ;
  int ic;

    if (!T)	return(-1);	/* Already free */

#if DEBUG
    osmtrace(1);
#endif

    if (!T->signature) {
        fprintf(stderr, "****tbf_free: can't free unsigned object 0x%p\n", T) ;
	return(-1) ;
    }

    if (T->signature[1] == 'M') { 	/* TABMAP Strucutre */
	/* Only 2 independent pieces of memory assigned in TABMAP:
	   the TABMAP + fields, and prefix
	*/
        M = (TABMAP *)T ;
	T = M->tabform ;
	pm = m = T->mapping ;
	if (pm == M) T->mapping = M->next ;
	else if (pm) {
	    m = pm->next;
	    while (m && (m != M))  pm=m, m=m->next ;
	    if (m) pm->next = M->next ;
	}
	if (!m) {
	    fprintf(stderr, "****tbf_free: bad mapping\n") ;
	    return(-1) ;
	}
	m->signature = (char *)0 ;
	if (m->prefix) free(m->prefix), m->prefix = (char *)0 ;
	free(m);
	return(0) ;
    }

    signature = cad_signature("\3TAB") ;
    if (T->signature != signature) {
        fprintf(stderr,
	    "****tbf_free: bad signature @%s for object  0x%p\n", 
	    T->signature, T);
	return(-1) ;
    }

    /* Free first the COLINFO */
    signature = cad_signature("\3COL") ;
    sigparam  = cad_signature("\3PAR") ;
    for (pc=T->afmt, ic=T->nfmt;--ic>=0; pc++) {
	if (pc->signature
	 && pc->signature != signature && pc->signature != sigparam) {
	    fprintf(stderr, "++++Non-standard '%s' column not freed\n",
		pc->label) ;
	    continue ;
	}
	if (pc->flags&F_refer) ;	/* pasted column */
	else {
    	    if (pc->lims)  free(pc->lims);
    	    if (pc->value) free(pc->value);
    	    if (pc->range) free(pc->range);
    	    if (pc->more)  free(pc->more);
    	    if (pc->comment) free(pc->comment);
	}
	if (pc->flags&F_Atext) free(pc->text);
	if (pc->flags&F_Aunit) free(pc->unit);
	if (pc->flags&F_Aname) free(pc->label);
	memset(pc, 0, sizeof(COLINFO));
    }

    /* Free the associated TABMAP's */
    pm = T->mapping;
    while (pm) {
	m = pm->next ;
	pm->signature = (char *)0 ;
        free(pm) ;
	pm = m;
    }

    T->nfmt = 0; T->afmt = (COLINFO *)0;
    free(T->afmt);
    T->signature = (char *)0 ;
    if (T->equinox) free(T->equinox), T->equinox = (char *)0;
    if (T->epoch)   free(T->epoch),   T->epoch = (char *)0;
    return(0);
}

int tbf_reset(T)
/*++++++++++++++++
.PURPOSE  Reinitialize a TABINFO
.RETURNS  0
-----------------*/
  TABINFO *T;	/* IN: Format to reset */
{
  COLINFO *pf; int i;
  int8 *pl ; double *pd;
    for (pf = T->afmt, i = T->nfmt; --i>=0; pf++) {
	if (pf->elims & F_SORT)	{	/* We must keep in memory last field */
	    bzero(pf->value, pf->len+1);
	    if (pf->elims & F_SRTD) *(pf->value) = '\177';
	}
	if (pf->range ) switch(pf->ff) {
	  case 'A': bzero(pf->range, 256);
	    break ;
	  case 'I': pl = (int8 *)pf->range;
	    pl[0] = ~NULL8; pl[1] = NULL8;
	    break ;
	  case 'E': case 'F': case 'D':
	    pd = (double *)pf->range;
	    pd[0] = 1.e40; pd[1] = -1.e40;
	}
    }
    return (0);
}

int tbf_check(T)
/*++++++++++++++++
.PURPOSE  Check various consistencies
.RETURNS  Number of warnings
-----------------*/
  TABINFO *T;	/* IN: Format to check */
{
  COLINFO *pf, *pe, *pd;
  LABEL *plab;
  char bufunit[32];
  char *p, *u;
  double fl, fe ;
  int nw, i;

#if DEBUG
    printf("....Starting checking of TABINFO %p = %s", T, T->section->header);
    inbyte = getchar();
#endif
    nw = 0;
    for (pf = T->afmt, pe = pf + T->nfmt; pf < pe; pf++) {
	/* Find out the missing notes */
    	if ((*pf->text == '*') && (!pf->note)) nw++,
    	    fprintf(stderr,
	    "++++ line %d: note (*) for '%s' not found ++++\n",
	    pf->lineno, pf->label);
    	if ((pf->elims & F_NOTE) && (!pf->note)) {
	    i = strbloc(pf->text, '(') ;
	    p = pf->text + i ;
	    if (i<0) p++ ;
    	    nw++,
	    fprintf(stderr, "++++ line %d: note '%s' not found ++++\n",
	    pf->lineno, p);
	}
	pf->elims &= ~F_NOTE;	/* Not useful any more */

	/* Check dependencies	*/
	nw += check_label(pf->label, pf->lineno) ;
	plab = lab_find(pf->label);	/* Standard label */
	if (pf->label[1] == '_') {
	    p = pf->label + 2;
	    pd = tbf_findlabel(T, p);
	    if (!plab) nw++, fprintf(stderr,
	    	"++++ line %d: unknown prefix in column label '%s' ++++\n",
	    	pf->lineno, pf->label);
	    if (!pd) nw++, fprintf(stderr,
	    	"++++ line %d: dependent column of '%s' (%s) not found ++++\n",
	    	pf->lineno, pf->label, p);
	}
	else pd = (COLINFO *)0;

	/* Check Units against assumed ones (implied by label) */
        if (plab) {
            if ((p = plab->unit)) {
            	if (*p == '=') p = pd ? pd->unit : (char *)0;
	    }

	    /* Check Compatible Units data / error units */
	    if (p && strdiff(p, pf->unit) && (pf->unit[0] != '%')) {
		u = unit_si(p, &fl) ; strcpy(bufunit, u) ;
		u = unit_si(pf->unit, &fe) ;
		if (strdiff(bufunit, u) || (fe > fl)) {
		    /* It's not an error if unitless and Relative Error,
		       or RA/LON with errors in arcsec */
		    if ((pf->label[0] == 'e') && (pf->label[1] == '_')
		      &&(!u[0]) && strstr(pf->text, "elative")) ;
		    else if ((strncmp(pf->label, "e_RA", 4) == 0)
			 && (!u[0])) ;	/* Case RA with no unit in error */
		    else nw++, fprintf(stderr,
	    	      "++++ line %d: '%s' has unit '%s' (should be '%s')\n",
		      pf->lineno, pf->label, pf->unit, p);
		}
	    }
	}

	/* Check Valid Units */
	if ((p = unitex(pf->unit, 0))) {
	    if (p[strloc(p, '?')]) nw++, fprintf(stderr,
	    	"++++ line %d: unit '%s' => %s\n", pf->lineno, pf->unit, p);
	}
    }
    return(nw);
}
