/*++++++++++++++
.IDENTIFICATION fits2a.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT
.KEYWORDS
.VERSION  1.0   24-Jul-1998
.VERSION  1.1   07-Apr-1999: Add Titles
.VERSION  1.2   03-Aug-2001: -max argument
.VERSION  1.3   30-Dec-2001: Problem with 0.123456D+8
.VERSION  1.4   06-Apr-2003: Accept BINTABLE
.VERSION  1.5   09-Jul-2004: Problem END_xx was recognized as END !
.VERSION  1.6   07-Feb-2005: More than 3 axises...
			Option -p1 to print the constant columns.
.VERSION  1.61  26-Feb-2005: Accept no argument if stdin is a tty.
.VERSION  1.62  10-Mar-2005: Know the BNULL BUNIT
.VERSION  1.63  20-Jan-2006: Accept to remove a column (format "%")
.VERSION  1.7   12-Apr-2006: Accept TABLE extension
.VERSION  1.71  20-Apr-2006: Reading one record can require several reads
.VERSION  1.72  06-May-2006: Problem when %-format specified
.VERSION  1.8   05-Jul-2006: Accept IMAGE extension
.VERSION  1.9   01-Dec-2006: Accept CDi_i in place of CDELT
				Estimate a format
.VERSION  2.0   07-Apr-2007: Option -q (query) in a range
.VERSION  2.1   15-Apr-2007: Option -axis1
.VERSION  2.11  20-Apr-2007: Bug in asci table exition
.VERSION  2.12  28-Apr-2007: Bug in finding column from number
.VERSION  2.13  10-Jun-2007: Cosmetics (-Wall); -D for a file returns 0D
.VERSION  2.2   19-Sep-2007: -axis2
.VERSION  2.3   24-Oct-2007: TDIM ; better edition of TDISP
.VERSION  2.4   09-Nov-2007: Option -meta (prepare a ReadMe file)
.VERSION  2.5   13-Jan-2008: -qs (sorted query)
.VERSION  2.6   25-Apr-2008: Better error message in case of bad FITS file;
				Accept URLs and compressed files .
.VERSION  2.61  12-Jul-2008: tdisp of 'A' fields
.VERSION  2.62  20-Oct-2008: Separate the edition (tdisp)
.VERSION  2.7   06-Sep-2009: Recognize MULTISPEC
.VERSION  2.8   31-Oct-2009: Added center of image + scale in toc
.VERSION  2.81  28-Jan-2010: Bug fixed in decimals()
.VERSION  2.82  06-Feb-2010: Ignore Multispec if npix not defined
.VERSION  2.9   18-Aug-2010: Accept a mainfits as Character data
				(example of POLLUX data)
.VERSION  2.91  02-Nov-2010: accept 64-bit ('K' integer)
.VERSION  2.92  16-Nov-2010: Pb sky_image with CD matrix
.VERSION  2.93  02-Feb-2011: take into account the input file to generate readme
.VERSION  2.94  09-Feb-2011: argument -vmax (limit array size) + MATRIX
		http://iram.fr/IRAMFR/GILDAS/doc/html/class-html/node81.html
.VERSION  2.95  17-Mar-2011: check_Cform (verify TDISP is compatible with TFORM)
.VERSION  2.96  28-Jun-2011: Added edition other axes
.VERSION  2.97  10-Aug-2011: Pb vectors in -tdisp option
.VERSION  2.98  19-Oct-2011: Pb image with WAT which are not multispec !
.VERSION  2.99  18-Nov-2011: accept single-byte header with '\n'
.VERSION  3.0   30-Jan-2012: accept VOTable FitsPlusTableWriter
.VERSION  3.01  27-Mar-2012: bug in title for -axisx=...
.VERSION  3.02  03-May-2012: compute from CD_matrix
.VERSION  3.1   07-Aug-2012: PCOUNT / GCOUNT
.VERSION  3.11  09-Aug-2012: Accept RA-- and DEC- in any axis
.VERSION  3.12  19-Sep-2012: Accept GLON- and GLAT- in axes
.VERSION  3.13  16-Oct-2012: Fixed bug
.VERSION  3.14  13-Nov-2012: Fixed bug
.VERSION  3.2   10-Jan-2013: Table with TTMIN/TTMAX (truncate min/max)
.VERSION  3.3   10-Jun-2013: 
		* List the ranges in toc() (e.g. spectra)
		* -par -axis1: lists all axes
		* -axis: can used 1/3 (list fits 1, then 3)
.VERSION  3.31  18-Jun-2013: bug for format %s or %10s
.VERSION  3.32  27-Jun-2013: big in bounded_K
.VERSION  3.33  01-Jul-2013: bug in -axis.=...
.VERSION  3.34  05-Aug-2013: Option -par in FITS table

.TODO: choice of columns from name (does not work)

.TODO: Option -l [list_of_lines|-] to list only specific lines (e.g. 1,10-5)
                                   or lines to list in standard input.
       with use of .gzi file if existing (or direct seek if file not compressed)
       (then option -skip n is equivalent to -l (n+1)- )

.COMMENTS       Translate a FITS file into ascii
	Returned code:
	0 = OK
	1 = Error
	2 = max attained
---------------*/

/*#define DEBUG */

#define VERSION "3.34 (05-Aug-2013)"

#include <filutil.h>
#include <strut.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef SOLARIS
#include <ieeefp.h>
#define isFinite(x) finite(x)
#else	/* Some compilers use finite, other use isfinite ... */
#ifdef  isfinite
#  define isFinite(x) isfinite(x)
#else
#  define isFinite(x) finite(x)
#endif
/* #define isnand(x) isnan(x) */
/* #define isnanf(x) isnan(x) */
#endif

#ifndef O_BINARY
#define O_BINARY 0
#endif

#ifndef int4
#define int4 int
#define int8 long long
#endif

#ifndef int8l	/* Format for 64-bit integer: 1=%d 2=%ll */
#define int8l  2
#endif

#ifndef MAX_AXIS	/* Can't work if NAXIS > this limit 	*/
#define MAX_AXIS  9
#endif
#ifndef MAX_QUERIES
#define MAX_QUERIES 99
#endif
static int bitpix, naxis, len1el, pcount, gcount;
static double bscale, bzp ;
static char *a2880[400];	/* Last FITS header read...	*/
static int  na2880 = 0;		/* How many a2880 were read...	*/
static char *fmt[1000];
static char sfmt[1000];		/* Status of format: 1 is given */
static int  nfmt=0;
static int  dim[MAX_AXIS] ;
static int dimm[MAX_AXIS] ;	/* Multiplicators for address 	*/
static int mdim[MAX_AXIS] ;	/* MAXIS (CLASS/Gildas)	 	*/
static int  xtno = -1;		/* Current   Extension number 	*/
static char *axten;		/* Requested -xten parameter	*/
static char *zindex;		/* Index generated by gzindex   */
static int  zpos;		/* Current position in zindex   */
static char xbyname;		/* Required xtension is a name  */
static int recno;		/* Record number (bintable)	*/
static int total_lines;		/* bintable asctable mainfits	*/
static double xaxis[MAX_AXIS] ;	/* Current   value along axis	*/
static double delta[MAX_AXIS] ;	/* Increment value along axis	*/
static double crval[MAX_AXIS] ;	/* Initial   value along axis	*/
static char   *type[MAX_AXIS] ;
static char   *unit[MAX_AXIS] ;
static char *input_name ;	/* Name of file being processed	*/
static int namelen=0;		/* Length of extension name     */
static int  blocno = 0;		/* Current position in file     */
static int  blocno_data;	/* Current position in file     */
static char *delim = " " ;	/* Delimiter bewteem cols	*/
static char *aheader;		/* Additional header		*/
static char *acols;		/* List of columns to output	*/
static char *axis_choice[MAX_AXIS];
/*#define axis1 axis_choice[0]	-- Selection along axis1	*/
/*#define axis2 axis_choice[1]	-- Selection along axis2	*/
static char *query[MAX_QUERIES];/* Query on columns		*/
static int  nquery = 0;
static char verbop ;		/* Verbose Option		*/
static char ohead ;		/* Option to write the Headers  */
static char oqs;		/* -qs: Sorted query		*/
static char opar ;		/* Option to write in parfile	*/
static char ometa ;		/* Option to write ReadMe templ */
static char odata = 1 ;		/* Option to write the Data  	*/
static char otoc  = 0 ;		/* Option -toc=1 -TOC=3		*/
static char op1   ;		/* Option to write constant cols*/
static char optb  ;		/* Option -b to remove blanks   */
static char oDim;		/* Option -D to issue just dim  */
static char tdisp ;		/* Option to evaluate formats   */

#define MAX(x,y)	((x) >= (y) ? (x) : (y))
#define MIN(x,y)	((x) <= (y) ? (x) : (y))
#define ITEMS(a)	sizeof(a)/sizeof(a[0])
#define starts_by(s,t)	(strncmp(s,t,sizeof(t)-1)==0)

typedef struct {
    char *name ;
    char *value ;
    char *comment ;
} FITSKEY ;
#define iTFIELDS	4	/* INDEX of TFIELDS  */
#define iXTENSION       5	/* INDEX of XTENSION */
#define iMAXIS          6	/* INDEX of MAXIS    */
#define iNULL           7	/* INDEX of BNULL    */
#define iEXTNAME       10
#define iUNIT           9 	/* INDEX of BUNIT    */
#define perAXIS	        6	/* #items per axis   */
#define iPCOUNT	       11	/* INDEX of NAXIS1   */
#define iGCOUNT	       12	/* INDEX of NAXIS1   */
#define iNAXIS1	       13	/* INDEX of NAXIS1   */
#define iNAXIS2	iNAXIS1+perAXIS	/* INDEX of NAXIS2   */
#define iMAXDIM 9*perAXIS+iNAXIS1  /* Upmost index of dimensions */
#define iCD     iMAXDIM+0	/* INDEX of CDi_j    */
#define iCD2        iCD+9     	/* 1,2 + 2,1 */
#define iCD3        iCD2+2     	/* 1,3 2,3 3,1 3,2   */
#define iCD4        iCD3+4     	/* 1,4 2,4 3,4 +rev  */
#define iMAXIS1     iCD4+6     	/* MAXIS1 */
#define iWAT	iMAXIS1+9     	/* Multispec (World Coo Sys) */
static FITSKEY keys[] = 	/* List of Keywords */	{
    { "iBITPIX  =", (char *)0} ,      /*  0 */
    { "iNAXIS   =", (char *)0} ,      /*  1 */
    { "dBSCALE  =", (char *)0} ,      /*  2 */
    { "dBZERO   =", (char *)0} ,      /*  3 */
    { "iTFIELDS =", (char *)0} ,      /*  4 */
    { "cXTENSION=", (char *)0} ,      /*  5 */
    { "iMAXIS   =", (char *)0} ,      /*  6 */
    { "dBNULL   =", (char *)0} ,      /*  7 = NULL value */
    { "dBLANK   =", (char *)0} ,      /*  8 = NULL value */
    { "cBUNIT   =", (char *)0} ,      /*  9 */
    { "cEXTNAME =", (char *)0} ,      /* 10 */
    { "iPCOUNT  =", (char *)0} ,      /* 11 */
    { "iGCOUNT  =", (char *)0} ,      /* 12 */
    { "iNAXIS1  =", (char *)0} ,      /* iNAXIS1+0 */
    { "dCRPIX1  =", (char *)0} ,      /* iNAXIS1+1 */
    { "dCRVAL1  =", (char *)0} ,      /* iNAXIS1+2 */
    { "dCDELT1  =", (char *)0} ,      /* iNAXIS1+3 */
    { "cCTYPE1  =", (char *)0} ,      /* iNAXIS1+4 */
    { "cCUNIT1  =", (char *)0} ,      /* iNAXIS1+5 */
    { "iNAXIS2  =", (char *)0} ,
    { "dCRPIX2  =", (char *)0} ,
    { "dCRVAL2  =", (char *)0} ,
    { "dCDELT2  =", (char *)0} ,
    { "cCTYPE2  =", (char *)0} ,
    { "cCUNIT2  =", (char *)0} ,
    { "iNAXIS3  =", (char *)0} ,
    { "dCRPIX3  =", (char *)0} ,
    { "dCRVAL3  =", (char *)0} ,
    { "dCDELT3  =", (char *)0} ,
    { "cCTYPE3  =", (char *)0} ,
    { "cCUNIT3  =", (char *)0} ,
    { "iNAXIS4  =", (char *)0} ,
    { "dCRPIX4  =", (char *)0} ,
    { "dCRVAL4  =", (char *)0} ,
    { "dCDELT4  =", (char *)0} ,
    { "cCTYPE4  =", (char *)0} ,
    { "cCUNIT4  =", (char *)0} ,
    { "iNAXIS5  =", (char *)0} ,
    { "dCRPIX5  =", (char *)0} ,
    { "dCRVAL5  =", (char *)0} ,
    { "dCDELT5  =", (char *)0} ,
    { "cCTYPE5  =", (char *)0} ,
    { "cCUNIT5  =", (char *)0} ,
    { "iNAXIS6  =", (char *)0} ,
    { "dCRPIX6  =", (char *)0} ,
    { "dCRVAL6  =", (char *)0} ,
    { "dCDELT6  =", (char *)0} ,
    { "cCTYPE6  =", (char *)0} ,
    { "cCUNIT6  =", (char *)0} ,
    { "iNAXIS7  =", (char *)0} ,
    { "dCRPIX7  =", (char *)0} ,
    { "dCRVAL7  =", (char *)0} ,
    { "dCDELT7  =", (char *)0} ,
    { "cCTYPE7  =", (char *)0} ,
    { "cCUNIT7  =", (char *)0} ,
    { "iNAXIS8  =", (char *)0} ,
    { "dCRPIX8  =", (char *)0} ,
    { "dCRVAL8  =", (char *)0} ,
    { "dCDELT8  =", (char *)0} ,
    { "cCTYPE8  =", (char *)0} ,
    { "cCUNIT8  =", (char *)0} ,
    { "iNAXIS9  =", (char *)0} ,
    { "dCRPIX9  =", (char *)0} ,
    { "dCRVAL9  =", (char *)0} ,
    { "dCDELT9  =", (char *)0} ,
    { "cCTYPE9  =", (char *)0} ,
    { "cCUNIT9  =", (char *)0} ,
    { "dCD1_1   =", (char *)0} ,	/* iCD+0 */
    { "dCD2_2   =", (char *)0} ,	/* iCD+1 */
    { "dCD3_3   =", (char *)0} ,	/* iCD+2 */
    { "dCD4_4   =", (char *)0} ,	/* iCD+3 */
    { "dCD5_5   =", (char *)0} ,	/* iCD+4 */
    { "dCD6_6   =", (char *)0} ,	/* iCD+5 */
    { "dCD7_7   =", (char *)0} ,	/* iCD+6 */
    { "dCD8_8   =", (char *)0} ,	/* iCD+7 */
    { "dCD9_9   =", (char *)0} ,	/* iCD+8 */
    { "dCD1_2   =", (char *)0} ,	/* iCD2+0 */
    { "dCD2_1   =", (char *)0} ,	/* iCD2+1 */
    { "dCD1_3   =", (char *)0} ,	/* iCD3+0 */
    { "dCD2_3   =", (char *)0} ,	/* iCD3+1 */
    { "dCD3_1   =", (char *)0} ,	/* iCD3+2 */
    { "dCD3_2   =", (char *)0} ,	/* iCD3+3 */
    { "dCD1_4   =", (char *)0} ,	/* iCD4+0 */
    { "dCD2_4   =", (char *)0} ,	/* iCD4+1 */
    { "dCD3_4   =", (char *)0} ,	/* iCD4+2 */
    { "dCD4_1   =", (char *)0} ,	/* iCD4+3 */
    { "dCD4_2   =", (char *)0} ,	/* iCD4+4 */
    { "dCD4_3   =", (char *)0} ,	/* iCD4+5 */
    { "iMAXIS1  =", (char *)0} ,
    { "iMAXIS2  =", (char *)0} ,
    { "iMAXIS3  =", (char *)0} ,
    { "iMAXIS4  =", (char *)0} ,
    { "iMAXIS5  =", (char *)0} ,
    { "iMAXIS6  =", (char *)0} ,
    { "iMAXIS7  =", (char *)0} ,
    { "iMAXIS8  =", (char *)0} ,
    { "iMAXIS9  =", (char *)0} ,
    { "mWAT0_001=", (char *)0} ,	/* iWAT+0 = World Coord. Multispec */
    { "mWAT1_001=", (char *)0} ,	/* iWAT+1 = World Coord. Multispec */
    { "mWAT2_001=", (char *)0} ,	/* iWAT+2 = World Coord. Multispec */
    { (char *)0 }
 } ;
#define XTENSION  keys[iXTENSION].value 	/* Current extension name */
#define TFIELDS  *(int *)(keys[iTFIELDS].value)
#define EXTNAME  keys[iEXTNAME].value 		/* Current extension name */

typedef struct {
    short no;	/* Col number   */
    short ano;	/* Order in -a fil */
    char form[12];    /* TFORM	*/
    int len1;	/* Length 1 el.	*/
    int len1a;  /* Length 1 el., edited */
    int count;	/* Count factor */
    int opos;	/* Position in row */
    int cnul;   /* Count of NULL values */
    int trunc[2]; /* Count of truncated values min/max */
    int flags;
#define FLAG_CONVERT		1	/* Ascii table: re-edit number */
#define FLAG_CONSTANT		2	/* Column is blank or constant */
#define FLAG_TRUNC_MIN		4	/* Column has a TTMIN value    */
#define FLAG_TRUNC_MAX		8	/* Column has a TTMAX value    */
#define FLAG_EXPONENT	     0x10	/* Number has an exponent...   */
#define FLAG_TRUNC_VMAX      0x80	/* Vector truncated (-vmax)    */
#define FLAG_SELECT_MIN     0x100
#define FLAG_SELECT_MAX     0x400
#define FLAG_SELECT_NOT    0x8000
#define FLAG_SELECT_float  0x1000	/* Bits concerned by selection */
#define FLAG_SELECT_char   0x4000	/* Bits concerned by selection */
#define FLAG_SELECT	   0xff00	/* Bits concerned by selection */
    char disp[16];     /* TDISP	*/
    char *type;	/* TTYPE (name)	*/
    char *unit;	/* TUNIT (unit)	*/
    char *null; /* TNULL	*/
    char *comment;
    char *value;
    char *tdim; /* TDIM 	*/
    char Cfmt[16];     /* TDISP	*/
    double min, max, nul, tmin, tmax;	/*  May be int4, int8, etc... */
    int cdec[16];	/* Number of decimals in floating-point number */
} FITSROW;
static FITSROW *fitsrow;	/* Allocated from TFIELDs   */
static FITSROW **arows;		/* The asked list of rows   */
static int      allorow;	/* Number of allocated rows */
static int colorder;		/* Column order number -col */
static FITSROW *tdelta[MAX_AXIS]; /* FIELD with increment (CLASS/Gildas) */
static FITSROW *tcrval[MAX_AXIS]; /* FIELD with reference (CLASS/Gildas) */
static char    *tparam[MAX_AXIS]; /* FIELD actual name    (CLASS/Gildas) */

/* The MULTISPEC stores one spectrum par line of 2-D image.
 * Each line i described by WAT2_xxx keywords e.g. wtype=multispec
 * spec1 = "3 65 0 11831.182617188 0.17006626186314 1002 0. 284.14 290.14"
 * Each spec# describes one line of the spectrum:
 * #1 = aperture# (int)
 * #2 = aperture# (int)
 * #3 = dtype (int)  0=linear 1=log-linear 2=non-linear
 * #4 = lambda1 (wavelength of pix#1)
 * #5 = dlambda (step in lambda)
 * #6 = Number of valid pixels
 * #7 = z (redshift)
 * #8 = aperture lower limit
 * #9 = aperture upper limit
 * See MULTISPEC at http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.46.2794&rep=rep1&type=ps
 */
static char *multiscale[] = { "linear", "log-linear", "non-linear" };
static char *multifunct[] = { "???", "chebyshev", "legendre",
    "cubic-spline", "linear-spline", "pixel-coordinate-array",
    "sampled-coordinate-array" };
typedef struct {
    int ftype;
    int order;
    double weight, olam;
    double pmin, pmax;
    double coeff[2];		/* Number of coeff = order */
} MULTIFUNCTION;
typedef struct {		/* MULTISPEC definition */
    int ap, beam, dtype; 	/* dtype=0 => linear */
    int npix;			/* In principle = NAXIS1 */
    double lam, dlam;		/* Wavelength + dWabelength */
    double z;			/* Generally 0 */
    double aplim[2];		/* Intersting ?? */
    MULTIFUNCTION *function;
} MULTISPEC;
static MULTISPEC *aspec;
static int nspec;		/* Number of allocated MULTISPEC's */

static char usage[] = "\
Usage: fits2a [-D|-toc|-TOC] [-a additional_header] [-b] [-dc] [-meta] [-head]\n\
       [-nodata|-tdisp] [-max lines] [-skip lines] [-top lines]\n\
       [-axis{1|2|...}[=list]] [-c {list_of_cols|=a}] [-p1] [-par]\n\
       [-vmax elements] [-q[s] col=min,max] [-xten list] \n\
       [n][%fmt] [%fmt]... [file[#xten]...]\n";
static int max_lines = 0x7fffffff ;
static int top_lines = 0x7fffffff ;
static int max_recno = 0x7fffffff ;
static int vmax      = 0;
static int nskip     = 0;
static char ruler[] = "\
#------------------------------------------------------------------------------\
-";

static char help[] = "\
This program converts a FITS array or binary table into an ASCII table.\n\
It can for instance convert a spectrum (1-D) into a 2-column table for a\n\
spreadsheet or for plotting with gnuplot-like utilities. Possible options:\n\
      -D : just issue the dimension, as 1D nnnn, 2D dim1xdim2, etc\n\
    -toc : issue a table of contents of the FITS file contents\n\
    -TOC : In addition to -toc, display the FITS headers\n\
 -a file : file containing additional FITS header metadata (example below)\n\
      -b : remove the useless blanks (end of strings, beginning of numbers)\n\
     -dc : define c as delimiter between columns (defaut: 1 blank; tab if -d)\n\
   -meta : print out column definitions (in catalog 'ReadMe' standard)\n\
  -nodat : do not print out the data\n\
  -tdisp : evaluate the TDISP formats from the data contents (tables only);\n\
           with the -v option, adds the min/max/NULs to each field.\n\
   -head : print column headers (default: only the data are printed)\n\
   -max  : stop after the specified number of lines to read, writes message\n\
   -skip : number of  input lines to skip before editing\n\
   -top  : number of output lines to edit (max. output lines)\n\
  -axis1 : edit as as set of vectors (1 line for values along axis1,\n\
           eventually restricted to specific components)                  (*)\n\
  -axis2 : edit as a set of vectors along axis2, eventually restricted\n\
 -c list : column selection as numbers (e.g. 1-9,12 selects cols 1,2..9,12)\n\
           or blank-separated TTYPExx names (e.g. \"RA Dec Flux\");\n\
              (pattern-matching when starts by * or -* e.g. \"-* r_*\")\n\
           =a restricts the output to the columns quoted in additional_file.\n\
     -p1 : print the constant columns (those with AXISx=1). By default the\n\
           constant columns are not printed, or printed as comments (#)\n\
    -par : print the output in paragraph format (1 line for each column);\n\
           associated with -axis, use the compoent along the axis as 'title'\n\
      -q : query, i.e. select rows by setting limits in a numeric column\n\
     -qs : as -q, but the file is sorted against this column\n\
   -vmax : maximal number of elements in vectors to print out (0 to get all)\n\
   -xten : choose the FITS extension selection (main FITS header is number 0)\n\
           e.g. 2,8-12. If not numeric, represents the EXTNAME\n\
  [n]%ft : C-format, to use for editing successive output columns, eventually\n\
           preceded by a repetition factor; \"%.\" stands for default edition\n\
           and \"%\" alone means that the corresponding column is NOT edited.\n\
    file : FITS input file; may be compressed, or a file delivered by HTTP;\n\
           specific extensions may also be chosen (with # or |)\n\
NULL values (NaN or specified via FITS NULL keywords) are edited as dashes\n\
\n\
(*) The choice of components of a vector may be specified with comma-dashes:\n\
    -axis1=1,3,5-8  means that the vector is made of components #1,3,5,6,7,8;\n\
    slashes indicate edition as separated tables:\n\
    -axis1=1,3/5,7  means edition of a vector made of components #1 and #3,\n\
                    followed by a blank line, and then components #5 and #7;\n\
    a percent (%) asks to edit 1-column tables (2-column with -par option):\n\
    -axis1=1%5  edit the values of element#1, a blank line, values of el.#2,\n\
                a blank line, ...etc... values of el#5\n\
    Associated to the -par option, values along axis are added as left column.\n\
\n\
Example of an additional_file which changes in a binary table the column\n\
names and formats of columns 2 and 3 (for details about FITS conventions, see\n\
http://fits.gsfc.nasa.gov/fits_documentation.html )\n\
TTYPE2 = 'RA(J2000)'   / Name of column#2\n\
TTYPE3 = 'Dec(J2000)'  / Name of column#3\n\
TDISP2 = 'F8.4'        / Fortran-like format. Note that C-like format can be\n\
TDISP3 = 'F8.4'        / used if starting by %, e.g. '%+08.4f' for a latitude\n\
TTMIN3 = '-999.999'     / asking for truncation lower limit\n\
TTMAX3 = ' 999.999'     / asking for truncation upper limit\n\
" ;
static int swapping ;   /* Set to 1 or 2 (swap) */
static double dexp[] = { 1., 1.e1, 1.e2, 1.e3, 1.e4, 1.e5, 1.e6,
  1.e7, 1.e8, 1.e9, 1.e10, 1.e11, 1.e12, 1.e13, 1.e14, 1.e15 };
static char sep_disp[] = "#..........................";
double atof(), fabs(), floor();
typedef int (*SORT_FCT)(/* const void *, const void * */) ;

#ifdef DEBUG
static int naloc = 0;
static void *aloc[1000];
void Free(void *p) {
  int i, n0;
    fprintf(stderr, "....Free..... %p", p);
    for (i=naloc, n0=0; --i>=0; ) {
	if (!aloc[i]) { n0++; continue; }	/* NULL address */
	if (p != aloc[i]) continue;
	fprintf(stderr, " #%03d (ok)", i);
	aloc[i] = (void *)0;
	break;
    }
    if (i<0) fprintf(stderr, " ****not allocated?");
    /* Count number of allocated pieces */
    while (i-->=0) if(!aloc[i]) n0++;
    fprintf(stderr, " [%d]\n", naloc-n0);
}
void *Malloc(int n) {
  void *p; int i;
    fprintf(stderr, "....Malloc");
    fflush(stderr);
    p = malloc(n);
    fprintf(stderr, " => %p", p);
    /* Save the allocated address */
    for (i=0; (i<ITEMS(aloc)) && aloc[i]; i++) ;
    if (i>=ITEMS(aloc))
	fprintf(stderr, " ****Too many (>=%d) allocations****", ITEMS(aloc));
    else {
	fprintf(stderr, " #%03d", i);
	aloc[i] = p;
	if (i == naloc) naloc++;
    }
    fprintf(stderr, " (%d bytes)\n", n);
    memset(p, 0, n);
    return(p);
}
char *Strdup(char *s) {
  char *p;
    p = Malloc(strlen(s)+1);
    strcpy(p, s);
    return(p);
}
#else
#define Malloc(n)	malloc(n)
#define Free(p)		free(p)
#define Strdup(p)	strdup(p)
#endif

/*==================================================================
		Get floating-point numbers
		(problem, atof doesn't recognize 0.12345D+08)
 *==================================================================*/

static char *ltrim(char *text)
/*++++++++++++++++
.PURPOSE  Skip the left blanks
.RETURNS  non-blank address
-----------------*/
{
  char *p;
    for (p=text; *p == ' '; p++) ;
    return(p);
}

static double a2f(char *text)
/*++++++++++++++++
.PURPOSE  Interpret a string for a double
.RETURNS  The value
-----------------*/
{
  double val, atof() ;
  char *p;
    /* Examine the number, until a non-valid char found */
    if ((p = strchr(text, 'D'))) *p = 'e' ;
    val = atof(text) ;
    if (p) *p = 'D' ;
    return(val) ;
}

static void f4fits(char *text, int datatype)
/*++++++++++++++++
.PURPOSE  Rewrite a string representing a number
.RETURNS  ---
.REMARKS  datatype may be 'D', 'F', etc
-----------------*/
{
  char *p;
    p = strchr(text, 'e'); if (!p) return;
    *p = datatype; 
    if (datatype != 'D') {
	*(p++) = 'E';
	if (*p == '+') p++;
	if (atoi(p) == 0) { /* Remove the exponent, useless */
	    for (*(p--) = 0; *p != 'E'; *(p--) = 0) ;
	    if (*p == 'E') *p = 0;
	}
    }
}

static double pow8(int bytes)
/*++++++++++++++++
.PURPOSE  Compute 8^bytes
.RETURNS  The value
-----------------*/
{
  double val;
    val = 1;
    while(bytes>0) {
	val *= 256. ;
	bytes--;
    }
    return(val);
}

/*==================================================================*/
#if DEBUG
static int check_unit(FITSROW **arows, int num)
{
    int nf=0;
    FITSROW *pr, **ar;
    for (ar=arows; *ar; ar++) {
	++nf; pr = *ar;
	if (pr->unit == (char *)1) { int c;
	    printf("#***row#%d[%p] unit=%p -- nul=%d",nf,pr,pr->unit, num);
	    c = getchar();
	}
    }
}
#endif
#if 0
int strloc(char *text, int c)
/*++++++++++++++++
.PURPOSE  Locate specified character
.RETURNS  Index of located char
-----------------*/
{
  char *s;
    for ( s = text; *s; s++)      if (*s == c)    break;
    return(s-text);
}
#endif

void ttrim(char *text)
/*++++++++++++++++
.PURPOSE  Remove trailing blanks
.RETURNS  ---
-----------------*/
{
  char *p;
    for (p=text+strlen(text)-1; (p>=text) && isspace(*p); p--) ;
    *++p = 0;
}

static int skip_num(char *text)
/*++++++++++++++++
.PURPOSE  skip the number (length of number + while space)
.RETURNS  ---
-----------------*/
{
  char *p = text;
    while (*p == ' ') p++;
    if ((*p == '-') || (*p == '+')) p++;
    while (isdigit(*p)) p++;
    if (*p == '.') for (++p; isdigit(*p); p++) ;
    if (isalpha(*p)) { int c = tolower(*p);
	if ((c=='e')||(c=='d')) p++;
	if ((*p == '-') || (*p == '+')) p++;
	while (isdigit(*p)) p++;
    }
    while (isspace(*p)) p++;
    return((p-text));
}

static int *get_range(char *str, int lower_limit, int upper_limit)
/*++++++++++++++++
.PURPOSE  Interpret numbers like 1,2,3-6,..a
.RETURNS  A new piece of memory with the list of numbers, preceded by its count
.REMARKS  On return  [0] = number of elements
	V3.3: there can be a continuation! The separator can be:
	  , / - %
	  / ==> insert a new element
	  % ==> insret a set of new elements
-----------------*/
{
  char *s;
  int *range, *v, n, add_sep, lower_range=lower_limit;
  int i = 1;

    /* Compute the number of values */
    s = str; n = 0;
    while (*s) {
	lower_range = add_sep = 0;
	if (*s == ',') s++;
	else if (*s == '-') {	/* A range */
	    lower_range = i;
	    s++;
	}
	else if (*s == '/') {	/* Insert a new line */
	    s++; n+=2;		/* Must add the length */
	    continue;
	}
	else if (*s == '%') { /* All elements as new records */
	    lower_range = i;
	    s++;
	    add_sep = 1;
	}
	if (*s)  {
	    if (!isdigit(*s)) {
	        fprintf(stderr, "++++Invalid number: %s\n", s);
	        return((int *)0);
	    }
	    i = atoi(s);
	}
	else i = upper_limit;
	if (i< lower_range) {
	    fprintf(stderr, "++++Invalid value '%s' in: %s\n", s, str);
	    return((int *)0);
	}
	n++;
	while (isdigit(*s)) s++;
	if (lower_range) n += (add_sep+1)*(i+1-lower_range);
    }

    range = v = Malloc((n+2)*sizeof(int));
    s = str; n = 0;
    while (*s) {
	lower_range = add_sep = 0;
	if (*s == ',') s++;
	else if (*s == '-') {
	    lower_range = i;
	    s++;
	}
	else if (*s == '/') {	/* New line... */
	    s++;
	    v[0] = n;
	    v += n+1; n=0;
	    continue;
	}
	else if (*s == '%') {
	    lower_range = i;
	    s++;
	    add_sep = 1;
	}
	i = *s ? atoi(s) : upper_limit;
	/* V3.33: the lower bound was countde twice... */
	if (i > upper_limit) {
	    fprintf(stderr, "++++Too large value %d (>'%d) in: %s\n",
		    i, upper_limit, str);
	    i = upper_limit;
	}
	if (lower_range) for (++lower_range; lower_range <= i; lower_range++) {
	    if (add_sep) {
	        if (n) { v[0] = n; v += n+1; n=0; }
	        v[n] = 1;
	    }
	    v[++n] = lower_range;
	}
	else v[++n ] = i;
	v[0] = n;
	while (isdigit(*s)) s++;
    }
    v[++n] = 0;
    return(range);
}

#if 0
static char *load_file(char *name)
/*++++++++++++++++
.PURPOSE  Load a file completely into memory
.RETURNS  The loaded text terminated by NULL / NULL if error
-----------------*/
{
  int  i;
  int4 count;
  char *buffer;

        /* Load file into Memory */
    i = open(name, O_BINARY);
    if (i <= 0) {
    	fprintf(stderr, "****");
        perror(name);
        return ((char *)0);
    }
    count = lseek(i, 0L, 2); lseek(i, 0L, 0);

    if ((buffer = Malloc(count+1))) {
        if (read(i, buffer, count) != count) {
   	    fprintf(stderr, "****");
            perror(name);
            Free(buffer);
            return ((char *)0);
        }
        buffer[count] = 0;
    }
    close(i);
    return(buffer);
}
#endif

static int isregfile(FILE *file)
/*++++++++++++++++
.PURPOSE  tests a regular file (can seek)
.RETURNS  0 = stream (no seek)
-----------------*/
{
  struct stat buf;
    if (fstat(fileno(file), &buf) < 0) {
	perror("****");
	return(0);
    }
    return(S_ISREG(buf.st_mode));
}

/*==================================================================
		Estimate the best formats for an edition
 *==================================================================*/
static int ndec(double value)
/*++++++++++++++++
.PURPOSE  Estimate the number of decimals to edit a number
.RETURNS  Number of decimals for a 'fair' edition; 0 for impossible (0, NaN..)
-----------------*/
{
  int nd;
    if ((value != 0) && isFinite(value)) {
        if(value<0) value = -value;
        for (nd=0; value<0.99; nd++) value *= 10;
	return(nd);
    }
    return(0);
}

static int ndig(double value)
/*++++++++++++++++
.PURPOSE  Estimate the number of digits for the integer part if a number
.RETURNS  0 for NaN/inf
-----------------*/
{
  int nd;
    if ((value != 0) && isFinite(value)) {
        if(value<0) value = -value;
        for (nd=0; value>0.99; nd++) value /= 10;
	return(nd);
    }
    return(0);
}

static char *guess_fmt(double val0, double val1, double delta, char *name)
/*++++++++++++++++
.PURPOSE  Estimate a format for an exis from min / max / delta
.RETURNS  The C-format (in a static piece)
-----------------*/
{
  static char fmt[20] = "%";
  int nd, ne;
    nd = ndig(val0);		/* Digits for Minimum */
    ne = ndig(val1);		/* Digits for Maximum */
    if (nd>ne) ne = nd;		/* Digits for integer part */
    if (ne < 1) ne = 1;		/* Integer part: at least one byte    */
    nd = ndec(delta);		/* Decimals required (from increment) */
    if (delta == 0) {
	if (!name) return("%g");
	if (starts_by(name, "RA-"))  return("%010.6f");
	if (starts_by(name, "DEC"))  return("%+010.6f");
	if (starts_by(name, "GLON")) return("%010.6f");
	if (starts_by(name, "GLAT")) return("%+010.6f");
        return("%g");
    }
    if ((nd+ne) < 18) {		/* Better to use a 'f' format */
	if ((nd+ne)<2) ne++;
	sprintf(fmt+1, "%d.%df", nd+ne+2, nd);
        return(fmt);
    }
    return("%20.12e");
}

/*==================================================================
		Swapping problems...
 *==================================================================*/
static void which_swapping()
/*++++++++++++++++
.PURPOSE  Verify which Swapping
.RETURNS  ---
-----------------*/
{
  static int value ;
  char *v ;
    value = 0x010203 ;
    v = (char *)(&value) ;
    v = (char *)(&value) ;
    if ((v[0] == 0) && (v[1] == 1) && (v[2] == 2))
         swapping = 0 ;  /* No swap necessary */
    else if ((v[0] == 1) && (v[1] == 0) && (v[2] == 3))
    	swapping = 1 ;  /* Half-word swap */
    else if ((v[0] == 3) && (v[1] == 2) && (v[2] == 1))
    	swapping = 2 ;  /* Full-word swap */
    else {
      	fprintf(stderr, "****Irrationnal Byte Swap %02x%02x%02X%02x\n",
	    v[0]&0xff, v[1]&0xff, v[2]&0xff, v[3]&0xff) ;
	exit(1) ;
    }
}

static void swap1 (short *array, int nshort)
/*++++++++++++++++
.PURPOSE  Swap the bytes in the array of shorts if necessary
.RETURNS  ---
.REMARKS  Useful for big-endian machines
-----------------*/
{
  register int m, n ;
  register short *p, *e ;
    for (p=array, e=p+nshort; p<e; p++) {
        n = *p ;
        m =  (n>>8) & 0xff ;
        m |= (n<<8) ;
        *p = m ;
    }
}

static void swap (int *array, int nint)
/*++++++++++++++++
.PURPOSE  Swap the bytes in the array of integers if necessary
.RETURNS  ---
.REMARKS  Useful for big-endian machines
-----------------*/
{
  char *p, *e ;
  int n ;

    p = (char *)array ; e = p + 4*nint ;
    if (swapping == 1) swap1((short *)array, 2*nint) ;
    else if (swapping == 2) while (p < e) {
        n = p[0] ; p[0] = p[3] ; p[3] = n ;
        n = p[1] ; p[1] = p[2] ; p[2] = n ;
        p += 4 ;
    }
}

static void swap8(void *array, int nint)
/*++++++++++++++++
.PURPOSE  Swap the bytes in the array of doubles if necessary
.RETURNS  ---
.REMARKS  Useful for big-endian machines
-----------------*/
{
  char *p, *e ;
  int n ;

    p = (char *)array ; e = p + 8*nint ;
    if (swapping == 1) swap1((short *)array, 4*nint) ;
    else if (swapping == 2) while (p < e) {
        n = p[0] ; p[0] = p[7] ; p[7] = n ;
        n = p[1] ; p[1] = p[6] ; p[6] = n ;
        n = p[2] ; p[2] = p[5] ; p[5] = n ;
        n = p[3] ; p[3] = p[4] ; p[4] = n ;
        p += 8 ;
    }
}

/*==================================================================
		Internal Routines
 *==================================================================*/
#if 0
static void toC(char *str)
/*++++++++++++++++
.PURPOSE  Convert a string according to C standards
.RETURNS  ---
.REMARKS  \octal or \f \b \r \n ...
-----------------*/
{
  int o; char *p ;
    for (p=str; *str; str++) {
        if (*str != '\\') { *(p++) = *str; continue; }
        switch(*++str) {
          case 'b': *(p++) = '\b' ; continue ;
          case 'f': *(p++) = '\f' ; continue ;
          case 'n': *(p++) = '\n' ; continue ;
          case 'r': *(p++) = '\r' ; continue ;
          case 't': *(p++) = '\t' ; continue ;
          case 'e': case 'E': *(p++) = '\033' ; continue ;
          default:
	    if (isdigit(*str))
	        for (o=0; isdigit(*str); str++) o = o*8 + (*str-'0');
	    else o = *(str++) ;
	    *(p++) = o ;
	    str-- ;
	}
    }
    *p = 0 ;
}
#endif

static char *key_value(char *text, int *ocomment, char *aheader)
/*++++++++++++++++
.PURPOSE  Get the value corresponding to a key (or from its alternative)
.RETURNS  The string + position of comment (if any) in ocomment
.REMARKS  Remove the trailing blanks. argument 'aheader' added in V2.51
-----------------*/
{
  static char buf[80];
  char *p, *e=(char *)0;
  int i, replaced=0;
    if (ocomment) *ocomment = 0;
    /* V2.51: is there an alternative ? */
    if (aheader) { int len=0;
	for (p=aheader; *p; p += len) {
	    if (iscntrl(*p)) { len = 1; continue; }
	    len = strloc(p, '\n');
	    if (!isalpha(*p)) continue;
	    for (i=0; i<8; i++) {
		if (text[i] == p[i]) continue;
		if ((text[i] == '=') && isspace(p[i])) i=8;
		break;
	    }
	    if (i>=8) break; 	/* Replacement found */
	}
	if (*p) { text = p; e = p+len; replaced = 1; }
    }
    if (!e) e = text + 80;
    for (p=text; (p<e) && (*p != '=') && isprint(*p); p++) ;
    if (*p != '=') return((char *)0);
    for (++p; (*p == ' ') || (*p == '\t'); p++) ;
    i = 0;
    if (*p == '\'')  {	/* Copy the Character String */
        for (++p; isprint(*p) && (p<e); p++) {
	    if (*p == '\'') {
	        if (p[1] == '\'') p++;
	        else break;
	    }
	    buf[i++] = *p;
    	}
	if (*p != '\'') {
	    fprintf(stderr, "****non-closed apostrophe in keyword:\n");
	    fwrite(text, 1, 80, stderr);
	    fputc('\n', stderr);
	}
	/* Remove the trailing blanks [only if length<=8]
	 * Important to keep the blanks in MULTISPEC!
	 * */
	if (i<9) {
	    for (--i; (i>=0) && buf[i] == ' '; i--) ; ++i;
	}
    }
    else  {		/* Copy the Number */
	while ((p<e) && isgraph(*p) && (*p != '/'))
	  buf[i++] = *(p++);
    }
    buf[i] = 0;
    if (ocomment) {	/* Copy the comment */
	while ((p<e) && isspace(*p)) p++;
	if ((p<e) && (*p == '/')) {
	    p++;
	    while ((p<e) && isspace(*p)) p++;
	    *ocomment = (p-text);
	}
    }
    if (replaced) {	/* Mark the additional keyword is used... */
	if (verbop) {
	    printf("#...Keyword was replaced by the following:\n#");
	    fwrite(text, 1, e-text, stdout);
	    fputc('\n', stdout);
	}
	text[0] = '/';	/* Replace e.g. BITPIX by /ITPIX, used only once */
    }
    return(buf);
}

static int print_a2880()
/*++++++++++++++++
.PURPOSE  Print the FITS header stored in a2880
.RETURNS  Number of lines issued
.REMARKS  Stops ant END
-----------------*/
{
  int len, i, j, lines=0;
  char *p;
    for (i=0; i<na2880; i++) {
	for (p = a2880[i], j = 2880/80; --j>=0; p += 80) {
	    for (len=80; --len>=0; ) {
		if (iscntrl(p[len])) p[len] = ' ';
		if (isspace(p[len])) continue;
		break;
	    }
	    ++len;
	    if (len == 0) 	/* empty line, replaced by . */
		putchar('.');
	    else {
	        fwrite(p, 1, len, stdout);
	        if ((p[0] == 'E') && (p[1] == 'N') && (p[2] == 'D') &&
		    isspace(p[3])) len = -1;
	    }
	    putchar('\n');
	    lines++;
	    if (len<0) break;	/* END was encountered */
	}
    }
    return(lines);
}

static int print_head(char *fmt, char *string)
/*++++++++++++++++
.PURPOSE  Print the String with the specified number of bytes
.RETURNS  Number of characters issued
.REMARKS  When string is NULL, issue -----------
-----------------*/
{
  int n, nb, nl ;
    while ((*fmt < '1') || (*fmt > '9')) fmt++ ;
    n = atoi(fmt) ;
    if (n == 0) return(0);
    if (!string) {
	for (nb=n; --nb>=0; ) putchar('-') ;
	return(n) ;
    }
    nb = n - strlen(string) ;	/* Number of blanks */
    if (nb < 0) { 		/* Truncate...      */
	for (nb=0; nb<n; nb++) putchar(string[nb]);
	return(n);
    }
    nl = nb/2 ;			/* Left Blanks	*/
    if (nb&1) nl++;		/* One more blank at left */
    nb -= nl ;			/* Right Blanks	*/
    while (--nl >= 0) putchar(' ') ;
    printf("%s", string) ;
    while (--nb >= 0) putchar(' ') ;
    return(n) ;
}

static int read_2880(FILE *in, char *buf)
/*++++++++++++++++
.PURPOSE  Read next 2880-byte buffer
.RETURNS  2880 / 0=EOF / -1=error
-----------------*/
{
  int n, k ;
    n = fread(buf, 1, 2880, in) ;
    if (n == 0) {
	if (ferror(in)) n = -1;
	else return(0);
    }
    /*
    {
	fprintf(stderr,
	    "****EOF found prematurely (after %d*2880) in file: %s\n",
            blocno, input_name) ;
	return(-1);
    }
    */
    if (n < 0) {	/* Error */
        fprintf(stderr, "****") ;
        perror(input_name) ;
        return(-1) ;
    }
    while (n < 2880) {
        k = fread(buf+n, 1, 2880-n, in) ;
	if (k <= 0) break ;
	n += k ;
    }
    if (verbop) fprintf(stderr, "----Read: %d bytes\n", n) ;
    blocno++;
    return(n) ;
}
static int read_fits(FILE *in, char *buf, int size) 
/*++++++++++++++++
.PURPOSE  Read muliple of FITS blocks
.RETURNS  Bytes read (in principle a multiple of 2880)
-----------------*/
{
  int n=0, bytes=0;
  char block[2880], *ebuf, *pbuf;
      for (pbuf=buf, ebuf=buf+size; pbuf<ebuf; pbuf += n) {
	  if((ebuf-pbuf)<2880)  {
	      n = read_2880(in, block);
	      if(n>0) memcpy(pbuf, block, ebuf-pbuf);
	  }
	  else n = read_2880(in, pbuf);
	  if (n!=2880) {
	      if(n>0) bytes += n;
	      break;
	  }
      }
      if (pbuf==buf) return(n);
      return(pbuf-buf);	/* Number of bytes */
}

static void keys_init()
/*++++++++++++++++
.PURPOSE  Empty all definitions
.RETURNS  ---
-----------------*/
{
static int mdone = 0;
  FITSROW *pr;
  FITSKEY *pk;
    pcount = 0; gcount = 1;
    memset( dim, 0, sizeof(dim));
    memset(mdim, 0, sizeof(dim));
    memset(tdelta, 0, sizeof(tdelta));
    memset(tcrval, 0, sizeof(tcrval));
    if (fitsrow) {
        for (pr=fitsrow; pr->form[0]; pr++) {
	    if (pr->type) Free(pr->type);
	    if (pr->unit) Free(pr->unit);
	    if (pr->tdim) Free(pr->tdim);
	    if (pr->null) { if (pr->null != (char *)1) Free(pr->null); }
	    if (pr->comment) Free(pr->comment);
	    if (pr->value)   Free(pr->value);
	    memset(pr, 0, sizeof(FITSROW));
	    /* Free(pr); */
	}
	Free(fitsrow);
	fitsrow = (FITSROW *)0;
    }
    for ( pk = keys; pk->name; pk++) {
        if (pk->value) Free(pk->value) ;
        if (pk->comment) Free(pk->comment) ;
        pk->value = pk->comment = (char *)0 ;
	if (pk->name[0] == 'm') {   /* Multi: reset to #1 */
	    if (mdone == 0) pk->name = strdup(pk->name);
	    else { char *p;
	        p = pk->name + strlen(pk->name)-1;
	        if (*p == '=') --p;
	        while (*p == ' ') --p;
	        if (isdigit(*p)) {
		    if (*p != '1') *p = '1';
		    for (--p; isdigit(*p); p--) { if (*p != '0') *p = '0'; }
	        }
	    }
	}
    }
    if (arows) {
        Free(arows);
	arows = (FITSROW **)0;
    }
    if (aspec) {
	Free(aspec); 
	aspec = 0;	/* Added V3.163 */
	nspec = 0;
    }
    mdone++;
}

static int keys_find(char *rec)
/*++++++++++++++++
.PURPOSE  Find a definition, save it
.RETURNS  1 (found) / 0 (ignored key)
-----------------*/
{
  static FITSKEY *p0 = keys ;
  FITSKEY *pk ;
  FITSROW *pr;
  char *p;
  int i, n, islash=0, len;

    /* Can it be a table definition ? */
    if (fitsrow && (*rec == 'T')) {
	n = 5;
	if      (strncmp(rec, "TTYPE", 5) == 0) i = 1;
	else if (strncmp(rec, "TFORM", 5) == 0) i = 2;
	else if (strncmp(rec, "TDISP", 5) == 0) i = 3;
	else if (strncmp(rec, "TNULL", 5) == 0) i = 4;
	else if (strncmp(rec, "TBCOL", 5) == 0) i = 5;
	else if (strncmp(rec, "TUNIT", 5) == 0) i = 6;
	else if (strncmp(rec, "TDIM",  4) == 0) i = 7, n=4;
	else if (strncmp(rec, "TTMIN", 5) == 0) i = 8;
	else if (strncmp(rec, "TTMAX", 5) == 0) i = 9;
	else i = n = 0;
	if (i) {
	    n = atoi(rec+n);
	    if ((n<1) || (n>=allorow)) {
	        fprintf(stderr, "#+++ TFIELD#%-3d (>%d) ignored in: %s\n",
		  n, allorow-1, input_name);
		return(-1);
	    }
	}
	if (i && n) {
	    /* Find the keyword contents */
	    pr = &fitsrow[n-1];
	    pr->no = n;
	    if (colorder == 0) pr->ano = 0;
	    else if (pr->ano == 0) {	/* Keep the requested order */
	        pr->ano = colorder;
		colorder++;
	    }
	    p = key_value(rec, &islash, aheader);
	    if (!p) {
		fprintf(stderr,
		  "****Can't find value (missing '='?) in FITS line:\n");
		fwrite(rec, 1, 79, stderr); fputc('\n', stderr);
		return(-1);
	    }
	    if (i == 1) {		/* TTYPE */
		if (pr->type) Free(pr->type);
		pr->type = Strdup(p);
		if (pr->comment) { Free(pr->comment); pr->comment = (char *)0; }
		if (pr->value)   { Free(pr->value);   pr->value = (char *)0; }
    		if ((islash>0) && (islash<80)) {
		    len = 80-islash;
		    pr->comment = Malloc(len+1) ;
		    strncpy(pr->comment, rec+islash, len);
		    pr->comment[len] = 0;
		    while (--len>0) { 	/* Remove trailing blanks */
			if (pr->comment[len] == ' ') pr->comment[len] = 0;
			else break;
		    }
		}
	    }
	    else if (i == 2) {		/* TFORM */
	        pr->count = atoi(p);
		if (!pr->count) pr->count = 1;
		if ((vmax>0) && (pr->count>vmax)) {
		    pr->flags |= FLAG_TRUNC_VMAX;
                    fprintf(stderr,
			"#+++Column#%03d: limiting arraysize %d/%d\n",
	                n, vmax, pr->count);
                    pr->count = vmax;
		}
		while (isdigit(*p)) p++;
		strncpy(pr->form, p, sizeof(pr->form));
		pr->form[sizeof(pr->form)-1] = 0;
		/* V1.7: Accept ascii table */
		if (isdigit(p[1])) {
		    if (!pr->disp[0]) {
			strncpy(pr->disp, p, sizeof(pr->disp));
			pr->disp[sizeof(pr->disp)-1] = 0;
		    }
		    pr->len1 = pr->len1a = atoi(p+1);
		}
	        if (pr->flags&FLAG_TRUNC_VMAX) {
	        }
	    }
	    else if (i == 3) {		/* TDISP */
		if (pr->disp[0] && strcmp(pr->disp, p))
		    /* Ascii table: flag a change of format */
		    pr->flags |= FLAG_CONVERT;
		strncpy(pr->disp, p, sizeof(pr->disp));
		pr->disp[sizeof(pr->disp)-1] = 0;
	    }
	    else if (i == 4) {		/* TNULL */
		if (pr->null && (pr->null != (char *)1)) Free(pr->null);
		/* Ignore the leading blanks.. */
		while (*p == ' ') p++;
		pr->null = Strdup(p);
	    }
	    else if (i == 5) {		/* TBCOL */
		/* The TBCOL value is stored here, must be between
		 * 1 and reclen. Change to offset and verification
		 * are done later, in asctable()
		**/
		pr->opos = atoi(p);
	    }
	    else if (i == 6) {		/* TUNIT */
		if (pr->unit) Free(pr->unit), pr->unit = (char *)0;
		while (*p == ' ') p++;
		if (*p) pr->unit = Strdup(p);
	    }
	    else if (i == 7) {		/* TDIM, e.g. (1200,25) */
		if (pr->tdim) Free(pr->tdim), pr->tdim = (char *)0;
		while (*p == ' ') p++;
		if (*p) pr->tdim = Strdup(p);
		/* No blank in the tdim field... */
		for (i=n=0, p=pr->tdim; p[n]; n++) {
		    if (p[n] == ' ') continue;
		    if (i!=n) p[i] = p[n];
		    i++;
		}
		p[i] = 0;
	    }
	    else if (i == 8) {		/* TTMIN (V3.2) */
		pr->flags |= FLAG_TRUNC_MIN;
		pr->tmin = a2f(p);
	    }
	    else if (i == 9) {
		pr->flags |= FLAG_TRUNC_MAX;
		pr->tmax = a2f(p);
	    }
	    return(1);
	}
    }

    for ( pk = p0; pk->name && strncmp(pk->name+1, rec, 9) ; pk++) ;
    if ((pk->name == (char *)0) && (p0 != keys)) {	/* Not Found */
        for (pk = keys; (pk < p0) && strncmp(pk->name+1, rec, 9); pk++) ;
	if (pk == p0) return(0) ;
    }
    else if (!pk->name) return(0);

    /* I've found the keyword... Fill now the value !   */
    p0 = pk ;
    if (pk->value && (pk->name[0] != 'm')) { 	/* Value already existing */
        fprintf(stderr, "++++FITS keyword duplicated: %s\n", pk->name+1) ;
        Free(pk->value) ;
    }
    p = key_value(rec, &islash, aheader);
    switch(pk->name[0]) {
      case 'i':
        pk->value = Malloc(sizeof(int)) ;
        *((int *)(pk->value)) = atoi(p);
	/* Special: TFIELDS keyword */
        if (pk == &(keys[iTFIELDS])) {
	    allorow = 1+TFIELDS;	/* fitsrow.form=0 for end of rows */
            n = allorow*sizeof(FITSROW);
            fitsrow = (FITSROW *)Malloc(n);
	    memset(fitsrow, 0, n);
        }
        break ;
      case 'd':
	/* NOTE: there are 2 synomyms for BNULL */
	if (pk == &(keys[iNULL+1])) pk--;
        pk->value = Malloc(sizeof(double)) ;
        *((double *)(pk->value)) = a2f(p);
        break ;
      case 'c':
	pk->value = Strdup(p);
	ttrim(pk->value);
        break ;
      case 'm':  /* Concatenate */
	if (pk->value) { char *o = pk->value;
	    pk->value = strdup2(pk->value, p);
	    free(o);
	}
	else {
	    if (pk == &(keys[iWAT+2])) { /* MULTISPEC definition */
		nspec = 1;
		if (keys[iNAXIS1+6].value) 
		    nspec = *(int *)(keys[iNAXIS1+6].value);
		aspec = (MULTISPEC *)malloc(nspec*sizeof(MULTISPEC));
		memset(aspec, 0, nspec*sizeof(MULTISPEC));
	    }
	    pk->value = strdup(p);
	}
	/* increment name */
	p = pk->name + strlen(pk->name)-1;
	if (*p == '=') --p; while (*p == ' ') --p;
	i=0; while (isdigit(*p)) { --p; i++; } ++p;
	n = atoi(p)+1;
	while (--i>=0) { p[i]='0'|(n%10); n /= 10; }
	break;
      default:
        fprintf(stderr, "****Internal Error, invalid FITSKEY: %s\n", pk->name) ;
        exit(1) ;
    }

    /* Store also the comment */
    if ((islash>0) && (islash<80)) {
	len = 80-islash;
	if (pk->comment) free(pk->comment);
        pk->comment = Malloc(len+1) ;
        strncpy(pk->comment, rec+islash, len);
        pk->comment[len] = 0;
        while (--len>0) { 	/* Remove trailing blanks */
	    if (pk->comment[len] == ' ') pk->comment[len] = 0;
	    else break;
        }
    }
    return(1) ;
}

static int add2header(char *aheader)
/*++++++++++++++++
.PURPOSE  Process the additional header
.RETURNS  Number of additional lines processed
-----------------*/
{
  int i, c, n=0;
  char *p;
  if (!aheader) return(0);
    /* Take care of the additional header lines
       Keep the order of the list
    */
    /* The order of aheader is kept with "-a ="  option */
    if (acols) colorder = *acols == '=' ;
    for (p=aheader; *p; p += i) {
        if (iscntrl(*p)) { i = 1; continue; }
        i = strloc(p, '\n');
        if (!isalpha(*p)) continue;
        c = p[i]; p[i] = 0;
        keys_find(p);
        p[i] = c;
	n++;
    }
    colorder = 0;	/* Reset the colorder flag */
    return(n);
}

static int get_header(FILE *in, int verify_xten)
/*++++++++++++++++
.PURPOSE  Interpret the header, save the result in the  'keys'
.RETURNS  1 (OK, to process) / 0 (OK, to skip) / -1 (error or EOF)
.REMARKS  a2880 and na2880 are set
-----------------*/
{
  static char *card1 = "SIMPLE  =";
  static double abuf[360] ;	/* 2880 bytes */
  char *buf ;
  char *p, *e;
  int  n, i, stop;


    /*printf("#...get_header from offset=%lld\n", ftell(in)) ;*/
    /* V1.6: Reinitialize */
    if (!in) {
	card1 = "SIMPLE  =";
	xtno = -1;
	blocno = 0;
	return(0);
    }
    /* Initialize the buffers for headers */
    buf = (char *)abuf ;
    buf[0] = 0 ;
    na2880 = 0;			/* Nothing  read  */
    if (!a2880[0]) a2880[0] = Malloc(2880);

    /* Find out the interesting parameters from Header */
    n = read_2880(in, buf) ;
    if (n == 0) return(-1);
    if (n < 12) {		/* EOF was Found! */
	if ((n==1) && (buf[0] == '\n'))	/* V2.99 */
	    return(-1);
	fprintf(stderr, "****Got %dbytes for %s#%d\n",
		n, input_name, xtno+1) ;
	if (n<0) perror("****");
	return(-1);
    }
    xtno++;			     /* In a new extension */
    memcpy(a2880[0], buf, 2880); na2880++;
    if (strncmp(buf, card1, 9)) {	/* Not a FITS file */
        fprintf(stderr, "****Not a %s file (file#%d): %s\n",
	  card1, xtno, input_name) ;
        exit(1) ;
    }
    card1 = "XTENSION=";
    keys_init() ;

    p = buf ; e = buf + n ;
    for (stop = 0; stop == 0; p += 80) {
        if (p >= e) {
	    n = read_2880(in, p=buf);
	    if (n<=0) {
	        if (n == 0) fprintf(stderr,
		    "****Unexpected EOF (after %d*2880) of FITS file: %s\n",
		    blocno, input_name);
		return(-1);
	    }
	    if (!a2880[na2880]) a2880[na2880] = Malloc(2880);
	    memcpy(a2880[na2880], buf, 2880); na2880++;
	    e = p + n;
	}
        if ((p[0] == 'E') && (p[1] == 'N') && (p[2] == 'D')) {
	    stop = iscntrl(p[3]) || isspace(p[3]);
        }
        else keys_find(p) ;
    }

#if 0	/* Done only when actual header requested */
    /* Take care of the additional header lines
       Keep the order of the list
    */
    if (aheader) {
	/* The order of aheader is kept with "-a ="  option */
	if (acols) colorder = *acols == '=' ;
        for (p=aheader; *p; p += i) {
	    if (iscntrl(*p)) { i = 1; continue; }
	    i = strloc(p, '\n');
	    if (!isalpha(*p)) continue;
	    c = p[i]; p[i] = 0;
	    keys_find(p);
	    p[i] = c;
        }
	colorder = 0;	/* Reset the colorder flag */
    }
#endif

    /* Verify Basic Keywords are here */
    for (i=0; i< 2; i++) {
        if (keys[i].value) continue ;
        fprintf(stderr, "****Missing keyword: %s\n", keys[i].name) ;
    }
    bitpix = *(int *)(keys[0].value) ;
    naxis  = *(int *)(keys[1].value) ;
    if (keys[iGCOUNT].value) gcount = *(int *)(keys[iGCOUNT].value);
    if (keys[iPCOUNT].value) pcount = *(int *)(keys[iPCOUNT].value);
    if (keys[2].value) bscale = *(double *)(keys[2].value) ;
    else bscale = 1. ;
    if (keys[3].value) bzp  = *(double *)(keys[3].value) ;
    else bzp = 0. ;
    len1el = (bitpix/8) ;
    if (len1el < 0) len1el = -len1el ;

    /* Save the current block address = address of beginning of data */
    blocno_data = blocno;

    /* Verify: if axten specified, is that OK ? (Mod. V3.14) */
    if ((!verify_xten) || (!axten)) {	/* Additional definitions */
	if (aheader)  add2header(aheader);	/* Additional keywords */
	return(1);
    }

    if (xbyname) {		/* axten specifies an extension name */
	if (EXTNAME && (strcmp(axten, EXTNAME) == 0)) {
	    add2header(aheader);
	    return(1);
	}
	return(0);
    }

    n = atoi(axten);
    if (isdigit(*axten)) {
	if (xtno < n) 	/* Required larger xtension */
	    return(0);
	if (xtno > n) {	/* Problem... */
	    fprintf(stderr, "++++fits2a: Can't process %s#%d\n",
		    input_name, n);
	    n = 0;
	}
	else n = 1;
	while(isdigit(*axten)) axten++;
	if (*axten == ',') axten++;
	add2header(aheader);	/* Insert additional keywords */
	return(n);
    }
    else if (*axten == '-') {
	n = -n; if (n == 0) n = 99999999;
	if (xtno == n) for (++axten; isdigit(*axten); axten++) ;
	if (xtno <= n) {
	    add2header(aheader);
	    return(1); 	/* OK */
	}
    }
    /* Bad definition ? */
    fprintf(stderr, "++++fits2a: Can't process %s#%s\n",
	input_name, axten);
    return(0);
}


/*==================================================================*/
static MULTIFUNCTION *get_multifunction(char *text)
/*++++++++++++++++
.PURPOSE  In the MULTISPEC context, interpret the function
.RETURNS  The function / NULL
.REMARKS  Stops when text='"'
-----------------*/
{
  MULTIFUNCTION mf, *new;
  char *p = text;
  int i, n;
    while (*p == ' ') p++;
    if (*p == '"') return(0);
    mf.weight = a2f(p);  p += skip_num(p);
    mf.olam   = a2f(p);  p += skip_num(p);
    mf.ftype  = atoi(p); p += skip_num(p);
    mf.order  = atoi(p); p += skip_num(p);
    mf.pmin   = a2f(p);  p += skip_num(p);
    mf.pmax   = a2f(p);  p += skip_num(p);
    n = mf.order;
    new = (MULTIFUNCTION *)malloc(sizeof(MULTIFUNCTION)+(n-2)*sizeof(double));
    memcpy(new, &mf, sizeof(MULTIFUNCTION));
    for (i=0; i<n; i++) {
	new->coeff[i] = a2f(p);
	p += skip_num(p);
    }
    if (isdigit(*p)) {
	fprintf(stderr, "#***Too complex multispec(%s):\n%s\n",
	    multifunct[mf.ftype], text);
    }
    return(new);
}

/*==================================================================*/
static int skip_data(FILE *in)
/*++++++++++++++++
.PURPOSE  Skip the current data part
.RETURNS  0 (OK) / EOF (error)
.REMARKS
-----------------*/
{
  int4 bytes;
  int  skip, i, j, dim, *a;
  char buf[2880];

    /* Compute the number of bytes to skip */
    if (naxis == 0) return(0);
    bytes = len1el;
    for (i=iNAXIS1, j=0; keys[i].name && (j<naxis); j++, i+=perAXIS) {
	a = (int *)(keys[i].value);
        dim = a ? *a : 1 ;
	if (dim==0) {
	    fprintf(stderr, "#+++%s %d ==> forced dim[%d]=1.\n",
		keys[i].name+1, dim, j+1);
	    dim = 1;
	}
	if (a) bytes *= dim;
	else   bytes = 0;
    }
    if (pcount) bytes += pcount*len1el;	/* Added V3.1 */
    if (gcount) bytes *= gcount;	/* Added V3.1 */

    skip = bytes/2880;		/* Number of 2880 blocks to skip */
    if (bytes%2880) skip++;	/* The last (incomplete) block   */
    if (blocno_data && (blocno > blocno_data))	/* A part of the data read */
	skip -= (blocno - blocno_data);
    /* V2.5: seek if possible */
    if ((skip>20) && isregfile(in)) {
	bytes = (int4)fseek(in, skip*2880, SEEK_CUR);
	if (bytes == (-1)) {
	    fprintf(stderr,
		"****Unexpected error seek(%d*2880) of FITS file: %s\n",
		skip, input_name);
	    return(-1);
	}
	blocno += skip;
	return(0);
    }

    while (--skip >= 0) {
	i = read_2880(in, buf);
        if (i <= 0) {
	    if (i == 0) fprintf(stderr,
		"****Unexpected EOF (after %d*2880) of FITS file: %s\n",
		blocno, input_name);
	    return(-1);
	}
    }
    return(0);
}

static void print_nan(char *fmt)
/*++++++++++++++++
.PURPOSE  Print the NULL value
.RETURNS  ---
.REMARKS  Print 3 dashes
-----------------*/
{
  char *p;
  int w, nb, na;
    for (p=fmt; *p && (!isdigit(*p)); p++) ;
    w = atoi(p);	/* Field Width */
    if (w == 0) return;
    nb = w-3;		/* Blanks before --- */
    na = 0;		/* Blanks  after --- */
    while (isdigit(*p)) p++;
    if (*p == '.') {	/* Align dash on decimal point */
	na = atoi(p+1) - 1; if (na<0) na=0;
	nb = w - 3 - na;
    }
    if (nb < 0) {	/* Can't issue --- */
	w = 3+nb;
	nb = 0;
    }
    else w = 3;

    while (--nb>=0) putchar(' ');
    while (--w>=0)  putchar('-');
    while (--na>=0) putchar(' ');
}

static void print_null(FITSROW *pr)
/*++++++++++++++++
.PURPOSE  Print the NULL value
.RETURNS  ---
.REMARKS  Print 3 dashes or nothing
-----------------*/
{
  int i; /* char *p; */
  if (optb) return;	/* -b option: Nothing out... */
    /* for (p = pr->Cfmt; ispunct(*p); p++) ;
    i = atoi(p); */
    i = pr->len1a;
    if (pr->form[0] != 'A') {	/* Numeric null */
	while (i>3) { putchar(' '); i--; }
	if (i>1) while (i>0) { putchar('-'); i--; }
    }
    while (i>0) { putchar(' '); i--; }
}

static int decimals(double val, int dmax)
/*++++++++++++++++
.PURPOSE  Process a BINARY TABLE
.RETURNS  Number of significant decimals
.REMARKS
-----------------*/
{
  double v, m;
  int i, max;
    if (dmax >= ITEMS(dexp)) dmax = ITEMS(dexp)-1;
    v = fabs(val);
    for (i=0, max=dmax; (max>0) && (v>=dexp[i]); i++) --max;
    if ((v*dexp[max]) <= 1.)	/* Very small number */
	return(dmax);
    for (i=0, m=1./dexp[max]; i<max; i++) {
	v -= floor(v);	/* Fraction */
	if (v > 0.5) v = 1.-v;
	if (v <= m) break;
	v *= 10;
	m *= 10;
    }
    return(i);
}

static int digits(int val)
/*++++++++++++++++
.PURPOSE  Compute the number of digits required in ascii
.RETURNS  Number of digits
.REMARKS
-----------------*/
{
  int i;
    i = (val < 0 ? 2 : 1);
    while (val >= 10) i++, val /= 10;
    return(i);
}

static int read_record(FILE *in, char *record, int reclen)
/*++++++++++++++++
.PURPOSE  Read a full record -- must be reclen bytes
.RETURNS  reclen, -1 for error
.REMARKS
-----------------*/
{
  int part, n;
    part = 0;
    while (part < reclen) {
       	n = fread(record+part, 1, reclen-part, in);
        if (n <= 0) {
            fprintf(stderr, "****");
            if (n < 0) perror(input_name);
            else fprintf(stderr, "unexpected EOF: %s\n", input_name);
    	    return(-1);
	}
        part += n;
    }
    return(part);
}

/*==================================================================
		Process BINTABLE and scii TABLE
 *==================================================================*/
static FITSROW *colfind(FITSROW **c1, char *name)
/*++++++++++++++++
.PURPOSE  Find a column from its name
.RETURNS  column / NULL
-----------------*/
{
  int i;

    /* Could be a column number ? */
    for (i=0; isdigit(name[i]); i++) ;
    if ((i>0) && (!name[i])) { int colno;
	colno = atoi(name);		/* Column number */
	/* V2.12: Columns may have been mixed!   */
	for (i=0; c1[i]; i++) {
	    if (c1[i]->no == colno)
	    	return(c1[i]);
	}
	return((FITSROW *)0);
    }

    /* Column specified by a name */
    for (i=0; c1[i]; i++) {
	if (strcmp(c1[i]->type, name) == 0)
	    return(c1[i]);
    }
    /* Nor found, try case insensitive */
    for (i=0; c1[i]; i++) {
	if (strcasecmp(c1[i]->type, name) == 0)
	    return(c1[i]);
    }
    return((FITSROW *)0);
}

static FITSROW **acolnext(FITSROW **c1, char *pattern)
/*++++++++++++++++
.PURPOSE  Find next column satisfying to a pattern
.RETURNS  column / NULL
-----------------*/
{
    while (*c1) {
	if (strmatch(pattern, (*c1)->type))
	    return(c1);
	c1++;
    }
    return(c1);
}

static FITSROW *colword(FITSROW **c1, char *name)
/*++++++++++++++++
.PURPOSE  Find a column from its name (only non-blanks)
.RETURNS  column / NULL
-----------------*/
{
  static char stops[] = "=,;";	/* These chars end the name */
  char buf[20]; int i;
    while (*name == ' ') name++;
    for (i=0; (i<sizeof(buf)) && isgraph(name[i]); i++) {
	buf[i] = name[i];
	if (!ispunct(buf[i])) continue;
	if (strchr(stops, buf[i])) break;
    }
    if (i>=sizeof(buf)) 	/* Too long name... */
	return(colfind(c1, name));
    buf[i] = 0;
    return(colfind(c1, buf));
}

static int colcmp(FITSROW **c1, FITSROW **c2)
/*++++++++++++++++
.PURPOSE  Compare 2 column for sorting on ano
.RETURNS  c1 - c2
-----------------*/
{
    if ((*c1)->ano == 0) {	/* Column to ignore */
	if ((*c2)->ano == 0) return((*c1)->no - (*c2)->no);
	else return(1);
    }
    if ((*c2)->ano == 0) return(-1);
    return((*c1)->ano - (*c2)->ano);
}

static double bound_F(FITSROW *pr, double D)
/*++++++++++++++++
.PURPOSE  Set value within specified limits
.RETURNS  The bounded value
-----------------*/
{
    /* V3.2: bound the value */
    if ((pr->flags&FLAG_TRUNC_MIN) && (D<pr->tmin)) {
	D = pr->tmin; pr->trunc[0] += 1;
    }
    if ((pr->flags&FLAG_TRUNC_MAX) && (D>pr->tmax)) {
	D = pr->tmax; pr->trunc[1] += 1;
    }
    return(D);
}

static double bounded_F(FITSROW *pr, double D)
/*++++++++++++++++
.PURPOSE  Check the range
.RETURNS  The bounded value
-----------------*/
{
    if (D > pr->max) {
	fprintf(stderr,
	  "++++Too large value line #%d '%s' col.%s, replaced by %g: %g\n",
          recno+1, pr->disp, pr->type, pr->max, D);
	return(pr->max);
    }
    if (D < pr->min) {
	fprintf(stderr,
          "++++Too small value line #%d '%s' col.%s, replaced by %g: %g\n",
          recno+1, pr->disp, pr->type, pr->min, D);
	return(pr->min);
    }
    return(D);
}

static int bound_J(FITSROW *pr, int J)
/*++++++++++++++++
.PURPOSE  Check the Integer
.RETURNS  The bounded value
-----------------*/
{
    /* V3.2: bound the value */
    if (pr->flags&(FLAG_TRUNC_MIN|FLAG_TRUNC_MAX)) { int *aJ;
	if (!(pr->flags&FLAG_CONVERT)) {	/* Keep values as int */
	    pr->flags |= FLAG_CONVERT;
	    aJ = (int *)&pr->tmin;
	    if (pr->flags&FLAG_TRUNC_MIN) *aJ = pr->tmin;
	    aJ = (int *)&pr->tmax;
	    if (pr->flags&FLAG_TRUNC_MAX) *aJ = pr->tmax;
	}
	aJ = (int *)&pr->tmin;
        if ((pr->flags&FLAG_TRUNC_MIN) && (J<*aJ)) {
	    if (verbop) printf("#...bounded_J(%d) ==> %d\n", J, *aJ);
	    J = *aJ; pr->trunc[0] += 1;
	}
        aJ = (int *)&pr->tmax;
        if ((pr->flags&FLAG_TRUNC_MAX) && (J>*aJ)) {
	    if (verbop) printf("#...bounded_J(%d) ==> %d\n", J, *aJ);
	    J = *aJ; pr->trunc[1] += 1;
	}
    }
    return(J);
}

static int bounded_J(FITSROW *pr, int J)
/*++++++++++++++++
.PURPOSE  Check the Integer
.RETURNS  The bounded value
-----------------*/
{
    if (J > *(int *)&pr->max) {
	fprintf(stderr,
          "++++Too large value line #%d '%s' col.%s, replaced by %d: %d\n",
          recno+1, pr->disp, pr->type, *(int *)(&pr->max), J);
	return(*(int *)&pr->max);
    }
    if (J < *(int *)&pr->min) {
	fprintf(stderr,
          "++++Too small value line #%d '%s' col.%s, replaced by %d: %d\n",
          recno+1, pr->disp, pr->type, *(int *)(&pr->min), J);
	return(*(int *)&pr->min);
    }
    return(J);
}

static int8 bound_K(FITSROW *pr, int8 K)
/*++++++++++++++++
.PURPOSE  Check the Integer
.RETURNS  The bounded value
-----------------*/
{
    /* V3.2: bound the value */
    if (pr->flags&(FLAG_TRUNC_MIN|FLAG_TRUNC_MAX)) { int8 *aK;
#if int8l == 1
	static char f8[] = "#...bounded_J(%ld) ==> %d\n";
#else
	static char f8[] = "#...bounded_J(%lld) ==> %d\n";
#endif
	if (!(pr->flags&FLAG_CONVERT)) {	/* Keep values as int8 */
	    pr->flags |= FLAG_CONVERT;
	    aK = (int8 *)&pr->tmin;
	    if (pr->flags&FLAG_TRUNC_MIN) *aK = pr->tmin;
	    aK = (int8 *)&pr->tmax;
	    if (pr->flags&FLAG_TRUNC_MAX) *aK = pr->tmax;
	}
	aK = (int8 *)&pr->tmin;
        if ((pr->flags&FLAG_TRUNC_MIN) && (K<*aK)) {
	    if (verbop) printf(f8, K, *aK);
	    K = *aK; pr->trunc[0] += 1;
	}
        aK = (int8 *)&pr->tmax;
        if ((pr->flags&FLAG_TRUNC_MAX) && (K>*aK)) {
	    if (verbop) printf(f8, K, *aK);
	    K = *aK; pr->trunc[1] += 1;
	}
    }
    return(K);
}

static int8 bounded_K(FITSROW *pr, int8 K)
/*++++++++++++++++
.PURPOSE  Check the Integer
.RETURNS  The bounded value
-----------------*/
{
    if (K > *(int8 *)&pr->max) {
#if int8l == 1
	fprintf(stderr,
          "++++Too large value line #%d '%s' col.%s, replaced by %ld: %ld\n",
          recno+1, pr->disp, pr->type, *(int8 *)(&pr->max), K);
#else
	fprintf(stderr,
          "++++Too large value line #%d '%s' col.%s, replaced by %lld: %lld\n",
          recno+1, pr->disp, pr->type, *(int8 *)(&pr->max), K);
#endif
	return(*(int *)&pr->max);
    }
    if (K < *(int8 *)&pr->min) {  /* Error fixed V3.32 */
#if int8l == 1
	fprintf(stderr,
          "++++Too small value line #%d '%s' col.%s, replaced by %ld: %ld\n",
          recno+1, pr->disp, pr->type, *(int8 *)(&pr->min), K);
#else
	fprintf(stderr,
          "++++Too small value line #%d '%s' col.%s, replaced by %lld: %lld\n",
          recno+1, pr->disp, pr->type, *(int8 *)(&pr->min), K);
#endif
	return(*(int *)&pr->min);
    }
    return(K);
}

static int check_Cform(FITSROW *pr)
/*++++++++++++++++
.PURPOSE  Verify C format compatible with datatype
.RETURNS  0 (ok) / -1 (error)
.REMARKS  Added V2.95
-----------------*/
{
  static char iformat[] = "dxou";	/* Acceptable Integer format */
  static char dformat[] = "efg";	/* Acceptable Real    format */
  char *p, fc, cc;
  int status = 0;

    if (!pr->disp) return(-1);
    for (p=pr->form; isspace(*p) || isdigit(*p); p++) ;
    fc = toupper(*p);	/* Letter of Fortran format */
    p = pr->Cfmt;
    if ((p[0]==0)||((p[1]==0)&&(p[0]=='%')))  /* Column not edited */
	return(status);
    p += strlen(p) -1; cc = tolower(*p);
    switch(fc) {
      case 'A':		/* Character */
	if (cc != 's') status = -1;
	break;
      case 'L': case 'B': case 'I': case 'J':  case 'K':
	if (!strchr(iformat, cc)) status = -1;
	break;
      case 'D': case 'E': case 'F':
	if (!strchr(dformat, cc)) status = -1;
	break;
      default:
	status = -1;

    }
    if (status<0) fprintf(stderr,
	"#***Incompatibility for col.#%d (%s): TFORM=%s TDISP=%s Cfmt=%s\n",
	pr->no, pr->type, pr->form, pr->disp, pr->Cfmt);
    return(status);
}

static int set_Cform(FITSROW *pr)
/*++++++++++++++++
.PURPOSE  Set the C format in a row.
.RETURNS  0 (nothing done) / 1
.REMARKS  Initialises also the min/max/null (when tdisp not set).
-----------------*/
{
  char abuf[40], *p, *e, fc;
  int i, minlen, set_minmax;
  int8 m;

    if (pr->Cfmt[0] == '%') {	/* Alreay in C */
	for (p=pr->Cfmt+1; *p && (!isdigit(*p)); p++) ;
	if ((pr->len1a = atoi(p)))  /* A length is specified ==> ok */
	    return(0);
	/* Here len1a is zero */
	if (pr->Cfmt[1]==0) return(0);	/* Format "%" => ignore column */
    }

    /* Do we need to set the min/max ? */
    set_minmax =  tdisp | (pr->flags&FLAG_SELECT) ? 0 : 1;

    /* Give the width = name */
    minlen = ohead ? strlen(pr->type) : 0 ;
    if (opar||(odata==0)) minlen = 0;
    				/* Horizontal Edition: no need to add blanks */

    for (p=pr->disp; isspace(*p) || isdigit(*p); p++) ;
    fc = toupper(*p);		/* Letter of Fortran format */
    if (*p) p++;		/* points to length */
    pr->Cfmt[0] = '%';
    switch(fc) {
      case 'L':
	i = MAX(minlen, 1);
	if (i >= sizeof(abuf)) i = sizeof(abuf)-1;
	pr->len1a = i;
	sprintf(pr->Cfmt+1, "%dd", i);
	return(1);
      case 'A':
        i = atoi(p);	/* Length in TDISP e.g. A16 */
	i = MAX(minlen, i);
	if (optb) strcpy(pr->Cfmt, "%s");
	else sprintf(pr->Cfmt+1, "-%ds", i);
	if (pr->len1a == 0) {
	    if (i < pr->count) {
		fprintf(stderr, "#+++ TFIELD '%s' truncated: %d/%d\n",
		    pr->type, i, pr->count);
		pr->len1a = i;
	    }
	    else {
		pr->len1a = pr->count;
		if (i > pr->count) fprintf(stderr,
		    "#... TFIELD '%s' padded with %d blanks\n",
		    pr->type, i-pr->count);
	    }
	    pr->count = 1;
	}
        /* Save the values:
           pr->min = number of bytes to copy
           pr->max = number of bytes in output (max-min = blanks to add)
           pr->nul = number of bytes that have to be null
        */
	if (set_minmax)  {
            *(int *)&pr->min = MIN(i, pr->len1a);
            *(int *)&pr->max = i;
            *(int *)&pr->nul = pr->len1a - i;
	}
        return(1);
      case 'B':
      case 'I':  case 'J': case 'K':
        /* Compute Min and Max */
        i = atoi(p);	/* How many bytes */
	i = MAX(minlen, i);
	if (i >= sizeof(abuf)) i = sizeof(abuf)-1;
	if (pr->len1 == 8)  {	/* 64-bit integer */
#if int8l == 1
	    if (optb) strcpy(pr->Cfmt, "%ld");
	    else sprintf(pr->Cfmt+1, "%dld", i);
#else
	    if (optb) strcpy(pr->Cfmt, "%lld");
	    else sprintf(pr->Cfmt+1, "%dlld", i);
#endif
	}
	else {			/* 32-bit integer */
	    if (optb) strcpy(pr->Cfmt, "%d");
	    else sprintf(pr->Cfmt+1, "%dd", i);
	}
	pr->len1a = i;
	if (set_minmax) {
	    if (pr->len1 == 8) {/* 64-bit integer */
                if (i >= 18) {
    	            *(int8 *)&pr->max =  0x7fffffffffffffffLL;
    	            *(int8 *)&pr->min = -0x7fffffffffffffffLL;
                }
                else {
                    for (m=1; --i>=0; m*=10);
    	            *(int8 *)&pr->max = m-1;
    	            *(int8 *)&pr->min = 1 - (m/10);
                }
	    }
	    else {		/* 32-bit integer */
                if (i >= 10) {
    	            *(int *)&pr->max =  0x7fffffff;
    	            *(int *)&pr->min = -0x7fffffff;
                }
                else {
                    for (m=1; --i>=0; m*=10);
    	            *(int *)&pr->max = m-1;
    	            *(int *)&pr->min = 1 - (m/10);
                }
	    }
	}
	return(1);
      case 'D':            fc = 'e'; break;
      case 'F': /* Compute Min and Max allowed for fixed formats */
       	    memset(abuf, '9', sizeof(abuf)-1);
        i = atoi(p);
	i = MAX(minlen, i);
	if (i >= sizeof(abuf)) i = sizeof(abuf)-1;
        abuf[i] = 0;
        if ((e = strchr(p, '.'))) i -= atoi(e+1)+1;
        if (i>=0) abuf[i] = '.';
	if (set_minmax) {
            pr->max = atof(abuf);
            abuf[0] = '-';
            pr->min = atof(abuf);
	}
    }
    fc = tolower(fc);
    if (optb) {		/* Ignore the leading blanks */
	while (isdigit(*p)) p++;
    }
    strcpy(pr->Cfmt+1, p);
    /* Note that for exponent-based values, C formats indicates the
       number of decimals, while in Fortran we have the number of
       significant digits. Therefore decrease the number of
       decimals in exponential notation.
    */
    if ((fc == 'e') && (e = strchr(pr->Cfmt, '.'))) {
        ++e;
        sprintf(e, "%d", atoi(e)-1) ;
    }
    e = pr->Cfmt + strlen(pr->Cfmt);
    *(e++) = fc; *e = 0;
    for (p=pr->Cfmt+1; *p && (!isdigit(*p)); p++) ;
    pr->len1a = atoi(p);
    return(1);
}

static int sort_cols(FITSROW **arows, char *acols)
/*++++++++++++++++
.PURPOSE  Sort the columns from a list
.RETURNS  -1 (error) / Number of columns selected
.REMARKS  The list of columns may start by
       	- for a list of columns to remove
   Columns specified by their number(s).
-----------------*/
{
  int colorder, i, o, ncol=0, pattern=0, from_names;
  FITSROW **ar, *pr;
  char *p;

    /* Choose and order the columns if asked */
    if ((!acols) || (!acols[0])) {	/* All columns */
	for (ar=arows; *ar; ar++) {
	    pr = *ar;
	    pr->ano = pr->no;
	    if (pr->ano>0) ncol++;
	}
	return(ncol);
    }
    p = acols;
    if (*p == '=') {		/* -c =a ==> ano already assigned */
	qsort(arows, allorow-1, sizeof(FITSROW *), (SORT_FCT)colcmp);
	for (ar=arows; *ar; ar++) {
	    pr = *ar;
	    if (pr->ano>0) ncol++;
	}
	return(ncol);
    }

    /* Columns given as names or numbers ? */
    for (from_names=0; *p && (from_names==0); p++) {
	if (isdigit(*p)) continue;
	if ((*p == ',') || (*p == '-') || isspace(*p)) continue;
	from_names++;
    }

    p = acols; colorder = 1;
    if (*p == '-') {		/* List of columns to remove */
        colorder = 0;
	for (ar=arows; *ar; ar++) {
	    if ((*ar)->ano == 0) (*ar)->ano = (*ar)->no;
	}
	p++;
    }
    else for (ar=arows; *ar; ar++) (*ar)->ano = 0;
    if (*p == '*') {
	pattern = 1;	/* Names with pattern matching */
	p++;
    }

    /* Set the column order */
    /* Accept numbers (e.g. 1, 2-5, 3, ...)
     * or nameswith '*' ==> matching "-" BUT ... */
    while(*p) {
        if (isspace(*p)) { p++; continue; }
	if (from_names) {
	    ar = arows;
	    for (i=0; isgraph(p[i]) && (p[i] != ','); i++) ;
	    if ((i>1) && (p[i-1] == ',')) --i;
	    o = p[i]; p[i] = 0;
	    if (pattern) for (ar=acolnext(arows,p); *ar; ar=acolnext(ar+1,p)) {
		pr = *ar;
		if (colorder == 0) pr->ano = 0;
		else pr->ano = pr->no;
	    }
	    else { 	/* Get name only */
		pr = colfind(arows, p);
		if (!pr) fprintf(stderr,
		    "++++Unknown column name in  -c list: %s\n", p);
		else if (colorder == 0) pr->ano = 0;
		else if (pr->ano  == 0) pr->ano = colorder++;
	    }
	    p += i; *p = o; if (*p == ',') p++;
	    continue;
	}
        if (ispunct(*p)) { p++; continue; }
	i = atoi(p);
	while (isdigit(*p)) p++;
	if (*p == '-') {	/* Range of Columns */
	    p++;
	    o = atoi(p);
	    while (isdigit(*p)) p++;
	}
	else o = i;
	--i; --o;
	if ((i<0) || (o<0) || (o>=allorow) || (i>=allorow) || (i>o)) {
	    fprintf(stderr,
	      "#***Only numbers are recognized in -c +%d: %s\n",
	      (int)(p-acols), acols);
	    return(-1);
	}
	while (i <= o) {
	    if (colorder == 0) arows[i]->ano = 0;
	    else {
	        if (arows[i]->ano) fprintf(stderr,
	          "++++Column %d'%s' asked twice as #%d and #%d: %s\n",
	          i, arows[i]->type, arows[i]->ano, colorder, acols);
	        else arows[i]->ano = colorder++;
	    }
	    i++;
	}
    }	/* loop on column specifications */
    qsort(arows, allorow-1, sizeof(FITSROW *), (SORT_FCT)colcmp);
    for (ar=arows; *ar; ar++) {
        pr = *ar;
        if (pr->ano>0) ncol++;
    }
    return(ncol);
}

static int qopt_cols(FITSROW **arows, char **cst)
/*++++++++++++++++
.PURPOSE  Install the contraints on columns.
.RETURNS  -1 (error) / 0 (nothing done) / 1 (sorting done)
.REMARKS  One constraint = column_name=[!]min[,max]
   Columns may be specified by its number
-----------------*/
{
  int qno, o;
  FITSROW *pr;
  char *p;
  double lims[2];

    /* Is there any constraint ? */
    if (!cst[0]) return(0);

    /* Install the constraints */
    lims[0] = lims[1] = 0;
    for (qno=0; cst[qno]; qno++) {
	p = cst[qno];
	o = strloc(p, '=');
	if (!p[o]) {	/* No constraint sign */
	    fprintf(stderr,
		    "++++Bad constraint (missing =), ignored: %s\n", p);
	    continue;
	}
	p[o] = 0;
	pr = colfind(arows, p) ; p[o] = '=';
	if (!pr) {
	    fprintf(stderr,
		    "++++Unknown column name, -q ignored: %s\n", p);
	    continue;
	}
	pr->flags &= ~FLAG_SELECT;	/* Remove all selections */

	/* Get the limits */
	p += o+1;
	if (*p == '!')		/* NOT limit */
	    pr->flags |= FLAG_SELECT_NOT;
	o = strloc(p, ',');	/* Separates the 2 limits */

	if (*p) {	/* There is a lower limit */
	    pr->flags |= FLAG_SELECT_MIN;
	    lims[0] = lims[1] = atof(p);
	    if (!p[o]) pr->flags |= FLAG_SELECT_MAX;
	}
	p += o;
	if (*p) p++;
	if (*p) {	/* There is an upper limit */
	    pr->flags |= FLAG_SELECT_MAX;
	    lims[1] = atof(p);
	}

	/* Set now the limits according to the data type */
	switch(pr->form[0]) {
	  case 'A':
	    pr->flags |= FLAG_SELECT_char;
	    fprintf(stderr,
		    "++++Constraint only on numeric column: %s\n",
		    pr->type);
	    break;
	  case 'B':
	  case 'I':
	  case 'J':
	    if (pr->flags&FLAG_SELECT_MIN)
	        *((int *)&pr->min) = (int)lims[0];
	    if (pr->flags&FLAG_SELECT_MAX)
	        *((int *)&pr->max) = (int)lims[1];
	    break;
	  case 'K':
	    if (pr->flags&FLAG_SELECT_MIN)
	        *((int8*)&pr->min) = (int8)lims[0];
	    if (pr->flags&FLAG_SELECT_MAX)
	        *((int8*)&pr->max) = (int8)lims[1];
	    break;
	  case 'E':
	  case 'D':
	    pr->flags |= FLAG_SELECT_float;
	    if (pr->flags&FLAG_SELECT_MIN) pr->min = lims[0];
	    if (pr->flags&FLAG_SELECT_MAX) pr->max = lims[1];
	    break;
	  /* case 'L': */
	  default:
	    fprintf(stderr, "++++Invalid TFORM(%c) in constraint:%s\n",
		    pr->form[0], cst[qno]);
	    return(-1);
	}
    }
    return(0);
}


static int print_thead(FITSROW **table, int opt)
/*++++++++++++++++
.PURPOSE  Print the titles of a table
.RETURNS  Number of lines (1/2/3)
.REMARKS  opt=2 => print units, opt=1 => print line of dashes
-----------------*/
{
  FITSROW **ar, *pr;
  char *p;
  int n, nb;
  int lines = 0;
  int colno;

    /* In paragraph mode, issue the Name=Unit=fmt=explain */
    if (opar) {
	printf("#List of columns\n");
	printf("%-20s %-8s #%-3s %s\n", "#TTYPE", "TDISP", "Col", "TUNIT");
	for (ar=table, colno=0; *ar; ar++) {
	    pr = *ar;
	    if (pr->ano == 0) break;	/* Ending columns to ignore */
	    if (!pr->Cfmt[1]) continue;	/* No format -- ignore! */
	    printf("%-20s %-8s #%-3d", pr->type, pr->Cfmt, pr->no);
	    if (pr->unit) printf(" %s", pr->unit);
	    putchar('\n');
	    lines++;
	}
	/* CLASS (Gildas): MATRIX ==> list TYPES */
	if (mdim[0]) { int c, i; double crpix; FITSROW *pd;
	    printf("#---Constants or MAXIS values:");
	    for (i=0, c=' '; i<MAX_AXIS; i++) {
		if (mdim[i]>0) { printf("%c%d", c, mdim[i]); c = 'x'; }
	    }
	    putchar('\n');
	    for (i=0; (i<MAX_AXIS) && mdim[i]; i++) {
                crpix = crval[i] = 0 ;
                delta[i] = 1 ;
		c = iNAXIS1+i*perAXIS;
                if ((p=keys[c+1].value)) crpix    = *((double *)p) -1. ;
                if ((p=keys[c+2].value)) crval[i] = *((double *)p) ;
                if ((p=keys[c+3].value)) delta[i] = *((double *)p) ;
		pr = colword(table, keys[c+2].name+1); /* overrides CRVALi */
		pd = colword(table, keys[c+3].name+1); /* overrides CDELTi */
		if (!pd) crval[i] -= crpix*delta[i];   /* Value at pixel#0 */
		p  = keys[c+4].value; if (!p) p = keys[c].name+2; /* AXISi */
		if (pr) 	/* CRVALi is overridden */
		     printf("%-8s = %17s", p, pr->type);
		else printf("%-8s = %17g", p, crval[i]);
		if (pd) 	/* CDELTi is overridden */
		    printf(" + %s", pd->type);
		else 		/* Value of parameter from CRVAL */
		    printf(" + (%g)x(pix)", delta[i]);
		if ((p = keys[c+5].value)) printf(" %s", p); /* Unit */
		putchar('\n');
	    }
	    printf("#---Constants or MAXIS values: (End)\n");
	}
	putchar('\n');
	lines++;
	return(lines);
    }

    /* Write out the labels */
    for (ar=table, colno=0; *ar; ar++) {
	pr = *ar;
	if (pr->ano == 0) break;	/* Ending columns to ignore */
	if (!pr->Cfmt[1]) continue;	/* No format -- ignore! */
	if (delim && (colno>0) && (!tdisp)) putchar(*delim); colno++;

	p = pr->type;
	if (optb) { printf("%s", p); continue; }
	n = pr->len1a;			/* Number of bytes for 1 edited num. */
	if (pr->form[0] != 'A') 	/* Full length saved in pr->len1a */
	    n *= pr->count;
	nb = n - strlen(p);
	if (nb<0) {			/* Truncated label */
	    for (nb=0; nb<n; nb++) putchar(p[nb]);
	    continue;
	}
	if (pr->form[0] != 'A') while (nb>0) { putchar(' '); --nb; }
	printf("%s", p); n -= strlen(p);
	while (nb>0) { putchar(' '); --nb; }
    }
    putchar('\n'); fflush(stdout);
    lines++;

    if (opt&2) {
        /* Write out the units */
        for (ar=table, colno=0; *ar; ar++) {
	    pr = *ar;
	    if (pr->ano == 0) break;	/* Ending columns to ignore */
	    if (!pr->Cfmt[1]) continue;	/* No format -- ignore! */
	    if (delim && (colno>0) && (!tdisp)) putchar(*delim); colno++;
	    p = pr->unit;
	    if (optb) { if (p) printf("%s", p); continue; }
	    if (!p) p = optb ? "" : "---";
	    n = pr->len1a;
	    if (pr->form[0] != 'A') n *= pr->count;
	    nb = n - strlen(p);
	    if (nb<0) {			/* Truncated label */
	        for (nb=0; nb<n; nb++) putchar(p[nb]);
	        continue;
	    }
	    if (pr->form[0] != 'A') while (nb>0) { putchar(' '); --nb; }
	    printf("%s", p); n -= strlen(p);
	    while (nb>0) { putchar(' '); --nb; }
	}
    }
    putchar('\n'); fflush(stdout);
    lines++;

    if (opt&1) {
        /* Write out the separator line */
        for (ar=table, colno=0; *ar; ar++) {
	    pr = *ar;
	    if (pr->ano == 0) break;	/* Ending columns to ignore */
	    if (!pr->Cfmt[1]) continue;	/* No format -- ignore! */
	    if (delim && (colno>0) && (!tdisp)) putchar(*delim); colno++;
	    p = pr->type;		/* Title of column  */
	    if (optb) { 		/* Take same as label */
		while(*p) { putchar('-'), p++; }
		continue;
	    }
	    n = pr->len1a;
	    if (pr->form[0] != 'A') n *= pr->count;
	    nb = n - strlen(p);
	    while (--nb>=0) putchar(' ');
	}
    }
    putchar('\n'); fflush(stdout);
    lines++;

    return(lines);
}

static int print_fmt(FITSROW **arows)
/*++++++++++++++++
.PURPOSE  Edit the format file (for ReadMe)
.RETURNS  Number of fields
.REMARKS
-----------------*/
{
  FITSROW *pr, **ar;
  int i, ldel, nf=0, i2, i3;
  int w2=14, w3=24;	/* Width of Format + Unit + Label */
  char *u;

    /* Estimate w3 = 3 + len(disp) + len(unit) + len(label) */
    for (ar=arows; *ar; ar++) {
        pr = *ar;
	if (!pr->Cfmt[1]) continue;	/* No format -- ignore! */
	if (pr->ano == 0) break;
	i = 2 + strlen(pr->disp);
	if (pr->unit)
	     i += strlen(pr->unit);
	else i += 3;
	if (i>w2) w2 = i;
	i += 1 + strlen(pr->type);
	if (i>w3) w3 = i;
    }

    /* Write format for ReadMe */
    ldel = 0;	/* Number of delimiter bytes */
    nf = 0;
    printf("\n#%s\n#%s\n", "Format for the ReadMe",
    	                   "---------------------");
    for (ar=arows; *ar; ar++) {
        pr = *ar;
	if (!pr->Cfmt[1]) continue;	/* No format -- ignore! */
	++nf;
        if (pr->ano == 0) break;
        printf("#%dX ", ldel); ldel = strlen(delim);
        if (pr->count > 1)
    	    i = printf("%d", pr->count);
        else { putchar(' '); i = 1; }
        i += printf("%s ", pr->disp);
	u = pr->unit ? pr->unit : "---";
	i2 = i + strlen(u);
	i3 = i2 + 1 + strlen(pr->type);
	while ((i3<w3) && (i2<w2) && (i<9)) { putchar(' '); i++; i2++; i3++; }
        i += printf("%s", u);
	while ((i3<w3) && (i<w2)) { putchar(' '); i++; i3++; }
        i += printf(" %s", pr->type);
        while (i<w3) { putchar(' '); i++; }
	putchar(' ');
	i = 0;				/* Set to 1 if blank needed */
        if (pr->form[0] == 'L') { printf("[0,1]"); i=1; }
	else if (pr->flags&FLAG_CONSTANT) {
	    switch(pr->form[0]) {
              case 'A': printf("[%s]", pr->value ? pr->value : " "); break;
	      case 'B': case 'J':
	      case 'I': printf("[%d]", *(int *)&pr->max); break;
#if int8l == 1
	      case 'K': printf("[%ld]", *(int8 *)&pr->max); break;
#else
	      case 'K': printf("[%lld]", *(int8 *)&pr->max); break;
#endif
	      default:  printf("[%g]", pr->max); break;
	    }
	    i = 1;			/* Must be followed by a blank */
	}
        if (pr->cnul && (pr->form[0] != 'A')) {
	    i = 1;
    	    putchar('?');
    	    if (pr->null && isgraph(pr->null[0]))
    	        printf("=%s", pr->null);
        }
	if (i) putchar(' ');
	if (pr->comment) printf("%s", pr->comment);
	else printf("Column#%d", nf);
	printf(" (%d)", pr->no);
        putchar('\n');
    }
    printf("#%s\n", "---------------------");
    return(nf);
}

static int bintable(FILE *in)
/*++++++++++++++++
.PURPOSE  Process a BINARY TABLE
.RETURNS  0 / 2 (truncated)
.REMARKS
-----------------*/
{
  static char bdisp[] = "LLll";		/* TDISP letter for boolean */
  static char abool[] = "FTft01";	/* how to edit false / true */
  FITSROW *pr, **ar;
  char *record, *p, *e, *a;
  int reclen, records, edited, colno, i, n, o, err=0;
  short I; int J, *aJ;
  int8  K, *aK;
  float E; double D;
  char tmin[60], tmax[60], *amin, *amax, *sep=delim;
  /* int4 here ; */
#if int8l == 1
  static char fmt8[] = "%21ld";
#else
  static char fmt8[] = "%21lld";
#endif

    if (!keys[iTFIELDS].value) {
        fprintf(stderr, "****Missing TFIELDS value for a BINTABLE: %s\n",
	  input_name);
	return(-1);
    }
    if (!keys[iNAXIS1].value) {
        fprintf(stderr, "****Missing NAXIS1 value for a BINTABLE: %s\n",
	  input_name);
	return(-1);
    }
    if (!keys[iNAXIS2].value) {
        fprintf(stderr, "****Missing NAXIS2 value for a BINTABLE: %s\n",
	  input_name);
	return(-1);
    }
    reclen  = *(int *)(keys[iNAXIS1].value);
    records = *(int *)(keys[iNAXIS2].value);
    total_lines = records;
    if (opar) sep = "\n";
    if (!sep) sep = " ";
    /* here    = lseek(in, 0L, 1); */

    if (arows) Free(arows);
    ar = arows = (FITSROW **)Malloc(n=allorow*sizeof(FITSROW *));
    memset(arows, 0, n);

    /* V2.94: Gildas stuff (MAXIS) */
    memset(mdim, 0, sizeof(mdim));
    if (keys[iMAXIS].value) { int c; double crpix;
	n = *(int *)(keys[iMAXIS].value);
	if (n>MAX_AXIS) n = MAX_AXIS;
	for (i=0; i<n; i++) {
	    mdim[i] = 1;
	    if (!keys[iMAXIS1+i].value) fprintf(stderr,
		    "#***fits2a: missing %s value, assume 1\n",
		    keys[iMAXIS1+i].name+1);
	    else mdim[i] = *(int *)(keys[iMAXIS1+i].value);
	    crpix = crval[i] = 0;
	    delta[i] = 1;
	    c = iNAXIS1+i*perAXIS;
	    if ((p=keys[c+1].value)) crpix    = *((double *)p) -1.;
	    if ((p=keys[c+2].value)) crval[i] = *((double *)p) ;
	    if ((p=keys[c+3].value)) delta[i] = *((double *)p) ;
	    tcrval[i] = colword(arows, keys[c+2].name+1); /* overrides CRVALi */
	    tdelta[i] = colword(arows, keys[c+3].name+1); /* overrides CDELTi */
	    if (!tdelta[i]) crval[i] -= crpix*delta[i];   /* Value at pixel#0 */
	    p  = keys[c+4].value; if (!p) p = keys[c].name+2;
	    tparam[i] = p;
	}
    }

    /* Define the length and position of all elements */
    for (pr=fitsrow, o=0; pr->form[0]; pr++) {
	*(ar++) = pr;
	pr->opos = o;
	switch(pr->form[0]) {
	  case 'A':
	    pr->len1 = 1;
	    if (!pr->disp[0]) sprintf(pr->disp, "A%d", pr->count);
	    if (!(pr->flags&FLAG_SELECT)) {
	        *((int *)&pr->min) =  0x7fffffff;
	        *((int *)&pr->max) = -0x7fffffff;
	    }
	    if (!tdisp) { D = pr->min; pr->min = pr->max; pr->max = D; }
	    break;
	  case 'B':
	    pr->len1 = 1;
	    if (!pr->disp[0]) strcpy(pr->disp, "I3");
	    if (pr->null)  *((int *)&pr->nul) = atoi(pr->null);
	    if (!(pr->flags&FLAG_SELECT)) {
	        *((int *)&pr->min) = 0xffff;
	        *((int *)&pr->max) = 0;
	    }
	    if (!tdisp) { D = pr->min; pr->min = pr->max; pr->max = D; }
	    break;
	  case 'I':
	    pr->len1 = 2;
	    if (!pr->disp[0]) strcpy(pr->disp, "I6");
	    if (pr->null)  *((int *)&pr->nul) = atoi(pr->null);
	    if (!(pr->flags&FLAG_SELECT)) {
	        *((int *)&pr->min) =  0x7fffffff;
	        *((int *)&pr->max) = -0x7fffffff;
	    }
	    if (!tdisp) { D = pr->min; pr->min = pr->max; pr->max = D; }
	    break;
	  case 'J':
	    pr->len1 = 4;
	    if (!pr->disp[0]) strcpy(pr->disp, "I11");
	    if (pr->null)  *((int *)&pr->nul) = atoi(pr->null);
	    if (!(pr->flags&FLAG_SELECT)) {
	        *((int *)&pr->min) =  0x7fffffff;
	        *((int *)&pr->max) = -0x7fffffff;
	    }
	    if (!tdisp) { D = pr->min; pr->min = pr->max; pr->max = D; }
	    break;
	  case 'K':
	    pr->len1 = 8;
	    if (!pr->disp[0]) strcpy(pr->disp, "I20");
	    if (pr->null)  *((int8 *)&pr->nul) = atoll(pr->null);
	    if (!(pr->flags&FLAG_SELECT)) {
	        *((int8 *)&pr->min) =  0x7fffffffffffffffLL;
	        *((int8 *)&pr->max) = -0x7fffffffffffffffLL;
	    }
	    if (!tdisp) { D = pr->min; pr->min = pr->max; pr->max = D; }
	    break;
	  case 'D':
	    pr->len1 = 8;
	    if (!pr->disp[0]) {
		strcpy(pr->Cfmt, "%24.16e");
		strcpy(pr->disp, "D24.17");
	    }
	    if (pr->null)  pr->nul = atof(pr->null);
	    if (!(pr->flags&FLAG_SELECT)) {
	        pr->min = 1.e99;
		pr->max = -pr->min;
	    }
	    if (!tdisp) { D = pr->min; pr->min = pr->max; pr->max = D; }
	    break;
	  case 'E':
	    pr->len1 = 4;
	    if (!pr->disp[0]) {
		strcpy(pr->Cfmt, "%13.6e");
		strcpy(pr->disp, "E13.6");
	    }
	    if (pr->null)  pr->nul = atof(pr->null);
	    if (!(pr->flags&FLAG_SELECT)) {
	        pr->min = 1.e38;
		pr->max = -pr->min;
	    }
	    if (!tdisp) { D = pr->min; pr->min = pr->max; pr->max = D; }
	    break;
	  case 'L':
	    if (!pr->disp[0]) strcpy(pr->disp, "L1");
	    pr->len1 = 1;
	    if (!pr->disp[0]) pr->disp[0]=pr->form[0], pr->disp[1] = 0;
	    break;
	  default:
	    fprintf(stderr, "++++Invalid TFORM: %c\n", pr->form[0]);
	    return(-1);
	}
	o += pr->len1*pr->count;
    }
    *ar = (FITSROW *)0;

    /* Choose and order the columns if asked */
    i = sort_cols(arows, acols);
    if (i == 0) fprintf(stderr,
	"#***No column to display from input table!\n");
    if (i<=0) return(i);

    /* Install the Query on Cols options */
    qopt_cols(arows, query);

    /* Convert the Formats into C ones */
    if (tdisp) opar = 0;	/* No edition */
    for (ar=arows; *ar; ar++) {
	pr = *ar;
	i = (ar - arows);
	    		/* Use format specified as arguments */
        if (fmt[i] && (sfmt[i]&2)) { free(fmt[i]); sfmt[i] = 0; }
	if ((sfmt[i]&1) && fmt[i]) {
	    memset(pr->Cfmt, 0, sizeof(pr->Cfmt));
	    strncpy(pr->Cfmt, fmt[i], sizeof(pr->Cfmt)-1);
	}
	/* printf(".... %d->%d %s\n", pr->no, pr->ano, pr->disp); */
	set_Cform(pr);
	err |= check_Cform(pr);	/* Added V2.95 */
    }
    if (err) exit(1);

    if(ohead && (!tdisp)) print_thead(arows, 3);
    else if (opar) putchar('\n');	/* Empty header paragraph */

    /* One record */
    record = Malloc(reclen);
    for (recno=edited=0; recno < records; recno++) {

	if ((!tdisp) && (recno >= max_recno)) break;
	if (edited >= top_lines) break;
	if(read_record(in, record, reclen)<0)
	    return(-1);
	if (recno < nskip) continue;	/* In case of -skip...	*/

	/* Version 2.0: Check column */
	if (query[0]) { int ok = 1;
	    if (oqs&2) {		/* Sorted query, max found */
		return(0);		/* No message but not read */
	    }
	    oqs &= ~3;			/* <=min => 1, >max => 2   */
            for (ar=arows, colno=i=0; *ar && ok && query[i]; ar++) {
	        pr = *ar;
	        if (pr->ano == 0) break;
		if ((pr->flags&FLAG_SELECT) == 0) continue;
		i++;
		p = record + pr->opos;
		switch(pr->form[0]) {
		  case 'B': J = *p&0xff;
		    break;
		  case 'I': memcpy(&I, p, sizeof(I));
		    if (swapping) swap1(&I, 1);
		    J = I;
		    break;
		  case 'J': memcpy(&J, p, sizeof(J));
		    if (swapping) swap(&J, 1);
		    break;
		  case 'D':
		    memcpy(&D, p, sizeof(D));
		    if (swapping) swap8(&D,1);
		    break;
		  case 'E': memcpy(&E, p, sizeof(E));
		    if (swapping) swap((int *)&E,1);
		    D = E;
		    break;
		  case 'K':
		    memcpy(&K, p, sizeof(K));
		    if (swapping) swap8(&K,1);
		    break;
		}
		if (pr->flags&FLAG_SELECT_float) {
		    if (oqs == 0x10) {	/* Sorted query: mask 2=end */
			if (D > pr->max) oqs |= 2;
			if (D < pr->min) oqs |= 1;
		    }
		    if (pr->flags&FLAG_SELECT_MIN) ok &= (D >= pr->min);
		    if (pr->flags&FLAG_SELECT_MAX) ok &= (D <= pr->max);
		}
		else if (pr->form[0] == 'K') {	/* 64-bit */
		    if (oqs == 0x10) {	/* Sorted query: mask 2=end */
			if (K > *((int8 *)&pr->max)) oqs |= 2;
			if (K < *((int8 *)&pr->min)) oqs |= 1;
		    }
		    if (pr->flags&FLAG_SELECT_MIN) ok &=
			(K >= *((int8 *)&pr->min));
		    if (pr->flags&FLAG_SELECT_MAX) ok &=
			(K <= *((int8 *)&pr->max));
		}
		else {		/* integer */
		    if (oqs == 0x10) {	/* Sorted query: mask 2=end */
			if (J > *((int *)&pr->max)) oqs |= 2;
			if (J < *((int *)&pr->min)) oqs |= 1;
		    }
		    if (pr->flags&FLAG_SELECT_MIN) ok &=
			(J >= *((int *)&pr->min));
		    if (pr->flags&FLAG_SELECT_MAX) ok &=
			(J <= *((int *)&pr->max));
		}
		if (pr->flags&FLAG_SELECT_NOT) ok ^= 1;
	    }
	    if (!ok) continue;
	}

        for (ar=arows, colno=0; *ar; ar++) {
	    pr = *ar;
	    if (pr->ano == 0) break;
	    if (!pr->Cfmt[1]) continue;	/* Column to ignore! */
	    if (sep/*d*/ && (colno>0) && (!tdisp)) putchar(*sep/*d*/); colno++;
	    p = record + pr->opos;
	    if(opar) printf("%s=", pr->type);

	    switch(pr->form[0]) {
	      case 'A':
		if (tdisp) {		 /* Count the longest string */
		    for (i=pr->len1a-1; (i>=0) && isspace(p[i]); i--) ;
		    ++i;
		    aJ = (int *)&pr->min;
		    if (i < *aJ) *aJ = i;
		    aJ = (int *)&pr->max;
		    if (i > *aJ) *aJ = i;
		    if (i == 0) pr->cnul += 1;
		    else {	/* Non-blank string:
		            Set the Constant flag for constant text */
			if (pr->value) {
			    if (pr->flags&FLAG_CONSTANT) {
				if (strncmp(pr->value, p, pr->count))
				    pr->flags &= ~FLAG_CONSTANT;
			    }
			}
			else {
			    pr->value = Malloc(pr->count+1);
			    strncpy(pr->value, p, pr->count);
			    pr->value[pr->count] = 0;
			    pr->flags |= FLAG_CONSTANT;
			}
			i = pr->count - i;	/* #tail blanks */
			if (i < ITEMS(pr->cdec)) pr->cdec[i] += 1;
		    }
		}
		else {
		    for (i = *(int *)&pr->min, e=p; --i>=0; e++)
			putchar(isprint(*e) ? *e : ' ');
		    for (i = *(int *)&pr->nul; --i>=0; ) {
			if (*e == ' ') continue;
			if (*e) fprintf(stderr,
			  "++++Non-blanks line #%d '%s' col.%s: %c at +%d\n",
			    recno+1, pr->disp, pr->type, *e, (int)(e-p));
		    }
		    /* Add the blanks if needed */
		    if (optb); else
		    for (i = *(int *)&pr->max - *(int *)&pr->min; --i>=0; e++)
			putchar(' ');
		}
	        break;
	      case 'B':
		for (i=pr->count, e=p; --i>=0; e++) {
		    I = *e&0xff;
		    aJ = (int *)&pr->nul;
		    if (pr->null && (I == *aJ)) {
			if (tdisp) pr->cnul += 1;
			else print_null(pr);
		    }
		    else if (tdisp) {
		        aJ = (int *)&pr->min;
			if (I < *aJ) *aJ = I;
		        aJ = (int *)&pr->max;
			if (I > *aJ) *aJ = I;
		    }
		    else printf(pr->Cfmt, I);
		}
	        break;
	      case 'I':
		for (i=pr->count, e=p; --i>=0; e += sizeof(I)) {
		    memcpy(&I, e, sizeof(I));
		    if (swapping) swap1(&I, 1);
		    aJ = (int *)&pr->nul;
		    if (pr->null && (I == *aJ)) {
			if (tdisp) pr->cnul += 1;
			else print_null(pr);
		    }
		    else if (tdisp) {
			J = I;
			if (pr->flags&(FLAG_TRUNC_MIN|FLAG_TRUNC_MAX))
			    J = bound_J(pr, J);
		        aJ = (int *)&pr->min;
			if (J < *aJ) *aJ = J;
		        aJ = (int *)&pr->max;
			if (J > *aJ) *aJ = J;
		    }
		    else printf(pr->Cfmt, bounded_J(pr, I));
		    if (i && (!tdisp)) putchar(' ');
		}
	        break;
	      case 'J':
		for (i=pr->count, e=p; --i>=0; e += sizeof(J)) {
		    memcpy(&J, e, sizeof(J));
		    if (swapping) swap(&J, 1);
		    aJ = (int *)&pr->nul;
		    if (pr->null && ( J == *aJ)) {
			if (tdisp) pr->cnul += 1;
			else print_null(pr);
		    }
		    else if (tdisp) {
			if (pr->flags&(FLAG_TRUNC_MIN|FLAG_TRUNC_MAX))
			    J = bound_J(pr, J);
		        aJ = (int *)&pr->min;
			if (J < *aJ) *aJ = J;
		        aJ = (int *)&pr->max;
			if (J > *aJ) *aJ = J;
		    }
		    else printf(pr->Cfmt, bounded_J(pr, J));
		    if (i && (!tdisp)) putchar(' ');
		}
	        break;
	      case 'K':	/* Added V2.91 */
		for (i=pr->count, e=p; --i>=0; e += sizeof(K)) {
		    memcpy(&K, e, sizeof(K));
		    if (swapping) swap8((double *)&K, 1);
		    aK = (int8 *)&pr->nul;
		    if (pr->null && ( K == *aK)) {
			if (tdisp) pr->cnul += 1;
			else print_null(pr);
		    }
		    else if (tdisp) {
			if (pr->flags&(FLAG_TRUNC_MIN|FLAG_TRUNC_MAX))
			    K = bound_K(pr, K);
		        aK = (int8 *)&pr->min;
			if (K < *aK) *aK = K;
		        aK = (int8 *)&pr->max;
			if (K > *aK) *aK = K;
		    }
		    else printf(pr->Cfmt, bounded_K(pr, K));
		    if (i && (!tdisp)) putchar(' ');
		}
	        break;
	      case 'D':
		for (i=pr->count, e=p; --i>=0; e += sizeof(D)) {
		    memcpy(&D,  e, sizeof(D));
		    if (swapping) swap8(&D,1);
		    if (pr->null && (pr->nul == D)) D = 0./0.;
		    if (!isFinite(D)) {
			if (tdisp) pr->cnul += 1;
			else print_null(pr);
		    }
		    else if (tdisp) {
			if (pr->flags&(FLAG_TRUNC_MIN|FLAG_TRUNC_MAX))
			    D = bound_F(pr, D);
		        if (D < pr->min) pr->min = D;
		        if (D > pr->max) pr->max = D;
			pr->cdec[decimals(D, 15)] += 1;
		    }
		    else {
			if (pr->min < pr->max)
			     n = printf(pr->Cfmt, bounded_F(pr, D));
			else n = printf(pr->Cfmt, D);
		    }
		    if (i && (!tdisp)) putchar(' ');
		}
	        break;
	      case 'E':
		for (i=pr->count, e=p; --i>=0; e += sizeof(E)) {
		    memcpy(&E,  e, sizeof(E));
		    if (swapping) swap((int *)&E,1);
		    if (pr->null && (pr->nul == E)) E = 0./0.;
		    if (!isFinite(E)) {
			if (tdisp) pr->cnul += 1;
			else print_null(pr);
		    }
		    else if (tdisp) {
			D = E;
			if (pr->flags&(FLAG_TRUNC_MIN|FLAG_TRUNC_MAX))
			    D = bound_F(pr, D);
		        if (D < pr->min) pr->min = D;
		        if (D > pr->max) pr->max = D;
			pr->cdec[decimals(E, 6)] += 1;
		    }
		    else {
			if (pr->min < pr->max)
			     n = printf(pr->Cfmt, bounded_F(pr, E));
			else n = printf(pr->Cfmt, E);
		    }
		    if (i && (!tdisp)) putchar(' ');
		}
	        break;
	      case 'L':
		/* Choice of the boolean edition: */
		a = abool + strloc(bdisp, pr->disp[0]);
		for (i=pr->count, e=p; --i>=0; e++) {
		    if (isalpha(*e))
		         I = toupper(*e) == 'T' ? 1 : 0;
		    else I = *e != 0;
		    if (tdisp) {
			I &= 1;
		        aJ = (int *)&pr->min;
			if (I < *aJ) *aJ = J;
			aJ = (int *)&pr->max;
			if (I > *aJ) *aJ = J;
			pr->cdec[I] += 1;
		    }
		    else putchar(a[I]);
		}
	        break;
	      default:
		fprintf(stderr,
		    "#***data type not yet possible: %c (%s)\n",
		    pr->form[0], pr->type);
	    }
	}
	if (tdisp);
	else { putchar('\n'); if(opar) putchar('\n'); ++edited; }
    }

#if 0
    if (top_lines && (edited > top_lines)) {
	printf("#!+++Stopped after %d lines edited (read = %d lines)\n",
		top_lines, recno);
	return(2);
    }
#else
    if (top_lines) ;
#endif
    else if (max_lines && (recno < records)) {	/* Truncation !!! */
	printf("#!+++Stopped after %d lines read (total is %d lines)\n",
	        max_lines, records) ;
	return(2);
    }

    if (tdisp) { 		/* Edit the Format */
	char notes[120], *pn;
	/* Guess the Format */
        for (ar=arows; *ar; ar++) {
	    pr = *ar; tmin[0] = tmax[0] = 0;
	    if (pr->ano == 0) break;
	    if (verbop) printf("#%6d %c %-20s ",
		    pr->no, pr->form[0], pr->type);
	    reclen = 15;	/* Decimals for D */
	    notes[0] = 0;	/* Notes about TDISP */
	    pn = notes;
	    if (pr->count>1) {
		sprintf(pn, "[%d]", pr->count);
		pn += strlen(pn);
		if (pr->tdim) {	/* Multidim */
		    sprintf(pn, "=%s", pr->tdim);
		    pn += strlen(pn);
		}
		else *(pn++) = 'x';
		*(pn++) = ' ';
		*pn = 0;
	    }
	    switch(pr->form[0]) {
	      case 'A':
		pn = notes; *pn = 0;	/* No multiplicity */
		if (verbop) {
		    printf("%12d %12d NULs=%-10d-------#\n#%8s",
		      *(int *)&pr->min, *(int *)&pr->max, pr->cnul, "");
		    for(i=0; i<ITEMS(pr->cdec); i++) {
			if (pr->cdec[i]) printf(" %d=%d",
			  pr->count-i, pr->cdec[i]);
		    }
		    printf(" tail blanks\n");
		}
		J = *(int *)&pr->max;
		sprintf(pr->Cfmt, "A%d", J);
		if (J == 0) {	/* Always empty */
		    strcpy(pn, "always blank");
		    pr->flags |= FLAG_CONSTANT;
		}
		else if (pr->flags&FLAG_CONSTANT) {
		    strcpy(pn, "always '");
		    strncpy(pn, pr->value, sizeof(notes)-strlen(notes));
		    pn[sizeof(notes)-2] = 0;
		    strcpy(pn+strlen(pn), "'");
		}
		else if (J < pr->count)
		    sprintf(pn, "shorter than claimed A%d", pr->count);
	        break;
	      case 'L':
		if (verbop) {
		    printf("%10dx0 %10dx1 NULs=%-10d-------#\n",
		      pr->cdec[0], pr->cdec[1], pr->cnul);
		}
		sprintf(pr->Cfmt, "I%d", pr->len1);
		/* Add count of true / false */
		sprintf(pn, "0(f)=%d 1(t)=%d", pr->cdec[0], pr->cdec[1]);
		pn += strlen(pn);
		if (pr->cnul) sprintf(pn, " (?)=%d", pr->cnul);
	        break;

	      case 'B':
	      case 'I': case 'J':
		if (verbop) {
		    sprintf(tmin, "%12d", *(int *)&pr->min);
		    sprintf(tmax, "%12d", *(int *)&pr->max);
		    printf("%12s %12s NULs=%-10d-------#\n",
		      tmin, tmax, pr->cnul);
		    if (pr->trunc[0]|pr->trunc[1]) printf("#%54s: %d/%d\n",
			    "Truncated", pr->trunc[0], pr->trunc[1]);
		}
		i = *(int *)&pr->min ;
		n = *(int *)&pr->max;
		if (pr->cnul) {
		    sprintf(pn, "NuL=%d ", pr->cnul);
		    pn += strlen(pn);
		}
		if (i == n) {	/* Constant */
		    sprintf(pn, "constant(%d)", n);
		    pr->flags |= FLAG_CONSTANT;
		    pn += strlen(pn);
		}
		else if (i<n) {
		    sprintf(pn, "range=(%d,%d)", i, n);
		    pn += strlen(pn);
		}
		else strcpy(pn, " (void?)");
		pn += strlen(pn); 
		if (pr->trunc[0]|pr->trunc[1]) sprintf(pn, "  truncated=%d/%d",
			pr->trunc[0], pr->trunc[1]);
		if (i < 0) {
		    if (n < -i) n = -i;
		    i = 2;
		}
		else i = 1;
		while (n >= 10) { i++; n /= 10; }
		sprintf(pr->Cfmt, "I%d", i);
	        break;

	      case 'K': 
		sprintf(tmin, fmt8, *(int8 *)&pr->min);
		sprintf(tmax, fmt8, *(int8 *)&pr->max);
		amin = ltrim(tmin); amax = ltrim(tmax);
		if (verbop) {
		    printf("%12s %12s NULs=%-10d-------#\n",
		      "", "", pr->cnul);
		    if (pr->trunc[0]|pr->trunc[1]) printf("#%54s: %d/%d\n",
			    "Truncated", pr->trunc[0], pr->trunc[1]);
		}
		if (pr->cnul) {
		    sprintf(pn, "NuL=%d ", pr->cnul);
		    pn += strlen(pn);
		}
		if (pr->min == pr->max) {	/* Constant */
		    sprintf(pn, "constant(%s)", amin);
		    pr->flags |= FLAG_CONSTANT;
		    pn += strlen(pn);
		}
		else if (*(int8 *)&pr->min > *(int8 *)&pr->max)
		    strcpy(pn, " (void?)");
		else {
		    sprintf(pn, "range=(%s,%s)", amin, amax);
		    pn += strlen(pn);
		}
		pn += strlen(pn); 
		if (pr->trunc[0]|pr->trunc[1]) sprintf(pn, "  truncated=%d/%d",
			pr->trunc[0], pr->trunc[1]);
		i = strlen(amin);
		n = strlen(amax);
		if (n>i) i=n;
		sprintf(pr->Cfmt, "I%d", i);
	        break;

	      case 'E': reclen = 6;
	      case 'D':
		if (verbop) { 	/* Mod. V3.2 */
		    sprintf(tmin, "%25.16e", pr->min); 
		    sprintf(tmax, "%25.16e", pr->max); 
		    f4fits(tmin, pr->form[0]);	/* Edit correctly (e --> D) */
		    f4fits(tmax, pr->form[0]);
		    printf(
		        "%-25s NULs=%-10d-------#\n#%29s %-25s Decimals:\n#%8s",
		        tmin, pr->cnul, "", tmax, "");
		}
		for (n=ITEMS(pr->cdec)-1; (n>0) && pr->cdec[n] == 0; n--);
		if (verbop) {
		    for(i=0; i<=n; i++) printf(" %d=%d", i, pr->cdec[i]);
		    putchar('\n');
		    if (pr->trunc[0]|pr->trunc[1]) printf("#%54s: %d/%d\n",
			    "Truncated", pr->trunc[0], pr->trunc[1]);
		}
		if (pr->cnul) {
		    sprintf(pn, "NaN=%d ", pr->cnul);
		    pn += strlen(pn);
		}
		if (pr->min >= pr->max) {
		    pr->flags |= FLAG_CONSTANT;
		    sprintf(pn, "constant(%g)",
			pr->min == pr->max ? pr->min : 0./0.);
		    pn += strlen(pn);
		}
		else {
		    sprintf(pn, "range=(%g,%g)", pr->min, pr->max);
		    pn += strlen(pn);
		}
		pn += strlen(pn); 
		if (pr->trunc[0]|pr->trunc[1]) sprintf(pn, "  truncated=%d/%d",
			pr->trunc[0], pr->trunc[1]);
		D = pr->max;
		if ((pr->min < 0) && (D < -pr->min)) D = -pr->min;
		if ((n == 0) && (D<=dexp[reclen]) &&
		   ((D==0) || (D>=0.1))) {	/* It's an integer number ! */
		    i = D;
		    i = digits(i);
		    if (pr->min < 0) i++;
		    sprintf(pr->Cfmt, "I%d", i);
		}
		else if (((pr->max - pr->min) < 1000) && (n<=6)
			&& ((D==0) || (D>=0.1))) {
		    i = D;
		    i = digits(i);	     /* Digits for integer part */
		    if (pr->min < 0) i++;    /* A minus is required ... */
		    sprintf(pr->Cfmt, "F%d.%d", 1+i+n, n);
		}
		else sprintf(pr->Cfmt, "%c%d.%d", pr->form[0],
		    reclen+6+(reclen>7), reclen);
	        break;
	    }
	    printf("TDISP%-3d= '%s'", pr->no, pr->disp);
	    for (i=strlen(pr->disp); i<9; i++) putchar(' ');
	    printf(" / %-8s", pr->Cfmt);
	    if (notes[0]) {
		/*for (i=strlen(pr->disp); i<9; i++) putchar(' '); */
		printf(" %s", notes);
	    }
	    /* V3.2: Edit also the NULL, TTMIN and TTMAX */
	    if (pr->null) printf("\nTNULL%-3d= '%s'  / NULL value as specified",
		    pr->no, pr->null);
	    if (pr->flags&FLAG_TRUNC_MIN) printf(
		    "\nTTMIN%-3d= '%g'  / Min Value as specified",
		    pr->no, pr->tmin);
	    else if (verbop && tmin[0]) printf(
		    "\nTTMIN%-3d= '%s'  / Min Value (computed)",
		    pr->no, tmin);
	    if (pr->flags&FLAG_TRUNC_MAX) printf(
		    "\nTTMAX%-3d= '%g'  / Max Value as specified",
		    pr->no, pr->tmax);
	    else if (verbop && tmax[0]) printf(
		    "\nTTMAX%-3d= '%s'  / Max Value (computed)",
		    pr->no, tmax);
	    printf("\n%s\n", ruler);
	    /*putchar('\n');*/
	}
        if (verbop) printf("%s\n\n", sep_disp); /* putchar('\n'); */
	print_fmt(arows);
    }
    else if (ometa) {
	print_fmt(arows);
    }

    Free(record);
    if ((n = (reclen*records)%2880)) {	/* Terminate Block */
        n = 2880-n;
	record = Malloc(n);
	fread(record, 1, n, in);
	Free(record);
    }
    return(0);
}

static int asctable(FILE *in)
/*++++++++++++++++
.PURPOSE  Process an ASCII TABLE
.RETURNS  0 / 2 (truncated)
.REMARKS  Added V1.7
-----------------*/
{
  FITSROW *pr, **ar;
  char *record, *p, *e, *sep=delim, x;
  int reclen, records, edited, colno, i, n, err=0;
  int errors = 0, isnull;
  int    ival, *aJ;
  double fval, *aD, D;
  /* int4 here ; */

    if (!keys[iTFIELDS].value) {
        fprintf(stderr, "****Missing TFIELDS value for a BINTABLE: %s\n",
	  input_name);
	return(-1);
    }
    if (!keys[iNAXIS1].value) {
        fprintf(stderr, "****Missing NAXIS1 value for a BINTABLE: %s\n",
	  input_name);
	return(-1);
    }
    if (!keys[iNAXIS2].value) {
        fprintf(stderr, "****Missing NAXIS2 value for a BINTABLE: %s\n",
	  input_name);
	return(-1);
    }
    reclen  = *(int *)(keys[iNAXIS1].value);
    records = *(int *)(keys[iNAXIS2].value);
    total_lines = records;
    if (opar) sep = "\n";
    if (!sep) sep = " ";
    /* here    = lseek(in, 0L, 1); */

    if (arows) Free(arows);
    ar = arows = (FITSROW **)Malloc(n=allorow*sizeof(FITSROW *));
    memset(arows, 0, n);

    /* Verify the length and position of all elements */
    for (pr=fitsrow; pr->form[0]; pr++) {
	*(ar++) = pr;
	if ((pr->opos<1) || (pr->opos>reclen)) {
	    fprintf(stderr, "****Invalid TBCOL%d, not in range [1,%d]: %d\n",
		pr->no, reclen, pr->opos);
	    errors++;
	}
	pr->opos -= 1;		/* Transform into position offset */
	if ((pr->len1<1) || ((pr->len1+pr->opos)> reclen)) {
	    fprintf(stderr, "****Invalid TFORM%d, indicates length=%d\n",
		pr->no, pr->len1);
	    errors++;
	}
	if(tdisp) {
	    pr->flags |= FLAG_CONVERT;
	    /* Simplify the NULL string: remove blanks */
	}
	if ((p = pr->null)) {
	    for (i=strlen(p)-1; (i>=0) && p[i] == ' '; i--) ;
	    p[++i] = 0;
	    for (i=0; p[i] == ' '; i++) ;
	    if (i && p[i]) {
		strcpy(p, p+i);
	    }
	}
	switch(pr->form[0]) {
	  case 'A':
	    if (!(pr->flags&FLAG_SELECT)) {
	        *((int *)&pr->min) =  0x7fffffff;
	        *((int *)&pr->max) = -0x7fffffff;
	    }
	    if (!tdisp) { D = pr->min; pr->min = pr->max; pr->max = D; }
	    break;
	  case 'I':
	    if (!(pr->flags&FLAG_SELECT)) {
	        *((int *)&pr->min) =  0x7fffffff;
	        *((int *)&pr->max) = -0x7fffffff;
	    }
	    if (!tdisp) { D = pr->min; pr->min = pr->max; pr->max = D; }
	    break;
	  case 'F':
	  case 'D':
	    if (!(pr->flags&FLAG_SELECT)) {
	        pr->min = 1.e99;
		pr->max = -pr->min;
	    }
	    if (!tdisp) { D = pr->min; pr->min = pr->max; pr->max = D; }
	    break;
	  case 'E':
	    if (!(pr->flags&FLAG_SELECT)) {
	        pr->min = 1.e38;
		pr->max = -pr->min;
	    }
	    if (!tdisp) { D = pr->min; pr->min = pr->max; pr->max = D; }
	    break;
	  default:
	    fprintf(stderr, "****Invalid TFORM%d: %c\n", pr->form[0], pr->no);
	    errors++;
	}
    }
    *ar = (FITSROW *)0;
    if (errors) return(-1);

    /* Choose and order the columns if asked */
    i = sort_cols(arows, acols);
    if (i == 0) fprintf(stderr,
	"#***No column to display from input table!\n");
    if (i<=0) return(i);

    /* Install the Query on Cols options */
    qopt_cols(arows, query);

    /* Convert the Formats into C ones */
    if (tdisp) opar = 0;	/* No edition */
    for (ar=arows; *ar; ar++) {
	pr = *ar;
	i = (ar - arows);
	    		/* Use format specified as arguments */
        if (fmt[i] && (sfmt[i]&2)) { free(fmt[i]); sfmt[i] = 0; }
	if ((sfmt[i]&1) && fmt[i]) {
	    memset(pr->Cfmt, 0, sizeof(pr->Cfmt));
	    strncpy(pr->Cfmt, fmt[i], sizeof(pr->Cfmt)-1);
	    pr->flags |= FLAG_CONVERT;	/* Requires a conversion */
	}
	set_Cform(pr);
	err |= check_Cform(pr);	/* Added V2.95 */
    }
    if (err) exit(1);

    if(ohead) print_thead(arows, 3);
    else if (opar) putchar('\n');	/* Empty header paragraph */

    /* One record */
    record = Malloc(reclen+1);
    for (recno=edited=0; recno < records; recno++) {

	if ((!tdisp) && (recno >= max_recno)) break;
	if (edited > top_lines) break;
	if (edited > max_lines) break;
	if(read_record(in, record, reclen)<0)
	    return(-1);
	if (recno < nskip) continue;	/* In case of -skip */
	record[reclen] = 0;				/* End-Of-String */

	/* Version 2.0: Check column */
	if (query[0]) { int ok = 1;
	    if (oqs&2) {		/* Sorted query, max found */
		return(2);		/* No message but not read */
	    }
	    oqs &= ~3;			/* <=min => 1, >max => 2   */
            for (ar=arows, colno=i=0; *ar && ok && query[i]; ar++) {
	        pr = *ar;
	        if (pr->ano == 0) break;
		if ((pr->flags&FLAG_SELECT) == 0) continue;
		i++;
		p = record + pr->opos;
		i = pr->len1;
		x = p[i]; p[i] = 0;
		if (pr->flags&FLAG_SELECT_float) { double v;
		    v = atof(p);
		    if (oqs == 0x10) {	/* Sorted query: mask 2=end */
			if (v > pr->max) oqs |= 2;
			if (v < pr->min) oqs |= 1;
		    }
		    if (pr->flags&FLAG_SELECT_MIN) ok &= (v >= pr->min);
		    if (pr->flags&FLAG_SELECT_MAX) ok &= (v <= pr->max);
		}
		else { int4 v;		/* integer */
		    v = atoi(p);
		    if (oqs == 0x10) {	/* Sorted query: mask 2=end */
			if (v > *((int *)&pr->max)) oqs |= 2;
			if (v < *((int *)&pr->min)) oqs |= 1;
		    }
		    if (pr->flags&FLAG_SELECT_MIN) ok &=
			(atoi(p) >= *((int *)&pr->min));
		    if (pr->flags&FLAG_SELECT_MAX) ok &=
			(atoi(p) <= *((int *)&pr->max));
		}
		if (pr->flags&FLAG_SELECT_NOT) ok ^= 1;
		p[i] = x;
	    }
	    if (!ok) continue;
	}

        for (ar=arows, colno=0; *ar; ar++) {
	    pr = *ar;
	    if (pr->ano == 0) break;		/* Ending columns to ignore */
	    if (!pr->Cfmt[1]) continue;		/* No format -- ignore! */
	    if (sep/*d*/ && (colno>0) && (!tdisp)) putchar(*sep/*d*/); colno++;
	    if(opar) printf("%s=", pr->type);
	    p = record + pr->opos;
	    if ((pr->flags&FLAG_CONVERT) == 0) {	/* Simple copy */
		if (optb) {			/* Remove useless blanks */
		    if (pr->form[0] == 'A') {	/* Remove trailing blanks */
			for (i=pr->len1-1; (i>=0) && isspace(p[i]); i--) ;
			++i;
		    }
		    else {			/* Remove trailing blanks */
			for (i=pr->len1; (i>0) && isspace(*p); i--, p++) ;
		    }
		}
		else i = pr->len1;
		/* Verify NULL */
		if (pr->null && (pr->form[0] != 'A')) {
		    for (n=0; (n<i) && (p[n] == ' '); n++) ;
		    e = pr->null - n;
		    while ((n<i) && e[n] && p[n] == e[n]) n++;
		    if (i==n) print_null(pr);
		    else fwrite(p, i, 1, stdout);
		}
		else fwrite(p, i, 1, stdout);
		continue;
	    }
	    /* Conversion required */
	    i = pr->len1;	/* Length of field */
	    switch(pr->form[0]) {
	      case 'A':
		/* Count the actual length */
		for (--i; (i>=0) && isspace(p[i]); i--) ;
		++i;
		if (tdisp) {	/* Count the longest string */
		    aJ = (int *)&pr->min;
		    if (i < *aJ) *aJ = i;
		    aJ = (int *)&pr->max;
		    if (i > *aJ) *aJ = i;
		    if (i == 0) pr->cnul += 1;
		    else {	/* Non-blank string:
		            Set the Constant flag for constant text */
			if (pr->value) {
			    if (pr->flags&FLAG_CONSTANT) {
				if (strncmp(pr->value, p, pr->len1))
				    pr->flags &= ~FLAG_CONSTANT;
			    }
			}
			else {
			    pr->value = Malloc(pr->len1+1);
			    strncpy(pr->value, p, pr->len1);
			    pr->value[pr->len1] = 0;
			    pr->flags |= FLAG_CONSTANT;
			}
			i = pr->len1 - i;	/* #tail blanks */
			if (i < ITEMS(pr->cdec)) pr->cdec[i] += 1;
		    }
		}
		else {
		    x = p[i]; p[i] = 0;
		    printf(pr->Cfmt, p);
		    p[i] = x;
		}
	        break;
	      case 'I':
		e = p+i; x = *e; *e = 0;
		/* Verify NULL */
		while (*p == ' ') p++;
		if (pr->null) 	/* Remember: no leading blanks in pr->null */
		     isnull = strcmp(p, pr->null) == 0;
		else isnull = *p == '\0';
		if (isnull) {
		    if (tdisp) pr->cnul += 1;
		    else print_null(pr);
		}
		else {
		    ival = atoi(p);
		    if (tdisp) {
		        aJ = (int *)&pr->min;
		        if (ival < *aJ) *aJ = ival;
		        aJ = (int *)&pr->max;
		        if (ival > *aJ) *aJ = ival;
		    }
		    else printf(pr->Cfmt, ival);
		}
		*e = x;
	        break;
	      default:		/* All the rest is double... */
		e = p+i; x = *e; *e = 0;
		/* Verify NULL */
		while (*p == ' ') p++;
		if (pr->null) 	/* Remember: no leading blanks in pr->null */
		     isnull = strcmp(p, pr->null) == 0;
		else isnull = *p == '\0';
		if (isnull) {
		    if (tdisp) pr->cnul += 1;
		    else print_null(pr);
		}
		else {
		    fval = atof(p);
		    if (tdisp) {
		        aD = (double *)&pr->min;
		        if (fval < *aD) *aD = fval;
		        aD = (double *)&pr->max;
		        if (fval > *aD) *aD = fval;
			n = strloc(p, '.');
			if (p[n]) {
			    for (i=++n; isdigit(p[i]); i++) ;
			    n = (i-n);	/* Number of decimals */
			    if (isalpha(p[i])) { /* Exponent ? */
				if (pr->flags&FLAG_EXPONENT);
				else if (atoi(p+i+1))
				    pr->flags |= FLAG_EXPONENT;
			    }
			}
			else n = 0;
			if (n < ITEMS(pr->cdec)) pr->cdec[n] += 1;
		    }
		    else printf(pr->Cfmt, fval);
		}
		*e = x;
	        break;
	    }
	}
	if (tdisp);
	else { putchar('\n'); if(opar) putchar('\n'); ++edited; }
    }

#if 0
    if (top_lines && (edited > top_lines)) {
	printf("#!+++Stopped after %d lines edited (read = %d lines)\n",
		top_lines, recno);
	return(2);
    }
#else
    if (top_lines) ;
#endif
    else if (recno < records) {	/* Truncation !!! */
	printf("#!+++Stopped after %d lines read (total is %d lines)\n",
	  max_lines, records) ;
	return(2) ;
    }

    if (tdisp) { 		/* Edit the Format */

	/* Guess the Format */
        for (ar=arows; *ar; ar++) {
	    pr = *ar;
	    if (pr->ano == 0) break;
	    if (verbop) printf("%s#%6d %c %-20s ",
		    sep_disp, pr->no, pr->form[0], pr->type);
	    reclen = 15;	/* Decimals for D */
	    switch(pr->form[0]) {
	      case 'A':
		if (verbop) {
		    printf("%12d %12d NULs=%-10d-------#\n#%8s",
		      *(int *)&pr->min, *(int *)&pr->max, pr->cnul, "");
		    for(i=0; i<ITEMS(pr->cdec); i++) {
			if (pr->cdec[i]) printf(" %d=%d",
			  i, pr->cdec[i]);
		    }
		    printf(" tail blanks");
		    if (*(int *)&pr->max == 0) {	/* Always empty */
			printf(" (always blank)");
		        pr->flags |= FLAG_CONSTANT;
		    }
		    else if (pr->flags&FLAG_CONSTANT) {
			printf(" [%s]", pr->value);
		    }
		    putchar('\n');
		}
		sprintf(pr->disp, "A%d", *(int *)&pr->max);
	        break;
	      case 'B':
	      case 'I': case 'J':
		if (verbop) printf("%12d %12d NULs=%-10d-------#\n",
		  *(int *)&pr->min, *(int *)&pr->max, pr->cnul);
		i = *(int *)&pr->min ;
		n = *(int *)&pr->max;
		if (i < 0) {
		    if (n < -i) n = -i;
		    i = 2;
		}
		else i = 1;
		while (n >= 10) { i++; n /= 10; }
		sprintf(pr->disp, "I%d", i);
	        break;
	      case 'E':
	      case 'D': case 'F':
		if (verbop) printf(
		    "%25.16e NULs=%-10d-------#\n#%29s %25.16e Decimals:\n#%8s",
		    pr->min, pr->cnul, "", pr->max, "");
		for (n=ITEMS(pr->cdec)-1; (n>0) && pr->cdec[n] == 0; n--);
		if (verbop) {
		    for(i=0; i<=n; i++) printf(" %d=%d", i, pr->cdec[i]);
		    putchar('\n');
		}
		fval = pr->max;
		if ((pr->min < 0) && (fval < -pr->min)) fval = -pr->min;
		if (pr->flags&FLAG_EXPONENT) ;	/* Keep TDISP=TFORM */
		else if (n == 0) {	/* It's an integer number ! */
		    i = fval;
		    i = digits(i);
		    if (pr->min < 0) i++;
		    sprintf(pr->disp, "I%d", i);
		}
		else if (((pr->max-pr->min)<1000) && (n<=6) && (fval>=0.1)) {
		    i = fval;
		    i = digits(i);	     /* Digits for integer part */
		    if (pr->min < 0) i++;    /* A minus is required ... */
		    sprintf(pr->disp, "F%d.%d", 1+i+n, n);
		}
		else sprintf(pr->disp, "%c%d.%d", pr->form[0],
		    reclen+6+(reclen>7), reclen);
	        break;
	    }
	    printf("TDISP%-3d= '%s'", pr->no, pr->disp);
	    /* V2.6: Indicate a difference ... */
	    if (pr->form[1] && strcmp(pr->form, pr->disp)) {
		i = 19 - strlen(pr->disp); 	/* Blanks for alignement */
		while (--i>=0) putchar(' ');
		printf(" / TFORM%-3d= '%s'", pr->no, pr->form);
	    }
	    putchar('\n');
	}
        if (verbop) printf("%s\n\n", sep_disp); /* putchar('\n'); */

	/* Write for ReadMe */
#if 0
	n = 0;	/* Number of delimiter bytes */
        for (ar=arows; *ar; ar++) {
	    pr = *ar;
	    if (pr->ano == 0) break;
	    printf("#%dX ", n); n = strlen(delim);
	    printf("%-6s %-8s %-12s ", pr->disp,
		    pr->unit ? pr->unit : "---", pr->type);
	    if (pr->form[0] == 'L') printf("[0,1]");
	    else if (pr->flags&FLAG_CONSTANT) switch(pr->form[0]) {
	      case 'A': if (pr->value) printf("[%s]", pr->value); break;
	      case 'B': case 'I': case 'J':
	      case 'I': printf("[%d]", *(int *)&pr->max); break;
	      default:  printf("[%g]", pr->max); break;
	    }
	    if (pr->cnul && (pr->form[0] != 'A')) printf("? ");
	    if (pr->comment) printf("%s", pr->comment);
	    printf("(%d)", pr->no);
	    putchar('\n');
	}
#else
	print_fmt(arows);
#endif
    }
    else if (ometa) {
	print_fmt(arows);
    }

    Free(record);
    if ((n = (reclen*records)%2880)) {	/* Terminate Block */
        n = 2880-n;
	record = Malloc(n);
	fread(record, 1, n, in);
	Free(record);
    }
    return(0);
}

static int mainfits(FILE *in)
/*++++++++++++++++
.PURPOSE  Process  a standard FITS file (or IMAGE extension)
.RETURNS  -1 (error) / 0 / 2 (truncated)
.REMARKS  Added V1.7
	delim1 = delimiter between columns in the case of axis.=... AND opar
-----------------*/
{
  static char Character_data[] = "Character_data";
  static char Multispec_fmt[]  = "";
  char buf[2880];
  int local_swapping = swapping;	/* To be sure once swapping only */
  char *p, *e, *a, *arec=0, *erec=0, *prec=0, *delim1=delim, c;
  int read_completed=0;
  int  n, i, j, k, colno, pixlen, nval=0, ispec=0, iaxis1, dtype;
  int add_sep=opar, iarray, kaxis=9999;
  double v, crpix;
  int *vaxis_choice[ITEMS(axis_choice)] ;	/* Selections */
/*#define vaxis1 vaxis_choice[0]*/
/*#define vaxis2 vaxis_choice[1]*/
  int *vaxis, iaxis=0, ivaxis=-1;
  short *aI; int4 *aJ; int8 *aK; float *aE; double *aD; unsigned char *aB;

    memset(delta, 0, sizeof(delta));
    memset(vaxis_choice, 0, sizeof(vaxis_choice));
    aB = 0; aI = 0; aJ = 0; aK = 0; aE = 0; aD = 0;
    pixlen = bitpix/8; if (pixlen<0) pixlen = -pixlen;
    if (delim1 && (*delim1 == '\n')) delim1 = " ";
    if ((naxis < 0) || (naxis > MAX_AXIS)) {
        fprintf(stderr, "****NAXIS not in range [0,%d]: %d\n",
            MAX_AXIS, naxis) ;
        exit(1) ;
    }
  Check_Multispec:
    if (verbop) fprintf(stderr, "....File %s: bitpix=%d, naxis=%d ",
		input_name, bitpix, naxis) ;
    if (aspec) { 	/* MULTISPEC */
	if (verbop) fprintf(stderr, "multispec");
	if (!fmt[1]) fmt[1] = Multispec_fmt;	/* 2nd is not real */
	if ((p = strstr(keys[iWAT+1].value, "label="))
	  && (!keys[iNAXIS1+4].value)) {
	    p += 6;
	    if (*p == '"') p++;
	    keys[iNAXIS1+4].value = Strdup(p);
	    for (p=keys[iNAXIS1+5].value; isgraph(*p) && (*p != '"'); p++) ;
	    *p = 0;
	}
	if ((p = strstr(keys[iWAT+1].value, "units="))
	  && (!keys[iNAXIS1+5].value)) {
	    p += 6;
	    if (*p == '"') p++;
	    keys[iNAXIS1+5].value = Strdup(p);
	    for (p=keys[iNAXIS1+5].value; isgraph(*p) && (*p != '"'); p++) ;
	    *p = 0;
	}
	n = 0;
	for (p=strstr(keys[iWAT+2].value, " spec"); p; p=strstr(p, " spec")) {
	    i = atoi(p+5); if (i<1) i=n; else --i; n++; /* Spec# */
	    if (i>=nspec) {
		fprintf(stderr, "#***Multispec: invalid WAT002 spec%d (>%d)\n",
			i+1, nspec);
		return(-1);
	    }
	    p += strloc(p, '"'); if (*p) p++; while (*p == ' ') p++;
	    aspec[i].ap    = atoi(p); p += skip_num(p);
	    aspec[i].beam  = atoi(p); p += skip_num(p);
	    aspec[i].dtype = atoi(p); p += skip_num(p);
	    aspec[i].lam   = a2f(p);  p += skip_num(p);
	    aspec[i].dlam  = a2f(p);  p += skip_num(p);
	    aspec[i].npix  = atoi(p); p += skip_num(p);
	    aspec[i].z     = a2f(p);  p += skip_num(p);
	    aspec[i].aplim[0] = a2f(p);  p += skip_num(p);
	    aspec[i].aplim[1] = a2f(p);  p += skip_num(p);
	    aspec[i].function = 0;
	    if (!fmt[0]) fmt[0] = Strdup(guess_fmt( aspec[i].lam,
			aspec[i].lam+(aspec[i].dlam*aspec[i].npix),
			aspec[i].dlam, 0));
#if 0
	    if (aspec[i].dtype != 0) {	/* Non-linear dispersion */
		fprintf(stderr, "#***Non-linear multispec, can't process***\n");
		return(-1);
	    }
#else
	    if ((dtype = aspec[i].dtype)) {
	        if ((dtype < 0) || (dtype >= ITEMS(multiscale))) {
		    fprintf(stderr, "#***Bad multispec#%d (dtype=%d)\n",
			    i, dtype);
		    return(-1);
	        }
		if (dtype == 2) {	/* Function: w olam ftype param coeff */
		    aspec[i].function = get_multifunction(p);
		    j = aspec[i].function->ftype;
		    k = aspec[i].function->order;
		    fprintf(stderr, "#...multispec#%d (%s) as %s[%d]\n",
			i, multiscale[dtype], multifunct[j], k);
		}
		else fprintf(stderr, "#...multispec#%d (%s)\n",
			i, multiscale[dtype]);
	    }
#endif
	    if (aspec[i].z) fprintf(stderr,
		"#+++z-factor (1+%f) ignored in multispec\n", aspec[i].z);
	}
	/* IS NOT A MULTISPEC in case lam/dlam=0 */
	if (aspec[0].dlam == 0) {	/* Ignore the MULTISPEC */
	    if (verbop) fprintf(stderr, "\n");
	    fprintf(stderr, "#+++ Ignore the multispec [%s=%s]\n",
		keys[iWAT].name, keys[iWAT].value);
	    free(aspec);
	    aspec = 0;
	    if (fmt[1] == Multispec_fmt) fmt[1] = (char*)0;
	    goto Check_Multispec;
	}
	else {
	    crval[0] = aspec[0].lam;
	    delta[0] = aspec[0].dlam;
	}
    }
    /*  V2.9: naxis=1 and pixlen = 8 and Character ==> write as Char data */
    else if ((naxis==1) && (bitpix==8) && (keys[0].comment) &&
	    (strncasecmp(keys[0].comment, "character", 9) == 0))
	    fmt[0] = fmt[1] = Character_data;

    p=buf;

    for (n=1, i=iNAXIS1, k=0, c='('; k < naxis; k++, i += perAXIS) {
        dim[k] = *(int *)(keys[i].value) ;
	if (dim[k]==0) {
	    fprintf(stderr, "#+++%s %d ==> forced dim[%d]=1.\n",
		keys[i].name+1, dim[k], k+1);
	    *(int *)(keys[i].value) = 1;
	    dim[k] = 1;
	}
	sprintf(p, k?"x%d":"%5d", dim[k]); p += strlen(p);	/* -D */
	dimm[k] = n;	/* @ = @0 + Sigma(dimm[i]*index[i] */
	n *= dim[k] ;
	if (verbop) fprintf(stderr, "%c%d", c, dim[k]), c = 'x' ;
	if (delta[k] == 0) {
            crpix = crval[k] = 0 ;
            delta[k] = 1 ;
            if (keys[i+1].value) crpix    = *((double *)(keys[i+1].value)) -1. ;
            if (keys[i+2].value) crval[k] = *((double *)(keys[i+2].value)) ;
            if (keys[i+3].value) delta[k] = *((double *)(keys[i+3].value)) ;
	    /* V1.9: CDELT may be replaced by CDi_i */
	    else if (keys[iCD+k].value)
		delta[k] = *((double *)(keys[iCD+k].value)) ;
	    crval[k] -= crpix*delta[k] ;
	}
	if ((type[k] = keys[i+4].value)) ; else type[k] = "(?)" ;
	if ((unit[k] = keys[i+5].value)) ; else unit[k] = "" /*"---"*/;
	xaxis[k] = crval[k] ;		/* Initial Values */
	v = dim[k]*delta[k] ;		/* Amplitude variation */
	/* if (axis1||axis2) continue;	--  Removed V3.3 */
	/* Free the allocated format(s) */
	if (fmt[k] && (sfmt[k]&2)) {
	    free(fmt[k]); fmt[k] = (char *)0 ;
	    sfmt[k] = 0;
	}
        if (!fmt[k]) {
	    fmt[k] = Strdup(guess_fmt(xaxis[k], xaxis[k]+v, delta[k], 
			keys[i+4].value));
	    sfmt[k] |= 2;
	}
    }
    vaxis = (int *)0;
    for (i=0; i<ITEMS(axis_choice); i++) {
	if (!axis_choice[i]) continue;
	if (verbop) fprintf(stderr, " (axis%d=%s)\n    ", i, axis_choice[i]);
	/* Convert to a list of columns -- lower limit is crval/delta */
	if (delta[i] != 1.0) {
	    fprintf(stderr, 
		"#***-axis%d has CDELT%d=%g (requires CDELT%d=1)!\n",
		    i, i, delta[i], i);
	    exit(1);
	}
	j = crval[i];
	vaxis_choice[i] = get_range(axis_choice[i], j, j+(dim[i]-1));
	if (*vaxis_choice[i] > ITEMS(fmt)) {
	    fprintf(stderr, "****axis%d limited to %d (%d asked)\n",
		    i, (int)(ITEMS(fmt)), *vaxis_choice[i]);
	    *vaxis_choice[i] = ITEMS(fmt);
	}
	if (vaxis) fprintf(stderr, "#***Only one (-vaxis.=) possible\n");
	else { ivaxis=i; vaxis = vaxis_choice[i]; }
    }
    k = naxis;				/* For the value of the array */
    dim[k] = 0;
    /* if (vaxis1||vaxis2) k = 0;	-- Removed V3.3 */
    if (fmt[k] && (sfmt[k]&2)) {
	free(fmt[k]); fmt[k] = (char *)0;
	sfmt[k] = 0;
    }
    if (!fmt[k]) {			/* Guess Format... */
	if (bitpix < 0) {		/* Floating-point  */
	    p = bitpix < -32 ? "%20.12e" : "%12.5e";
	    fmt[k] = Strdup(p);
	}

	else {
	    v = pow8(len1el)*bscale;
	    if (bitpix <= 8) p = guess_fmt(bzp, bzp+v, bscale, 0);
	    else p = guess_fmt(bzp-(v/2), bzp+(v/2), bscale, 0);
	    fmt[k] = Strdup(p);
	    sfmt[k] |= 2;		/* Need to free!   */
	}
	if (k != naxis) fmt[naxis] = Strdup(fmt[k]);
    }
    /*vaxis = vaxis2 ? vaxis2 : vaxis1;*/
#if 0	/* Removed V3.3 */
    if (vaxis) {			/* Propagate the format */
	for (k=1; k<=vaxis[0]; k++) {
	    if (fmt[k] && (sfmt[k]&2)) {
		free(fmt[k]); fmt[k] = (char *)0;
		sfmt[k] = 0;
	    }
	    if (!fmt[k]) fmt[k] = strdup(fmt[0]);
	}
    }
#endif

    if (verbop) {
	fprintf(stderr, "=%d)\n----", n) ;
	if (ivaxis>=0) {
	    i = ivaxis;
	    vaxis = vaxis_choice[i];
	    fprintf(stderr, "axis%d=", i);
	    /* List the values chosen */
	    for (k=1; k<=vaxis[0]; k++) {
		if (k>1) fputc(',', stderr);
		fprintf(stderr, "%d", vaxis[k]);
	    }
	}
	else {
	    for (k=0; k<naxis; k++) fprintf(stderr,
	        " %s[%s]%s", type[k], unit[k], fmt[k]) ;
	    fprintf(stderr, " (Value)%s", fmt[k]) ;
	}
	if (pcount) fprintf(stderr, " + %dPcount", pcount);
	fputc('\n', stderr);
    }
    total_lines = n ;		/* Product of all dimensions */
#if 0
    if (vaxis2) total_lines /= dim[1];
    else if (vaxis1) total_lines /= dim[0];
#else
    if (ivaxis>=0) total_lines /= dim[ivaxis];
#endif

#if 0
    /* Change the dimensions (remove those with 1) */
    if (!op1) for (k=0;  k < naxis; k++) {
	if (dim[k] > 1) continue;
	if (!oDim) {		/* Show the constants (not in -D opion) */
	    printf("#%s = ", type[k]);
	    printf(fmt[k],  xaxis[k]);
	    printf(" %s",    unit[k]);
	    putchar('\n');
	}
	/* --- NOTE: The TYPE, DISP, NULL, UNIT are not freed... */
	for (i=iNAXIS1+k*perAXIS; i<iMAXDIM-perAXIS; i++)
	    keys[i].value = keys[i+perAXIS].value;
	while (i<iMAXDIM) keys[i++].value = (char *)0;

	for (i=k; i<ITEMS(dim)-1;   i++)   dim[i] = dim[i+1];   dim[i]=0;
	for (i=k; i<ITEMS(type)-1;  i++)  type[i] = type[i+1];  type[i]=0;
	for (i=k; i<ITEMS(crval)-1; i++) crval[i] = crval[i+1]; crval[i]=0;
	for (i=k; i<ITEMS(delta)-1; i++) delta[i] = delta[i+1]; delta[i]=0;
	for (i=k; i<ITEMS(xaxis)-1; i++) xaxis[i] = xaxis[i+1]; xaxis[i]=0;
	for (i=k; i<ITEMS(unit)-1;  i++)  unit[i] = unit[i+1];  unit[i]=0;
	for (i=k; i<ITEMS(unit)-1;  i++)   fmt[i] =  fmt[i+1];  fmt[i]=0;
	if (k<ivaxis) {
	    for (i=k; i<ITEMS(vaxis_choice)-1;   i++) 
		vaxis_choice[i] = vaxis_choice[i+1];
	    vaxis_choice[i] = 0;
	    ivaxis--;
	}
	--naxis; --k;
    }
#endif
    if (oDim) {
	printf("%3dD%c%-12s", naxis, aspec ? '-' : ' ', buf);
	if (input_name) printf(" %s", input_name);
	if (xtno>0) printf("#%d", xtno);
	putchar('\n');
    }

    /* Print the titles (headers); fixed V3.01 */
    /* kaxis indicates which value should be printed */
#if 0
    if (vaxis2) { vaxis = vaxis2; iaxis = 2; kaxis=0; }
    else        { vaxis = vaxis1; iaxis = 1; kaxis=1; }
#else
    if (ivaxis>=0) {
	vaxis = vaxis_choice[ivaxis];
	iaxis = ivaxis+1;
	kaxis = 0; if (kaxis == ivaxis) kaxis++;
    }
#endif
    /*vaxis = vaxis2 ? vaxis2 : vaxis1;
    iaxis = 1 + (&vaxis - vaxis_choice); */

    if (tdisp) {		/* Print out the formats */
	if (vaxis) for (k=1; k<=vaxis[0]; k++) {
	    i = 0;		/* Bytes written */
	    if (k<vaxis[0]) i += printf("#X%-2d", k+1);
	    else            i += printf("#===");
	    i += printf(" %s", fmt[k]);
	    while (i<12) { putchar(' '); i++; }
	    i += printf(" %s" , unit[k] ? unit[k] : "---");
	    while (i<22) { putchar(' '); i++; }
	    i += printf(" %s", type[k] ? type[k] : "Value");
	    if (k<vaxis[0]) {
	        i += printf("[%d]", dim[k]);
	        while (i<40) { putchar(' '); i++; }
	        i += printf(" [%g,%g]", crval[k], crval[k]+dim[k]*delta[k]);
		while (i<62) { putchar(' '); i++; }
		printf(" dx=%g", delta[k]);
	    }
	    else {
	        i += printf("[]");
	        while (i<40) { putchar(' '); i++; }
		printf(" (bitpix=%d)", bitpix);
	    }
	    putchar('\n');
	}
	else for (k=colno=0; k <= naxis; k++) {
	    if (!(fmt[k][0]&fmt[k][1])) 	/* Format is just "%" */
		continue;
	    i = 0;		/* Bytes written */
	    if (k<naxis) i += printf("#X%-2d", k+1);
	    else         i += printf("#===");
	    i += printf(" %s", fmt[k]);
	    while (i<12) { putchar(' '); i++; }
	    i += printf(" %s" , unit[k] ? unit[k] : "---");
	    while (i<22) { putchar(' '); i++; }
	    i += printf(" %s", type[k] ? type[k] : "Value");
	    if (k<naxis) {
		i += printf("[%d]", dim[k]);
		while (i<40) { putchar(' '); i++; }
	        i += printf(" [%g,%g]", crval[k], crval[k]+dim[k]*delta[k]);
		while (i<62) { putchar(' '); i++; }
		printf(" dx=%g", delta[k]);
	    }
	    else {
		i += printf("[]");
		while (i<40) { putchar(' '); i++; }
		printf(" (bitpix=%d)", bitpix);
	    }
	    putchar('\n');
	}
	putchar('\n') ;
    }

    if (ohead) {
	if (vaxis) {
	    if (opar)  {	/* V3.3: Give the value along the axis */
		print_head(fmt[kaxis], type[kaxis]);
		add_sep = 1;
	    }
	    else add_sep = 0;
	    for (k=1; k<=vaxis[0]; k++) {
		/* if (k>1) printf("%s", delim); -- V3.3 */
	        if (add_sep) printf("%s", delim1) /*putchar(' ')*/; add_sep=1;
	        sprintf(buf, "axis%d=%d", iaxis, vaxis[k]);
	        print_head(fmt[naxis/*k-1*/], buf);
	    }
	}
	else for (k=colno=0; k <= naxis; k++) {
	    /* Ignore single-element axis */
	    if (dim[k]==1) { if (!op1) continue; }
	    if (!(fmt[k][0]&fmt[k][1])) /* Format is just "" or "%" */
		continue;
	    if (colno>0) /*putchar(' ') */ printf("%s", delim1); /* V3.3 */
	    print_head(fmt[k], k<naxis ? type[k] : "Value");
	    colno++;
	}
	putchar('\n') ;
        /* for (k=0; k < naxis; k++) print_head(fmt[k], unit[k]) ;
	   print_head(fmk[k], "Value") ; putchar('\n') ;
	*/
	if (vaxis) {
	    if(opar) { 	/* V3.3: Give the value along the axis */
		print_head(fmt[kaxis], (char *)0);
		add_sep=1;
	    }
	    else add_sep=0;
	    for (k=1; k<=vaxis[0]; k++) {
	        if(add_sep) printf("%s", delim1) /*putchar(' ')*/; add_sep=1;
		/* if (k>1) printf("%s", delim); -- V3.3 */
	        print_head(fmt[naxis/*k-1*/], (char *)0);
	    }
	}
	else for (colno=k=0; k <= naxis; k++) {
	    if (!(fmt[k][0]&fmt[k][1]))	/* Format is just "%" */
		continue;
	    if (colno>0) /*putchar(' ')*/ printf("%s", delim1); /* V3.3 */
	    print_head(fmt[k], (char *)0) ;
	    colno++;
	}
	putchar('\n') ;
    }

    if ((odata==0) || tdisp)  	{	/* Option -nodata */
        for (i=0; i<ITEMS(vaxis_choice); i++) {
            if (vaxis_choice[i]) free(vaxis_choice[i]);
        }
	return(2);
    }

    p = e = buf;		/* Nothing read (yet) */
    i = j = k = 0 ;		/* Coordinates */


    /* Skip the data */
    iarray = nskip%dim[0];	/* V3.3: position in array */
    if (nskip>0) {
	total_lines -= nskip;

	/* Compute the coordinates (values on the axes) after the skip */
	n = nskip;
	i = n%dim[0]; n /= dim[0];
	if (dim[1]) { j = n%dim[1]; n /= dim[1]; }
	if (dim[2]) { k = n%dim[2]; n /= dim[2]; }
	xaxis[0] = crval[0] + i * delta[0];
	xaxis[1] = crval[1] + j * delta[1];
	xaxis[2] = crval[2] + k * delta[2];

	n = nskip;
        while (n>0) {
	    if (p >= e) p = buf, e = p + read_2880(in, buf);
	    if (p >= e) {
		if (p == e) fprintf(stderr,
		  "****Unexpected EOF (after %d*2880) of FITS file: %s\n",
		  blocno, input_name);
		exit(1);
	    }
	    i = (e-p)/len1el;	/* Number of values in the buffer read */
	    if (i > nskip) i = nskip;
	    n -= i;
	    p += i*len1el;
	}
    }

    /* Position of first  non-degenerate axis (normally 0): */
    for (iaxis1=0; dim[iaxis1]==1; iaxis1++) ;
    /* How many values in array: */
    for (nval=1, k=0; k<naxis; k++) nval *= dim[k];
    if (vaxis) { 	/* Edit by line: read all */
	int bytes = nval*pixlen;
	arec = Malloc(bytes);
	erec = arec + bytes;
	n = read_fits(in, arec, bytes);
	if (n >= bytes) read_completed = 1;
	prec = arec;
	erec = arec + n;
    }

    /* Read the Data (n bytes), and write out the values.
       Remind: first axis is varying most rapidly !
    */

  Edit_Element:
    n = total_lines;
    if (n > max_lines) n = max_lines ;
    while (--n >= 0) {
	colno = 0;
      Read_Next:
	if (read_completed == 0) {	/* Still things to read */
            if (p >= e) p = buf, e = p + read_2880(in, buf) ;
	    if (p >= e) {
	        if (p == e) fprintf(stderr,
		    "****Unexpected EOF (after %d*2880) of FITS file: %s\n",
		    blocno, input_name);
	        if (vaxis)  free(arec);
                for (i=0; i<ITEMS(vaxis_choice); i++) {
                    if (vaxis_choice[i]) free(vaxis_choice[i]);
                }
	        /*if (vaxis1) free(vaxis1);*/
	        /*if (vaxis2) free(vaxis2);*/
	        return(-1);
	    }
	}
	if (fmt[0] == Character_data) {
	    if (isascii(*p)) putchar(*p);
	    else if (*p) putchar('?');
	    p++;
	    continue;
	}
	if (vaxis) {		/*  Copy FITS buffer into arec */
	    if (!read_completed) {
	        k = erec - prec;	/* Remaining bytes to copy */
	        if ((p+k) >= e) k = (e-p);
	        memcpy(prec, p, k);
	        p += k;
	        prec += k;
	        if (prec < erec) goto Read_Next;
		else read_completed++;
	    }
	    a = arec;		/* Address of pixels to edit */
	}
	else {
	    /* V3.3: insert a blank line when first dim is exhausted */
	    if (opar && (iarray == dim[iaxis1])) { add_sep=1; iarray=0; }
	    if (add_sep) { putchar('\n'); add_sep=0; }
	    iarray++;
	    for (k=colno=0; k<naxis; k++) {
		if ((dim[k]==1)&&(!op1)) continue;
	        if (!(fmt[k][0]&fmt[k][1]))	/* No edition (multispec) */
		    continue;
	        if (colno>0) /*putchar(' ')*/ printf("%s", delim1); /* V3.3 */
	        printf(fmt[k], xaxis[k]) ;
	        colno++;
	    }
	    a = p;		/* Address of pixels to edit */
	    nval = 1;		/* Number of values */
        }
        switch(bitpix) {
          case 8:
            aB = (unsigned char *)a;
	    if (vaxis) ;
	    else v = *aB;	break ;
          case 16:
            aI = (short *)a;
	    if (local_swapping) swap1(aI, nval) ;
	    if (vaxis) ;
	    else v = *aI;	break ;
          case 32:
            aJ = (int4 *)a;
	    if (local_swapping) swap(aJ, nval) ;
	    if (vaxis) ;
	    else v = *aJ;	break ;
          case 64:
            aK = (int8 *)a;
	    if (local_swapping) swap8((double *)aK, nval) ;
	    if (vaxis) ;
	    else v = *aK;	break;
          case -32:
            aE = (float *)a;
	    if (local_swapping) swap((int4 *)aE, nval) ;
	    if (vaxis) ;
	    else v = *aE; 	break;
          case -64:
            aD = (double *)a;
	    if (local_swapping) swap8(aD, nval) ;
	    if (vaxis) ;
	    else v = *aD;	break;
          default:
            fprintf(stderr, "****Bad bitpix: %d\n", bitpix) ;
            exit(1) ;
        }
	if (vaxis) { int jm;	/* Edit the array of pixels */
	    jm = ivaxis == iaxis1 /*vaxis2*/ ? dim[iaxis1] : 1;
	  Next_Row:
	    if (opar) {
		printf(fmt[kaxis], xaxis[kaxis]); add_sep=1;
	        /* Prepare Next value to edit */
                xaxis[kaxis] += delta[kaxis] ;
	    }
	    else add_sep=0;
	    for (k=1; k<=vaxis[0]; k++) {
		if(add_sep) /*putchar(' ')*/ printf("%s", delim1); add_sep=1;
                /* if (k>1) putchar(' '); -- printf("%s", delim) ; -- V3.3 */
		j = dimm[ivaxis]*vaxis[k]; /*i = k-1; j = (vaxis[k]-1)*jm;*/
	        switch(bitpix) {
	          case 8:    printf(fmt[naxis], aB[j]);	break;
	          case 16:   printf(fmt[naxis], aI[j]);	break;
	          case 32:   printf(fmt[naxis], aJ[j]);	break;
	          case 64:   printf(fmt[naxis], aK[j]); break;
	          case -32:  printf(fmt[naxis], aE[j]);	break;
	          case -64:  printf(fmt[naxis], aD[j]);	break;
	        }
	    }
	    if (ivaxis>0/*vaxis2*/) {	/* Do it on 2nd axis */
		if (--n >= 0) {
		    putchar('\n') ;
		    aB++; aI++; aJ++; aK++; aE++; aD++;
		    goto Next_Row;
		}
	    }
	}
	else if ((fmt[k][0]&fmt[k][1])) {	/* Valid format */
            if (colno>0) /*putchar(' ')*/ printf("%s", delim1) ; /* V3.3 */
	    if (keys[iNULL].value) {	/* V1.62: Verify NULL */
	        if (*(double *)(keys[iNULL].value) == v) v = 0./0.;
	    }
	    v = bscale*v + bzp;
	    if (!isFinite(v)) print_nan(fmt[naxis]);
            else printf(fmt[naxis], v);
	    /* Prepare Next value to edit */
            i++, xaxis[0] += delta[0] ;
	    if (i == dim[0]) {
	        i = 0 ; xaxis[0] = crval[0] ;
	        j++ ;  xaxis[1] += delta[1] ;
	        if (j == dim[1]) {
	            j = 0 ; xaxis[1] = crval[1] ;
	            k++ ;  xaxis[2] += delta[2] ;
	        }
	    }
            p += len1el ;
	}
	if (aspec && ((n%dim[0] == 0))) {  /* Multispec: Change ! */
	    ispec++;
	    if (ispec < nspec) {
		putchar('\n');
		crval[0] = aspec[ispec].lam;
	        delta[0] = aspec[ispec].dlam;
		xaxis[0] = crval[0];
	    }
	}
	putchar('\n') ;
    }
    /* V3.3: Maybe there are continuation(s)? */
    if (vaxis) {
	vaxis += vaxis[0]+1;
	if (vaxis[0]) { 	/* A continuation */
	    putchar('\n');
	    prec = arec;
	    local_swapping = 0;
	    xaxis[kaxis] = crval[kaxis];
	    goto Edit_Element;
	}
    }

    if (vaxis)  free(arec);
    for (i=0; i<ITEMS(vaxis_choice); i++) {
        if (vaxis_choice[i]) free(vaxis_choice[i]);
    }

    if (total_lines > max_lines) 	/* Truncation !!! */
	return(2);
    return(0);
}


/*==================================================================
		Process a file
 *==================================================================*/
static double sky_image(double *center, int *axes)
/*++++++++++++++++
.PURPOSE  Find center of image + scale + indexes of axes in "keys"
.RETURNS  Scale in arcsec/pix -- 0 if not a sky image
.REMARKS  Added V2.8
-----------------*/
{
  double scale = 0;
  double *ascale;
  int axRA=-1, axDE=-1, i, k;
    /* Are there axes for an image ? Not necessary 1 and 2 ... (V3.11) */
    for (i=iNAXIS1+4, k=0; keys[i].value && (i<iMAXDIM); i += perAXIS) {
	if (starts_by(keys[i].value, "RA-"))      axRA=i-4;
	else if (starts_by(keys[i].value, "DEC")) axDE=i-4;
	else if (starts_by(keys[i].value, "GLON")) axRA=i-4;
	else if (starts_by(keys[i].value, "GLAT")) axDE=i-4;
	else continue;
	k++;
    }
    /* I have an image on the sky IF k==2 */
    if (k==2) {
	if (axes) { axes[0] = axRA; axes[1] = axDE; }
	/* Takes scale from Dec */
	ascale = (double *)keys[axDE+3].value;
	if (!ascale) {
	    ascale = (double *)keys[iCD].value;
	    scale = *ascale * *((double *)keys[iCD+1].value)
		- *((double *)keys[iCD2].value) *
		  *((double *)keys[iCD2+1].value) ;
	    scale = sqrt(fabs(scale));
	}
	if (!ascale) {
	    fprintf(stderr, "#+++sky_image: no CDELT2, no CD2_2 ? \n");
	    return(0);
	}
	if (scale==0) scale = fabs(*ascale);
	if (center) { int n; double cdel, rpix;	/* Compute the center */
	    n = *((int *)(keys[axRA].value));
	    rpix = *((double *)(keys[axRA+1].value));		/* CRPIX */
	    /* V2.92: CDELT may be replaced by CD matrix */
	    cdel = keys[axRA+3].value ?
		*((double *)(keys[axRA+3].value)) :		/* CDELT */
		*((double *)(keys[iCD+0].value));		/* CD1_1 */
	    center[0] = *((double *)(keys[axRA+2].value))	/* CRVAL */
		      + cdel*(0.5*(n+1) - rpix);
	    n = *((int *)(keys[axDE].value));
	    rpix = *((double *)(keys[axDE+1].value));		/* CRPIX */
#if 0
	    cdel = keys[axDE+3].value ?
		*((double *)(keys[axDE+3].value)) : 		/* CDELT */
		*((double *)(keys[iCD+1].value));		/* CD2_2 */
#else
	    cdel = *ascale;
#endif
	    center[1] = *((double *)(keys[axDE+2].value))	/* CRVAL */
		      + cdel*(0.5*(n+1) - rpix);
	}
    }
    return(scale*3600.);
}

static int print_axis(FITSKEY *keys, FITSKEY *CDkey)
/*++++++++++++++++
.PURPOSE  Print the summary of one axis
.RETURNS  0 (no axis) / number of bytes written out
.REMARKS  Added V2.96
-----------------*/
{
  double *adouble, delt, rpix, rval, lim[2]; /* 0=ref, 1=min, 2=max */
  int dim, bytes=0;

    dim = *(int *)(keys[0].value);
    if (dim < 1) return(0);
    if (keys[4].value) bytes += printf("%s=", keys[4].value);
    adouble = (double *)keys[1].value; rpix = adouble ? *adouble : 1.0;
    adouble = (double *)keys[2].value; rval = adouble ? *adouble : 0.0;
    adouble = (double *)keys[3].value; /* CDELT */
    if (!adouble) adouble = (double *)CDkey[0].value;
    delt = adouble ? *adouble : 1.0;
    lim[0] = rval + delt*(1.0-rpix);
    lim[1] = rval + delt*(dim-rpix);
    if (lim[0]>lim[1]) { double x;
	x = lim[0]; lim[0] = lim[1]; lim[1] = x;
	delt = -delt;
    }
    bytes += printf("%g", lim[0]);
    if(lim[1]>lim[0]) bytes += printf("/%g(%g)", lim[1], delt);
    if (keys[5].value) bytes += printf("%s", keys[5].value);
    return(bytes);
}

static int toc(FILE *in)
/*++++++++++++++++
.PURPOSE  Write a TOC summary
.RETURNS  Number of FITS files described. -1 if error
.REMARKS  otoc&1 ==> headline of TOC; otoc&2 ==> list headers
-----------------*/
{
  double center[2], scale;
  char fmt[20];
  int nf=0;
  int4 pos;
  int c, i, k, nb, axes[2], nfmax=999999999;

    /* additional header are not taken into account... */
    center[0] = center[1] = 0;
    axes[0] = axes[1] = -1;
    if (aheader) fprintf(stderr,
	    "++++additional header (-a option) ignored in -toc option\n");

    /* -xten asks to stop at specified extension */
    if (axten && isdigit(*axten))
	nfmax = atoi(axten);

    pos = blocno;
    if (input_name) printf("#--- %8s\n", input_name);
    if (otoc&1) printf("#ext %8s bpx %-18s%6s extname\n",
	"xtension", "(dim)", "x2880");
    if (otoc&2) printf("%s\n", ruler);
    while ((nf<=nfmax) && (get_header(in, 0) >= 0)) {
        if (otoc&2) putchar('\n');
    	printf("#%03d %8s %3d ", nf,
		XTENSION ? XTENSION : "main", bitpix);
	memset(dim, 0, sizeof(dim));
	putchar('(');
	if (pcount) nb += printf("%d+", pcount);
        for (nb=0, i=iNAXIS1, k=0, c=0; k < naxis; k++, i += perAXIS) {
	    if (c) nb++, putchar(c);
            dim[k] = *(int *)(keys[i].value) ;
	    nb += printf("%d", dim[k]);
	    c = 'x';
	}
	if (keys[iXTENSION].value && keys[iTFIELDS].value) {
	    if (TFIELDS)
	      nb += printf("/%df", TFIELDS);
	}
	putchar(')');
	if (gcount>1) nb += printf("x%d", gcount);
	putchar(' ');
	sprintf(fmt, "%%%ddb", 20-nb);
	/* while (nb<16) putchar(' '), nb++; */
	nf++;

	if (nf > nfmax) {
	    c = strloc(fmt, 'd'); fmt[strloc(fmt, 'd')] = 's';
	    printf(fmt, "...");
	}
	else if (skip_data(in)) {
	    printf("(%d)\n", blocno-pos);
	    fprintf(stderr, "****Error skip_fits\n");
	    exit(1);
	}
	else printf(fmt, blocno-pos);
	scale = sky_image(center, axes); /* V2.8: write center+scale */
	if (EXTNAME) { i = strlen(EXTNAME);
	    printf(" %s", EXTNAME);
	    while (i<namelen) { putchar(' '); i++; }
	    if (namelen<i) namelen = i;
	}
	else if (aspec && aspec->npix &&  aspec->dlam) /* Mod. V2.82 */
	    printf(" %s", "MULTISPEC");
	if ((axes[0]>0)&&(axes[1]>0)) {
	    if (!EXTNAME) {
		for(i=0; i<namelen; i++) putchar(' ');
		printf("  ");
	    }
#if 0
	    printf(" [%09.5f%+09.5f,scale=%.3g\"/pix]",
		center[0], center[1], scale); 
	    for (i=3; i<=naxis; i++) {
		printf(" [");
		print_axis(keys+iNAXIS1+(i-1)*perAXIS, keys+iCD+(i-1));
		printf("]");
	    }
#else
	    i=axes[0]+4;	/* Type of coo? */
	    if (keys[i].value) {
	      if (keys[i].value[2] == '-')    /* RA- DEC */
	         printf(" [%09.5f%+09.5f", center[0], center[1]);
	      else printf(" [%c%09.5f%+09.5f", keys[i].value[0], 
		                          center[0], center[1]);
	      if (scale>0) printf(",scale=%.3g\"/pix", scale);
	      printf("]");
	    }
            for (i=iNAXIS1, k=0; k<naxis; i+=perAXIS, k++) {
		if (i==axes[0]) continue;
		if (i==axes[1]) continue;
		printf(" [");
		print_axis(keys+i, keys+iCD+k);
		printf("]");
	    }
#endif
	}
	else for (i=iNAXIS1, k=0; k<naxis; i+=perAXIS, k++) {
	    printf(" [");
	    print_axis(keys+i, keys+iCD+k);
	    printf("]");
	}
	putchar('\n');
	pos = blocno;
	if (otoc&2) {	/*  Print the Header */
	    printf("%s\n", ruler);
	    print_a2880();
	    /* printf("%s\n", ruler); */
	}
    }
    return(nf);
}

static int trf(FILE *in)
/*++++++++++++++++
.PURPOSE  Execute transformations on an opened FITS file
.RETURNS  0 / 2 (truncated) / 1 (EOF found)
.REMARKS
-----------------*/
{
  char *p;
  static int ntit = 0;		/* Number of files edited */
  int  get_status, ret_status=0;

    /* Decide how extensions to process are specified:
     * EXTNAMEs or extension number ?
     */
    xbyname = 0;		/* -xten specifies ranges of numbers */
    if (axten && *axten) {
	for (p = axten; isdigit(*p) || (*p == '-') || (*p == ','); p++) ;
	if (*p) xbyname = 1;
    }

    /* Read the Header */
    get_status = get_header(in, 1);
    if (get_status<0) return(1);

  Exec1Xtension:
    if ((naxis == 0) && (xtno == 0)) {	/* It's a generic header */
	if(verbop) fprintf(stderr,
	    "....Main header has NAXIS=0, skipped\n");
	if (oDim);		/* Ignore generic header, EXCEPT in -D option */
	else get_status = 0;	/* xten = 1; */
    }

    while (get_status == 0) { 	/* (xtno < xten) */
	if (skip_data(in)) {
	    fprintf(stderr, "****Error skip_fits\n");
	    exit(1);
	}
	/* xtno++; */
	get_status = get_header(in, 1);
    }

    if (get_status < 0) { /* xtno != xten) */
	fprintf(stderr, "****Can't reach Extension#%d: %s\n",
	  xtno, input_name);
	return(-1);
    }

    total_lines = 0;	/* Changed by specific routines */
#if 0
    if (nskip && max_lines)
	max_recno = max_lines + nskip;
    else if (nquery == 0)
	max_recno = max_lines;
#else
    max_recno = max_lines + nskip;
    if (max_recno<0) max_recno = max_lines;
#endif

    /* Write out the name(s) */
    if (ohead || tdisp || verbop) {
	if (ntit) putchar('\n');
	ntit++;
	printf("#---File: %s", input_name);
	if (xtno) printf("[#%d]", xtno);
	if (XTENSION) printf(" (%s)", XTENSION);
	putchar('\n');
    }

    /* Tabular data: use dedicated routine */
    if (XTENSION) {
	if (strcmp(XTENSION, "BINTABLE") == 0)
	    get_status |= bintable(in);
	else if (strcmp(XTENSION, "TABLE") == 0)
	    get_status |= asctable(in);
	else if (strcmp(XTENSION, "IMAGE") == 0)
	    get_status |= mainfits(in);
	else {
	    fprintf(stderr, "****Can't process XTENSION='%s', file: %s\n",
	        XTENSION, input_name);
	    return(-1);
	}
    }
    else 	/* Main FITS file */
	get_status |= mainfits(in);

    /* Verify the status */
    if (get_status < 0) return(-1);
    /* ret_status |= get_status; */

    /* Did a truncation occur ? (already edited)
    if (get_status == 2) {
	printf("#!+++Stopped after %d lines read (total is %d lines)\n",
	    max_lines, total_lines) ;
    }
    */

    /* Other extensions to process ? */
    if (axten && *axten) {
	putchar('\n');		/* Blank line to separate extensions */
	if (skip_data(in)) 	/* Error in this part ... */
	    return(ret_status|1);
        get_status = get_header(in, 1);
        if (get_status<0) return(ret_status);
	goto Exec1Xtension;
    }
    return(ret_status);
}

/*==================================================================
		Program
 *==================================================================*/

/* Note about compressed and indexed FITS files.
 * The grogram "gzindex" compresses a FITS file and generates a "*.gzi"
 * file which looks like:
#       +seek_offset(1.0) grid.fit.gz
#       -v -211F grid.fit
#. Index to be used for a seek in a gzipped file:
#. position in uncompressed file is associated to +seek_offset in gzipped file
#: FITS(2880c) +offset
#u=2880
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
: 1 +0                  / SIMPLE  =            (dummy FITS header)
: 2 +191                / XTENSION= 'BINTABLE' (14352 FITS blocks) 3376x12243x(8)
: 4 +1446
: 215 +231024
: 426 +461359
: 637 +690825
....
; 14355 (last_record)
; ~~~~~~~ +15668741
*/
#define set_param(argname) while (isalpha(*p)) p++; \
    if (*p) { if (*p == '=') p++; } else { p = *++argv; --argc; }\
    if (!p) { fprintf(stderr, "#***bad '-%s' argument\n%s\n", argname, usage);\
	exit(1); }
#define set_numparam(argname, param) set_param(argname) param = atoi(p)
int main (int argc, char **argv)
{
  int   rep, i, ndashes;
  int   files, status;
  FILE *infile = (FILE *)0;
  char  *p, x;

    /* Check the Arguments */
    which_swapping();
    files = status = 0;

    if ((argc < 2) && isatty(0)) {
	fprintf(stderr, "%s\n",
	    "****Missing arguments (fits2a -help for details)");
	exit(1);
    }

    while (-- argc > 0) {
	p = *++argv;
	if (verbop) fprintf(stderr, " %s", p);
	if ((ndashes = *p == '-')) while (p[1] == '-') p++, ndashes++;
	if (*p == '-') switch(*++p) {
	  case 0:		/* stdin */
	    break;
	  case 'v':
	    if ((ndashes>1) || (strncmp(p, "ver", 3) == 0)) {
	        printf("fits2a (CDS) Version %s\n", VERSION);
		exit(0);
	    }
	    else if (p[1] == 'm') {	/* -vmax */
		set_numparam("vmax", vmax);
	    }
	    else { verbop = 1 ; fprintf(stderr, "#...fits2a arguments"); }
	    continue ;
	  case 'h':				/* Help */
	    if (p[1]) {
	        if (p[2] == 'a') { ohead = 1 ; continue ; }
		if (p[2] == 'l') { printf("%s\n%s", usage, help) ; exit(0); }
	    }
	    fprintf(stderr, "****Ambiguous option: -%s\n%s\n", p, usage);
	    exit(1) ;
	  case 'a':				/* Additional header ,axis1 */
	    if ((x = p[1])) {
		if (strncmp(p, "axis", 4)) goto case_default;
		if (isdigit(p[4]));   else goto case_default;
		i = p[4] - '1';			/* 1 or 2 */
		if ((i<0) || (i >= ITEMS(axis_choice)))
					   goto case_default;
		p += strloc(p, '=');
		axis_choice[i] = *p ? p+1 : "1-";
		continue;
	    }
	    set_param("a");
	    if (aheader) Free(aheader);
	    aheader = load_file(*argv, 1);
	    if (!aheader) exit(1);
	    continue;
	  case 'b':				/* Remove blanks   */
	    optb = 1;
	    continue;
	  case 'c':				/* List of columns */
	    set_param("cols");
	    acols = p;
	    if ((*acols == '=') && (!aheader)) {
	        fprintf(stderr, "-cols =a requires the -a option%s\n", usage);
		exit(1);
	    }
	    continue;
	  case 'q':				/* Query on columns */
	    while (isalpha(*p)) { if (*p == 's') oqs=0x10; p++; }
	    set_param("q");
	    if (nquery >= MAX_QUERIES-1) {
		fprintf(stderr, "****Too many -q options (max=%d)\n",
			MAX_QUERIES);
		exit(1);
	    }
	    query[nquery++] = *argv;
	    continue;
	  case 'D':				/* Just dimensions */
	    oDim = 1;
	    odata = 0;
	    continue;
	  case 'd':				/* Delimiter...	*/
	    delim = Strdup(p+1) ;
	    if (*delim == 0) { delim[0] = '\t'; delim[1] = 0; }
	    continue ;
	  case 't':				/* Evaluate Format */
	    if (strncmp(p, "toc", 3) == 0)
		 otoc = 1;
	    else if (strncmp(p, "top", 3) == 0) {
		set_numparam("top", top_lines);
	    }
	    else tdisp = 1;
	    continue;
	  case 'T':
	    otoc = 3;
	    continue;
	  case 'n':				/* Assume -nodata  */
	    odata = 0;
	    max_lines = 0;
	    continue;
	  case 'p':				/* Print the Constant Col. */
	    if (p[1] == '1') op1 = 1;
	    else { 				/* Write as paragraphs */
		opar = 1;
		if (!delim) delim = "\n";
	    }
	    continue;
	  case 'm':				/* Delimiter...	*/
	    if (strncmp(p, "meta", 4) == 0) {
		ometa = 1;
		odata = 0;
		max_lines = 0;
		continue;
	    }
	    set_numparam("max", max_lines);
	    if (max_lines <= 0) max_lines=1, fprintf(stderr,
	      "++++maxlines value '%s' interpreted as 1\n", *argv);
	    continue ;
	  case 's':
	    set_numparam("skip", nskip);
	    continue;
	  case 'x':				/* Xtrension #	*/
	    set_param("xten");
	    if (axten) free(axten);
	    axten = strdup(p);			/* e.g. 0-5,12- */
	    continue;
	    /*
	    if (--argc > 0) xten = atoi(*++argv);
	    if (xten > 0) continue;
	    fprintf(stderr, "****Non-positive -xten number\n%s\n", usage);
	    exit(1);
	    */
          case_default:
	  default:
	    if (isdigit(*p)) {			/* Assume max lines */
		max_lines = atoi(p);
		continue;
	    }
	    fprintf(stderr, "****Unknown option: %s\n", p-1) ;
	    exit(1) ;
	}
	else if (strchr(p, '%')) {		/* A Fmt */
	    rep = 1;
	    if (*p != '%') {
		rep = atoi(p);
		while (isdigit(*p)) p++;
		if (*p != '%') rep = 0;
	    }
	    if (rep <= 0) goto case_isaFile;
	    /* A format "%." means "take default". */
	    if ((p[1] == '.') && (!p[2] )) p = (char *)0;
	    while (--rep >= 0) {
		if (nfmt >= ITEMS(fmt)) {
	            fprintf(stderr, "****Can't have more than %d formats\n",
	                (int)(ITEMS(fmt))) ;
	            exit(1) ;
	        }
		fmt[nfmt] = p;
		if (p) sfmt[nfmt] = 1;
		nfmt++;
	    }
	    continue;
	}

    case_isaFile:
	if (verbop) fprintf(stderr, "[file]\n");
	/* The argument represents a file name. Open it and convert */
	if (input_name) { free(input_name); input_name = (char *)0; namelen=0; }
	if (zindex) { free(zindex); zindex = (char *)0; }
	/* Write a blank line between tables */
	if (files && (oDim == 0)) putchar('\n') ;
	/* File name may include the xtension, the compress suffix, etc... */
	i = strlen(*argv); input_name = malloc(i+12); /* Accept suffix */
	strcpy(input_name, *argv);
	if ((p = strchr(input_name, '#')) ||
	    (p = strchr(input_name, '|'))) {	/* Look for extensions */
	    *(p++) = 0;
	    if (axten) { free(axten);  axten = (char *)0; }
	    axten = strdup(p);
	}
	if ((i>3) && (strcmp(input_name+i-3, ".gz") == 0)) /* Remove suffix */
	    input_name[i-3] = 0;
	if (*input_name == '-') {
	    infile = stdin;
	    if (isatty(0) && isatty(2)) fprintf(stderr,
	      "----FITS file input on the terminal...\n");
	}
	else if ((infile = fzopen(input_name))) {
	    i = strlen(input_name);
	    if (isregfile(infile)) ;
	    else {	/* Load the "gzi" index if existing */
		strcpy(input_name+i, ".gzi");
		zindex = load_file(input_name, 0);
		zpos = 0;
		if (verbop) fprintf(stderr, "#...try load_file(%s): %d bytes\n",
			input_name, zindex ? (int)strlen(zindex) : 0);
		input_name[i] = 0;
	    }
	}
	if (!infile) {
	    fprintf(stderr, "****"); perror(input_name);
	    exit(1);
	}
	if (otoc) status |= toc(infile)<=0 ;
	else status |= trf(infile) ;
	if (infile) fzclose(infile);	/* close(in); */
	files++;
	get_header((FILE *)0, 0);	/* Reinitialize */
	nfmt = 0;		/* Ready for another file... */
    }

    /* Execute on stdin if necessary */

    if (files == 0) {
        input_name = "(stdin)" ;
	if (otoc) status |= toc(stdin)<=0 ;
	else status |= trf(stdin);
	fclose(stdin);
    }

    exit(status<0 ? 1 : status);
}
